#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from htf.daqmx.core import DAQError
from htf.daqmx.task import Task
from ctypes import *  # noqa: F403
from htf.daqmx.api import int32, float64, DAQmx_Val_Volts
from htf.mixins.delegation import DelegateMixin


class AnalogVoltageOutputSingleLine(DelegateMixin):
    """
    ``AnalogVoltageOutputSingleLine`` is used to access a single analog output
    line on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    def __init__(
        self,
        descriptor: str,
        minimum_voltage: float = -10.0,
        maximum_voltage: float = 10.0,
    ) -> None:
        """
        Args:
            descriptor: The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`
            minimum_voltage=-10.0: the minimum voltage.
            maximum_voltage=10.0: the maximum voltage.
        """
        task_name = "AO_" + descriptor.replace("/", "_")
        self.value = 0.0
        self.tmp_value = float64(0.0)
        self.written = int32(0)
        self.read = int32(0)
        self.task = Task(task_name)

        if not isinstance(descriptor, bytes):
            descriptor = descriptor.encode()  # type: ignore

        self.task.CreateAOVoltageChan(
            descriptor,
            None,
            float64(minimum_voltage),
            float64(maximum_voltage),
            DAQmx_Val_Volts,
            None,
        )

    def get(self) -> float:
        """
        Return the last written analog value.
        """

        return self.value

    def set(self, value: float) -> None:
        """
        Set the pin value to ``value``.

        Args:
            value: new state for output pin.
        """
        self.value = value

        self.task.WriteAnalogScalarF64(True, float64(1.0), float64(value), None)
        self.task.StopTask()

    def close(self) -> None:
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self) -> None:
        self.close()

    def __str__(self) -> str:
        return str(self.get())


if __name__ == "__main__":
    ao0 = AnalogVoltageOutputSingleLine("Dev1/ao0", minimum_voltage=0, maximum_voltage=5.0)

    for ii in range(100):
        ao0.set(0.0)
        value = ao0.get()
        print(value)
        # ao1.set(3.0)
        # time.sleep(1.0)
        # ao0.set(3.0)
        # ao1.set(0.0)
        # time.sleep(1.0)
