#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from htf.daqmx.core import DAQError
from htf.daqmx.task import Task
from ctypes import *  # noqa: F403
from htf.daqmx.api import uInt32, int32, DAQmx_Val_ChanForAllLines, float64
from htf.mixins.delegation import DelegateMixin


class DigitalInputSingleLine(DelegateMixin):
    """
    ``DigitalInputSingleLine`` is used to access a single digital input line
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    N = 2

    def __init__(self, descriptor: str) -> None:
        """

        Args:
            descriptor: The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/line0`.
        """
        task_name = "DI_" + descriptor.replace("/", "_")
        self.state = 0x00000000
        self.data = cast((uInt32 * self.N)(), POINTER(uInt32))  # noqa: F405
        self.read = int32(0)
        self.task = Task(task_name)

        if not isinstance(descriptor, bytes):
            descriptor = descriptor.encode()  # type: ignore

        self.task.CreateDIChan(descriptor, None, DAQmx_Val_ChanForAllLines)

    def get(self) -> bool:
        """
        Return the current pin state.
        """
        self.task.ReadDigitalU32(
            self.N,
            float64(1.0),
            DAQmx_Val_ChanForAllLines,
            self.data,
            self.N,
            byref(self.read),  # noqa: F405
            None,
        )
        self.task.StopTask()
        return bool(self.data[0])

    def set(self, state: bool = True) -> None:
        raise NotImplementedError(
            "There is no meaningful implementation for " "a 'set' method in the context of this class."
        )

    def close(self) -> None:
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self) -> None:
        self.close()

    def __str__(self) -> str:
        return str(self.get())


if __name__ == "__main__":
    import time

    l7 = DigitalInputSingleLine("Dev1/line0")
    l8 = DigitalInputSingleLine("Dev1/line1")

    for ii in range(10):
        print(l7.get())
        print(l8.get())
        time.sleep(0.5)

        print(l8.get())
        print(l7.get())
        time.sleep(0.5)
