#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from htf.daqmx.core import DAQError
from htf.daqmx.task import Task
from ctypes import *  # noqa: F403
from htf.daqmx.api import (
    uInt32,
    int32,
    DAQmx_Val_ChanForAllLines,
    float64,
    DAQmx_Val_GroupByChannel,
)
from htf.mixins.delegation import DelegateMixin


class DigitalOutputPort(DelegateMixin):
    """
    ``DigitalOutputPort`` is used to access a digital output port
    on a DAQmx device.

    A DAQmx task is automatically created and deleted in the constructor and
    destructor.
    """

    N = 2

    def __init__(self, descriptor: str) -> None:
        """
        Args:
            descriptor (str): The DAQmx device and line descriptor, eg.
                `cDaq1Mod1/port0`.
        """
        task_name = "DO_" + descriptor.replace("/", "_")
        self.value = 0x00000000
        self.data = cast((uInt32 * self.N)(), POINTER(uInt32))  # noqa: F405
        self.written = int32(0)
        self.read = int32(0)
        self.task = Task(task_name)

        if not isinstance(descriptor, bytes):
            descriptor = descriptor.encode()  # type: ignore

        self.task.CreateDOChan(descriptor, None, DAQmx_Val_ChanForAllLines)

    def get(self) -> int:
        """
        Return the current port state.
        """
        self.task.ReadDigitalU32(
            self.N,
            float64(1.0),
            DAQmx_Val_GroupByChannel,
            self.data,
            self.N,
            byref(self.read),  # noqa: F405
            None,
        )
        return int(self.data[0])

    def set(self, value: int) -> None:
        """
        Set the port to ``value``.

        Args:
            value (float): new value for output port
        """

        self.value = value

        self.data[0] = self.data[1] = value

        self.task.WriteDigitalU32(
            self.N,
            True,
            float64(1.0),
            DAQmx_Val_ChanForAllLines,
            self.data,
            byref(self.written),  # noqa: F405
            None,
        )
        self.task.StopTask()

    def close(self) -> None:
        """
        Close the abstractor by clearing the task.
        """
        try:
            self.task.ClearTask()
        except DAQError:
            pass

    def __del__(self) -> None:
        self.close()

    def __str__(self) -> str:
        return str(self.get())


if __name__ == "__main__":
    a = DigitalOutputPort("Dev1/port1")
    a.set(0b10101)
    print(bin(a.get()))
    exit(0)
    a = DigitalOutputPort("cDaq4Mod1/port0")
    b = DigitalOutputPort("cDaq4Mod1/port0")

    a.set(0x12345678)
    assert a.get() == 0x12345678
    assert b.get() == 0x12345678

    a.set(0)
    assert a.get() == 0
    assert b.get() == 0

    exit(0)
