#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#
from .api import *
import random
import string


class Task:
    """
    ``Task`` automatically creates a DAQmx task.

    Every method that is called does not need the prefix ``DAQmx`` and you must not
    supply the first argument if it is a taskhandle.

    The task is automatically cleared when the destructor is called.
    """

    def __init__(self, name: str) -> None:
        self.name = (
            name.replace(":", "_")
            + "_"
            + "".join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in range(10))
        )
        if not isinstance(self.name, bytes):
            self.name = self.name.encode()  # type: ignore
        self.task_handle = TaskHandle()
        DAQmxCreateTask(self.name, byref(self.task_handle))

    def AddGlobalChansToTask(self, channelNames):
        return DAQmxAddGlobalChansToTask(self.task_handle, channelNames)

    def StartTask(self):
        return DAQmxStartTask(self.task_handle)

    def StopTask(self):
        return DAQmxStopTask(self.task_handle)

    def ClearTask(self):
        return DAQmxClearTask(self.task_handle)

    def WaitUntilTaskDone(self, timeToWait):
        return DAQmxWaitUntilTaskDone(self.task_handle, timeToWait)

    def IsTaskDone(self, isTaskDone):
        return DAQmxIsTaskDone(self.task_handle, isTaskDone)

    def TaskControl(self, action):
        return DAQmxTaskControl(self.task_handle, action)

    def GetNthTaskChannel(self, index, buffer, bufferSize):
        return DAQmxGetNthTaskChannel(self.task_handle, index, buffer, bufferSize)

    def GetNthTaskDevice(self, index, buffer, bufferSize):
        return DAQmxGetNthTaskDevice(self.task_handle, index, buffer, bufferSize)

    def GetTaskAttribute(self, attribute, value):
        return DAQmxGetTaskAttribute(self.task_handle, attribute, value)

    def CreateAIVoltageChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        customScaleName,
    ):
        return DAQmxCreateAIVoltageChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            customScaleName,
        )

    def CreateAICurrentChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        shuntResistorLoc,
        extShuntResistorVal,
        customScaleName,
    ):
        return DAQmxCreateAICurrentChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            shuntResistorLoc,
            extShuntResistorVal,
            customScaleName,
        )

    def CreateAIVoltageRMSChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        customScaleName,
    ):
        return DAQmxCreateAIVoltageRMSChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            customScaleName,
        )

    def CreateAICurrentRMSChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        shuntResistorLoc,
        extShuntResistorVal,
        customScaleName,
    ):
        return DAQmxCreateAICurrentRMSChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            shuntResistorLoc,
            extShuntResistorVal,
            customScaleName,
        )

    def CreateAIThrmcplChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        thermocoupleType,
        cjcSource,
        cjcVal,
        cjcChannel,
    ):
        return DAQmxCreateAIThrmcplChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            thermocoupleType,
            cjcSource,
            cjcVal,
            cjcChannel,
        )

    def CreateAIRTDChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        rtdType,
        resistanceConfig,
        currentExcitSource,
        currentExcitVal,
        r0,
    ):
        return DAQmxCreateAIRTDChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            rtdType,
            resistanceConfig,
            currentExcitSource,
            currentExcitVal,
            r0,
        )

    def CreateAIThrmstrChanIex(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        resistanceConfig,
        currentExcitSource,
        currentExcitVal,
        a,
        b,
        c,
    ):
        return DAQmxCreateAIThrmstrChanIex(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            resistanceConfig,
            currentExcitSource,
            currentExcitVal,
            a,
            b,
            c,
        )

    def CreateAIThrmstrChanVex(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        resistanceConfig,
        voltageExcitSource,
        voltageExcitVal,
        a,
        b,
        c,
        r1,
    ):
        return DAQmxCreateAIThrmstrChanVex(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            resistanceConfig,
            voltageExcitSource,
            voltageExcitVal,
            a,
            b,
            c,
            r1,
        )

    def CreateAIFreqVoltageChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        thresholdLevel,
        hysteresis,
        customScaleName,
    ):
        return DAQmxCreateAIFreqVoltageChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            thresholdLevel,
            hysteresis,
            customScaleName,
        )

    def CreateAIResistanceChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        resistanceConfig,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateAIResistanceChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            resistanceConfig,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateAIStrainGageChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        strainConfig,
        voltageExcitSource,
        voltageExcitVal,
        gageFactor,
        initialBridgeVoltage,
        nominalGageResistance,
        poissonRatio,
        leadWireResistance,
        customScaleName,
    ):
        return DAQmxCreateAIStrainGageChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            strainConfig,
            voltageExcitSource,
            voltageExcitVal,
            gageFactor,
            initialBridgeVoltage,
            nominalGageResistance,
            poissonRatio,
            leadWireResistance,
            customScaleName,
        )

    def CreateAIRosetteStrainGageChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        rosetteType,
        gageOrientation,
        rosetteMeasTypes,
        numRosetteMeasTypes,
        strainConfig,
        voltageExcitSource,
        voltageExcitVal,
        gageFactor,
        nominalGageResistance,
        poissonRatio,
        leadWireResistance,
    ):
        return DAQmxCreateAIRosetteStrainGageChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            rosetteType,
            gageOrientation,
            rosetteMeasTypes,
            numRosetteMeasTypes,
            strainConfig,
            voltageExcitSource,
            voltageExcitVal,
            gageFactor,
            nominalGageResistance,
            poissonRatio,
            leadWireResistance,
        )

    def CreateAIForceBridgeTwoPointLinChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        firstElectricalVal,
        secondElectricalVal,
        electricalUnits,
        firstPhysicalVal,
        secondPhysicalVal,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAIForceBridgeTwoPointLinChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            firstElectricalVal,
            secondElectricalVal,
            electricalUnits,
            firstPhysicalVal,
            secondPhysicalVal,
            physicalUnits,
            customScaleName,
        )

    def CreateAIForceBridgeTableChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        electricalVals,
        numElectricalVals,
        electricalUnits,
        physicalVals,
        numPhysicalVals,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAIForceBridgeTableChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            electricalVals,
            numElectricalVals,
            electricalUnits,
            physicalVals,
            numPhysicalVals,
            physicalUnits,
            customScaleName,
        )

    def CreateAIForceBridgePolynomialChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        forwardCoeffs,
        numForwardCoeffs,
        reverseCoeffs,
        numReverseCoeffs,
        electricalUnits,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAIForceBridgePolynomialChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            forwardCoeffs,
            numForwardCoeffs,
            reverseCoeffs,
            numReverseCoeffs,
            electricalUnits,
            physicalUnits,
            customScaleName,
        )

    def CreateAIPressureBridgeTwoPointLinChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        firstElectricalVal,
        secondElectricalVal,
        electricalUnits,
        firstPhysicalVal,
        secondPhysicalVal,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAIPressureBridgeTwoPointLinChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            firstElectricalVal,
            secondElectricalVal,
            electricalUnits,
            firstPhysicalVal,
            secondPhysicalVal,
            physicalUnits,
            customScaleName,
        )

    def CreateAIPressureBridgeTableChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        electricalVals,
        numElectricalVals,
        electricalUnits,
        physicalVals,
        numPhysicalVals,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAIPressureBridgeTableChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            electricalVals,
            numElectricalVals,
            electricalUnits,
            physicalVals,
            numPhysicalVals,
            physicalUnits,
            customScaleName,
        )

    def CreateAIPressureBridgePolynomialChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        forwardCoeffs,
        numForwardCoeffs,
        reverseCoeffs,
        numReverseCoeffs,
        electricalUnits,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAIPressureBridgePolynomialChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            forwardCoeffs,
            numForwardCoeffs,
            reverseCoeffs,
            numReverseCoeffs,
            electricalUnits,
            physicalUnits,
            customScaleName,
        )

    def CreateAITorqueBridgeTwoPointLinChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        firstElectricalVal,
        secondElectricalVal,
        electricalUnits,
        firstPhysicalVal,
        secondPhysicalVal,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAITorqueBridgeTwoPointLinChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            firstElectricalVal,
            secondElectricalVal,
            electricalUnits,
            firstPhysicalVal,
            secondPhysicalVal,
            physicalUnits,
            customScaleName,
        )

    def CreateAITorqueBridgeTableChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        electricalVals,
        numElectricalVals,
        electricalUnits,
        physicalVals,
        numPhysicalVals,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAITorqueBridgeTableChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            electricalVals,
            numElectricalVals,
            electricalUnits,
            physicalVals,
            numPhysicalVals,
            physicalUnits,
            customScaleName,
        )

    def CreateAITorqueBridgePolynomialChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        forwardCoeffs,
        numForwardCoeffs,
        reverseCoeffs,
        numReverseCoeffs,
        electricalUnits,
        physicalUnits,
        customScaleName,
    ):
        return DAQmxCreateAITorqueBridgePolynomialChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            forwardCoeffs,
            numForwardCoeffs,
            reverseCoeffs,
            numReverseCoeffs,
            electricalUnits,
            physicalUnits,
            customScaleName,
        )

    def CreateAIBridgeChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        nominalBridgeResistance,
        customScaleName,
    ):
        return DAQmxCreateAIBridgeChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            nominalBridgeResistance,
            customScaleName,
        )

    def CreateAIVoltageChanWithExcit(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        bridgeConfig,
        voltageExcitSource,
        voltageExcitVal,
        useExcitForScaling,
        customScaleName,
    ):
        return DAQmxCreateAIVoltageChanWithExcit(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            bridgeConfig,
            voltageExcitSource,
            voltageExcitVal,
            useExcitForScaling,
            customScaleName,
        )

    def CreateAITempBuiltInSensorChan(self, physicalChannel, nameToAssignToChannel, units):
        return DAQmxCreateAITempBuiltInSensorChan(self.task_handle, physicalChannel, nameToAssignToChannel, units)

    def CreateAIAccelChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        sensitivity,
        sensitivityUnits,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateAIAccelChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            sensitivity,
            sensitivityUnits,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateAIVelocityIEPEChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        sensitivity,
        sensitivityUnits,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateAIVelocityIEPEChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            sensitivity,
            sensitivityUnits,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateAIForceIEPEChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        sensitivity,
        sensitivityUnits,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateAIForceIEPEChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            sensitivity,
            sensitivityUnits,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateAIMicrophoneChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        units,
        micSensitivity,
        maxSndPressLevel,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateAIMicrophoneChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            units,
            micSensitivity,
            maxSndPressLevel,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateAIPosLVDTChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        sensitivity,
        sensitivityUnits,
        voltageExcitSource,
        voltageExcitVal,
        voltageExcitFreq,
        ACExcitWireMode,
        customScaleName,
    ):
        return DAQmxCreateAIPosLVDTChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            sensitivity,
            sensitivityUnits,
            voltageExcitSource,
            voltageExcitVal,
            voltageExcitFreq,
            ACExcitWireMode,
            customScaleName,
        )

    def CreateAIPosRVDTChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        sensitivity,
        sensitivityUnits,
        voltageExcitSource,
        voltageExcitVal,
        voltageExcitFreq,
        ACExcitWireMode,
        customScaleName,
    ):
        return DAQmxCreateAIPosRVDTChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            sensitivity,
            sensitivityUnits,
            voltageExcitSource,
            voltageExcitVal,
            voltageExcitFreq,
            ACExcitWireMode,
            customScaleName,
        )

    def CreateAIPosEddyCurrProxProbeChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        sensitivity,
        sensitivityUnits,
        customScaleName,
    ):
        return DAQmxCreateAIPosEddyCurrProxProbeChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            sensitivity,
            sensitivityUnits,
            customScaleName,
        )

    def CreateAIDeviceTempChan(self, physicalChannel, nameToAssignToChannel, units):
        return DAQmxCreateAIDeviceTempChan(self.task_handle, physicalChannel, nameToAssignToChannel, units)

    def CreateTEDSAIVoltageChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIVoltageChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            customScaleName,
        )

    def CreateTEDSAICurrentChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        shuntResistorLoc,
        extShuntResistorVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAICurrentChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            shuntResistorLoc,
            extShuntResistorVal,
            customScaleName,
        )

    def CreateTEDSAIThrmcplChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        cjcSource,
        cjcVal,
        cjcChannel,
    ):
        return DAQmxCreateTEDSAIThrmcplChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            cjcSource,
            cjcVal,
            cjcChannel,
        )

    def CreateTEDSAIRTDChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        resistanceConfig,
        currentExcitSource,
        currentExcitVal,
    ):
        return DAQmxCreateTEDSAIRTDChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            resistanceConfig,
            currentExcitSource,
            currentExcitVal,
        )

    def CreateTEDSAIThrmstrChanIex(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        resistanceConfig,
        currentExcitSource,
        currentExcitVal,
    ):
        return DAQmxCreateTEDSAIThrmstrChanIex(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            resistanceConfig,
            currentExcitSource,
            currentExcitVal,
        )

    def CreateTEDSAIThrmstrChanVex(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        resistanceConfig,
        voltageExcitSource,
        voltageExcitVal,
        r1,
    ):
        return DAQmxCreateTEDSAIThrmstrChanVex(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            resistanceConfig,
            voltageExcitSource,
            voltageExcitVal,
            r1,
        )

    def CreateTEDSAIResistanceChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        resistanceConfig,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIResistanceChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            resistanceConfig,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateTEDSAIStrainGageChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        voltageExcitSource,
        voltageExcitVal,
        initialBridgeVoltage,
        leadWireResistance,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIStrainGageChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            voltageExcitSource,
            voltageExcitVal,
            initialBridgeVoltage,
            leadWireResistance,
            customScaleName,
        )

    def CreateTEDSAIForceBridgeChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        voltageExcitSource,
        voltageExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIForceBridgeChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            voltageExcitSource,
            voltageExcitVal,
            customScaleName,
        )

    def CreateTEDSAIPressureBridgeChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        voltageExcitSource,
        voltageExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIPressureBridgeChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            voltageExcitSource,
            voltageExcitVal,
            customScaleName,
        )

    def CreateTEDSAITorqueBridgeChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        voltageExcitSource,
        voltageExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAITorqueBridgeChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            voltageExcitSource,
            voltageExcitVal,
            customScaleName,
        )

    def CreateTEDSAIBridgeChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        voltageExcitSource,
        voltageExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIBridgeChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            voltageExcitSource,
            voltageExcitVal,
            customScaleName,
        )

    def CreateTEDSAIVoltageChanWithExcit(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        voltageExcitSource,
        voltageExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIVoltageChanWithExcit(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            voltageExcitSource,
            voltageExcitVal,
            customScaleName,
        )

    def CreateTEDSAIAccelChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIAccelChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateTEDSAIForceIEPEChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        minVal,
        maxVal,
        units,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIForceIEPEChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            minVal,
            maxVal,
            units,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateTEDSAIMicrophoneChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        terminalConfig,
        units,
        maxSndPressLevel,
        currentExcitSource,
        currentExcitVal,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIMicrophoneChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            terminalConfig,
            units,
            maxSndPressLevel,
            currentExcitSource,
            currentExcitVal,
            customScaleName,
        )

    def CreateTEDSAIPosLVDTChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        voltageExcitSource,
        voltageExcitVal,
        voltageExcitFreq,
        ACExcitWireMode,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIPosLVDTChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            voltageExcitSource,
            voltageExcitVal,
            voltageExcitFreq,
            ACExcitWireMode,
            customScaleName,
        )

    def CreateTEDSAIPosRVDTChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        voltageExcitSource,
        voltageExcitVal,
        voltageExcitFreq,
        ACExcitWireMode,
        customScaleName,
    ):
        return DAQmxCreateTEDSAIPosRVDTChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            voltageExcitSource,
            voltageExcitVal,
            voltageExcitFreq,
            ACExcitWireMode,
            customScaleName,
        )

    def CreateAOVoltageChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        customScaleName,
    ):
        return DAQmxCreateAOVoltageChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            customScaleName,
        )

    def CreateAOCurrentChan(
        self,
        physicalChannel,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        customScaleName,
    ):
        return DAQmxCreateAOCurrentChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            customScaleName,
        )

    def CreateAOFuncGenChan(self, physicalChannel, nameToAssignToChannel, type, freq, amplitude, offset):
        return DAQmxCreateAOFuncGenChan(
            self.task_handle,
            physicalChannel,
            nameToAssignToChannel,
            type,
            freq,
            amplitude,
            offset,
        )

    def CreateDIChan(self, lines, nameToAssignToLines, lineGrouping):
        return DAQmxCreateDIChan(self.task_handle, lines, nameToAssignToLines, lineGrouping)

    def CreateDOChan(self, lines, nameToAssignToLines, lineGrouping):
        return DAQmxCreateDOChan(self.task_handle, lines, nameToAssignToLines, lineGrouping)

    def CreateCIFreqChan(
        self,
        counter,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        edge,
        measMethod,
        measTime,
        divisor,
        customScaleName,
    ):
        return DAQmxCreateCIFreqChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            edge,
            measMethod,
            measTime,
            divisor,
            customScaleName,
        )

    def CreateCIPeriodChan(
        self,
        counter,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        edge,
        measMethod,
        measTime,
        divisor,
        customScaleName,
    ):
        return DAQmxCreateCIPeriodChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            edge,
            measMethod,
            measTime,
            divisor,
            customScaleName,
        )

    def CreateCICountEdgesChan(self, counter, nameToAssignToChannel, edge, initialCount, countDirection):
        return DAQmxCreateCICountEdgesChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            edge,
            initialCount,
            countDirection,
        )

    def CreateCIPulseWidthChan(
        self,
        counter,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        startingEdge,
        customScaleName,
    ):
        return DAQmxCreateCIPulseWidthChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            startingEdge,
            customScaleName,
        )

    def CreateCISemiPeriodChan(self, counter, nameToAssignToChannel, minVal, maxVal, units, customScaleName):
        return DAQmxCreateCISemiPeriodChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            customScaleName,
        )

    def CreateCITwoEdgeSepChan(
        self,
        counter,
        nameToAssignToChannel,
        minVal,
        maxVal,
        units,
        firstEdge,
        secondEdge,
        customScaleName,
    ):
        return DAQmxCreateCITwoEdgeSepChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            minVal,
            maxVal,
            units,
            firstEdge,
            secondEdge,
            customScaleName,
        )

    def CreateCIPulseChanFreq(self, counter, nameToAssignToChannel, minVal, maxVal, units):
        return DAQmxCreateCIPulseChanFreq(self.task_handle, counter, nameToAssignToChannel, minVal, maxVal, units)

    def CreateCIPulseChanTime(self, counter, nameToAssignToChannel, minVal, maxVal, units):
        return DAQmxCreateCIPulseChanTime(self.task_handle, counter, nameToAssignToChannel, minVal, maxVal, units)

    def CreateCIPulseChanTicks(self, counter, nameToAssignToChannel, sourceTerminal, minVal, maxVal):
        return DAQmxCreateCIPulseChanTicks(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            sourceTerminal,
            minVal,
            maxVal,
        )

    def CreateCILinEncoderChan(
        self,
        counter,
        nameToAssignToChannel,
        decodingType,
        ZidxEnable,
        ZidxVal,
        ZidxPhase,
        units,
        distPerPulse,
        initialPos,
        customScaleName,
    ):
        return DAQmxCreateCILinEncoderChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            decodingType,
            ZidxEnable,
            ZidxVal,
            ZidxPhase,
            units,
            distPerPulse,
            initialPos,
            customScaleName,
        )

    def CreateCIAngEncoderChan(
        self,
        counter,
        nameToAssignToChannel,
        decodingType,
        ZidxEnable,
        ZidxVal,
        ZidxPhase,
        units,
        pulsesPerRev,
        initialAngle,
        customScaleName,
    ):
        return DAQmxCreateCIAngEncoderChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            decodingType,
            ZidxEnable,
            ZidxVal,
            ZidxPhase,
            units,
            pulsesPerRev,
            initialAngle,
            customScaleName,
        )

    def CreateCIGPSTimestampChan(self, counter, nameToAssignToChannel, units, syncMethod, customScaleName):
        return DAQmxCreateCIGPSTimestampChan(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            units,
            syncMethod,
            customScaleName,
        )

    def CreateCOPulseChanFreq(
        self,
        counter,
        nameToAssignToChannel,
        units,
        idleState,
        initialDelay,
        freq,
        dutyCycle,
    ):
        return DAQmxCreateCOPulseChanFreq(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            units,
            idleState,
            initialDelay,
            freq,
            dutyCycle,
        )

    def CreateCOPulseChanTime(
        self,
        counter,
        nameToAssignToChannel,
        units,
        idleState,
        initialDelay,
        lowTime,
        highTime,
    ):
        return DAQmxCreateCOPulseChanTime(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            units,
            idleState,
            initialDelay,
            lowTime,
            highTime,
        )

    def CreateCOPulseChanTicks(
        self,
        counter,
        nameToAssignToChannel,
        sourceTerminal,
        idleState,
        initialDelay,
        lowTicks,
        highTicks,
    ):
        return DAQmxCreateCOPulseChanTicks(
            self.task_handle,
            counter,
            nameToAssignToChannel,
            sourceTerminal,
            idleState,
            initialDelay,
            lowTicks,
            highTicks,
        )

    def GetAIChanCalCalDate(self, channelName, year, month, day, hour, minute):
        return DAQmxGetAIChanCalCalDate(self.task_handle, channelName, year, month, day, hour, minute)

    def SetAIChanCalCalDate(self, channelName, year, month, day, hour, minute):
        return DAQmxSetAIChanCalCalDate(self.task_handle, channelName, year, month, day, hour, minute)

    def GetAIChanCalExpDate(self, channelName, year, month, day, hour, minute):
        return DAQmxGetAIChanCalExpDate(self.task_handle, channelName, year, month, day, hour, minute)

    def SetAIChanCalExpDate(self, channelName, year, month, day, hour, minute):
        return DAQmxSetAIChanCalExpDate(self.task_handle, channelName, year, month, day, hour, minute)

    def GetChanAttribute(self, channel, attribute, value):
        return DAQmxGetChanAttribute(self.task_handle, channel, attribute, value)

    def SetChanAttribute(self, channel, attribute):
        return DAQmxSetChanAttribute(self.task_handle, channel, attribute)

    def ResetChanAttribute(self, channel, attribute):
        return DAQmxResetChanAttribute(self.task_handle, channel, attribute)

    def CfgSampClkTiming(self, source, rate, activeEdge, sampleMode, sampsPerChan):
        return DAQmxCfgSampClkTiming(self.task_handle, source, rate, activeEdge, sampleMode, sampsPerChan)

    def CfgHandshakingTiming(self, sampleMode, sampsPerChan):
        return DAQmxCfgHandshakingTiming(self.task_handle, sampleMode, sampsPerChan)

    def CfgBurstHandshakingTimingImportClock(
        self,
        sampleMode,
        sampsPerChan,
        sampleClkRate,
        sampleClkSrc,
        sampleClkActiveEdge,
        pauseWhen,
        readyEventActiveLevel,
    ):
        return DAQmxCfgBurstHandshakingTimingImportClock(
            self.task_handle,
            sampleMode,
            sampsPerChan,
            sampleClkRate,
            sampleClkSrc,
            sampleClkActiveEdge,
            pauseWhen,
            readyEventActiveLevel,
        )

    def CfgBurstHandshakingTimingExportClock(
        self,
        sampleMode,
        sampsPerChan,
        sampleClkRate,
        sampleClkOutpTerm,
        sampleClkPulsePolarity,
        pauseWhen,
        readyEventActiveLevel,
    ):
        return DAQmxCfgBurstHandshakingTimingExportClock(
            self.task_handle,
            sampleMode,
            sampsPerChan,
            sampleClkRate,
            sampleClkOutpTerm,
            sampleClkPulsePolarity,
            pauseWhen,
            readyEventActiveLevel,
        )

    def CfgChangeDetectionTiming(self, risingEdgeChan, fallingEdgeChan, sampleMode, sampsPerChan):
        return DAQmxCfgChangeDetectionTiming(
            self.task_handle, risingEdgeChan, fallingEdgeChan, sampleMode, sampsPerChan
        )

    def CfgImplicitTiming(self, sampleMode, sampsPerChan):
        return DAQmxCfgImplicitTiming(self.task_handle, sampleMode, sampsPerChan)

    def CfgPipelinedSampClkTiming(self, source, rate, activeEdge, sampleMode, sampsPerChan):
        return DAQmxCfgPipelinedSampClkTiming(self.task_handle, source, rate, activeEdge, sampleMode, sampsPerChan)

    def GetTimingAttribute(self, attribute, value):
        return DAQmxGetTimingAttribute(self.task_handle, attribute, value)

    def SetTimingAttribute(self, attribute):
        return DAQmxSetTimingAttribute(self.task_handle, attribute)

    def ResetTimingAttribute(self, attribute):
        return DAQmxResetTimingAttribute(self.task_handle, attribute)

    def GetTimingAttributeEx(self, deviceNames, attribute, value):
        return DAQmxGetTimingAttributeEx(self.task_handle, deviceNames, attribute, value)

    def SetTimingAttributeEx(self, deviceNames, attribute):
        return DAQmxSetTimingAttributeEx(self.task_handle, deviceNames, attribute)

    def ResetTimingAttributeEx(self, deviceNames, attribute):
        return DAQmxResetTimingAttributeEx(self.task_handle, deviceNames, attribute)

    def DisableStartTrig(self):
        return DAQmxDisableStartTrig(self.task_handle)

    def CfgDigEdgeStartTrig(self, triggerSource, triggerEdge):
        return DAQmxCfgDigEdgeStartTrig(self.task_handle, triggerSource, triggerEdge)

    def CfgAnlgEdgeStartTrig(self, triggerSource, triggerSlope, triggerLevel):
        return DAQmxCfgAnlgEdgeStartTrig(self.task_handle, triggerSource, triggerSlope, triggerLevel)

    def CfgAnlgWindowStartTrig(self, triggerSource, triggerWhen, windowTop, windowBottom):
        return DAQmxCfgAnlgWindowStartTrig(self.task_handle, triggerSource, triggerWhen, windowTop, windowBottom)

    def CfgDigPatternStartTrig(self, triggerSource, triggerPattern, triggerWhen):
        return DAQmxCfgDigPatternStartTrig(self.task_handle, triggerSource, triggerPattern, triggerWhen)

    def DisableRefTrig(self):
        return DAQmxDisableRefTrig(self.task_handle)

    def CfgDigEdgeRefTrig(self, triggerSource, triggerEdge, pretriggerSamples):
        return DAQmxCfgDigEdgeRefTrig(self.task_handle, triggerSource, triggerEdge, pretriggerSamples)

    def CfgAnlgEdgeRefTrig(self, triggerSource, triggerSlope, triggerLevel, pretriggerSamples):
        return DAQmxCfgAnlgEdgeRefTrig(
            self.task_handle,
            triggerSource,
            triggerSlope,
            triggerLevel,
            pretriggerSamples,
        )

    def CfgAnlgWindowRefTrig(self, triggerSource, triggerWhen, windowTop, windowBottom, pretriggerSamples):
        return DAQmxCfgAnlgWindowRefTrig(
            self.task_handle,
            triggerSource,
            triggerWhen,
            windowTop,
            windowBottom,
            pretriggerSamples,
        )

    def CfgDigPatternRefTrig(self, triggerSource, triggerPattern, triggerWhen, pretriggerSamples):
        return DAQmxCfgDigPatternRefTrig(
            self.task_handle,
            triggerSource,
            triggerPattern,
            triggerWhen,
            pretriggerSamples,
        )

    def DisableAdvTrig(self):
        return DAQmxDisableAdvTrig(self.task_handle)

    def CfgDigEdgeAdvTrig(self, triggerSource, triggerEdge):
        return DAQmxCfgDigEdgeAdvTrig(self.task_handle, triggerSource, triggerEdge)

    def GetTrigAttribute(self, attribute, value):
        return DAQmxGetTrigAttribute(self.task_handle, attribute, value)

    def SetTrigAttribute(self, attribute):
        return DAQmxSetTrigAttribute(self.task_handle, attribute)

    def ResetTrigAttribute(self, attribute):
        return DAQmxResetTrigAttribute(self.task_handle, attribute)

    def SendSoftwareTrigger(self, triggerID):
        return DAQmxSendSoftwareTrigger(self.task_handle, triggerID)

    def ReadAnalogF64(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadAnalogF64(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadAnalogScalarF64(self, timeout, value, reserved):
        return DAQmxReadAnalogScalarF64(self.task_handle, timeout, value, reserved)

    def ReadBinaryI16(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadBinaryI16(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadBinaryU16(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadBinaryU16(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadBinaryI32(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadBinaryI32(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadBinaryU32(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadBinaryU32(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadDigitalU8(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadDigitalU8(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadDigitalU16(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadDigitalU16(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadDigitalU32(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadDigitalU32(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadDigitalScalarU32(self, timeout, value, reserved):
        return DAQmxReadDigitalScalarU32(self.task_handle, timeout, value, reserved)

    def ReadDigitalLines(
        self,
        numSampsPerChan,
        timeout,
        fillMode,
        readArray,
        arraySizeInBytes,
        sampsPerChanRead,
        numBytesPerSamp,
        reserved,
    ):
        return DAQmxReadDigitalLines(
            self.task_handle,
            numSampsPerChan,
            timeout,
            fillMode,
            readArray,
            arraySizeInBytes,
            sampsPerChanRead,
            numBytesPerSamp,
            reserved,
        )

    def ReadCounterF64(
        self,
        numSampsPerChan,
        timeout,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadCounterF64(
            self.task_handle,
            numSampsPerChan,
            timeout,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadCounterU32(
        self,
        numSampsPerChan,
        timeout,
        readArray,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadCounterU32(
            self.task_handle,
            numSampsPerChan,
            timeout,
            readArray,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadCounterScalarF64(self, timeout, value, reserved):
        return DAQmxReadCounterScalarF64(self.task_handle, timeout, value, reserved)

    def ReadCounterScalarU32(self, timeout, value, reserved):
        return DAQmxReadCounterScalarU32(self.task_handle, timeout, value, reserved)

    def ReadCtrFreq(
        self,
        numSampsPerChan,
        timeout,
        interleaved,
        readArrayFrequency,
        readArrayDutyCycle,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadCtrFreq(
            self.task_handle,
            numSampsPerChan,
            timeout,
            interleaved,
            readArrayFrequency,
            readArrayDutyCycle,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadCtrTime(
        self,
        numSampsPerChan,
        timeout,
        interleaved,
        readArrayHighTime,
        readArrayLowTime,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadCtrTime(
            self.task_handle,
            numSampsPerChan,
            timeout,
            interleaved,
            readArrayHighTime,
            readArrayLowTime,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadCtrTicks(
        self,
        numSampsPerChan,
        timeout,
        interleaved,
        readArrayHighTicks,
        readArrayLowTicks,
        arraySizeInSamps,
        sampsPerChanRead,
        reserved,
    ):
        return DAQmxReadCtrTicks(
            self.task_handle,
            numSampsPerChan,
            timeout,
            interleaved,
            readArrayHighTicks,
            readArrayLowTicks,
            arraySizeInSamps,
            sampsPerChanRead,
            reserved,
        )

    def ReadCtrFreqScalar(self, timeout, frequency, dutyCycle, reserved):
        return DAQmxReadCtrFreqScalar(self.task_handle, timeout, frequency, dutyCycle, reserved)

    def ReadCtrTimeScalar(self, timeout, highTime, lowTime, reserved):
        return DAQmxReadCtrTimeScalar(self.task_handle, timeout, highTime, lowTime, reserved)

    def ReadCtrTicksScalar(self, timeout, highTicks, lowTicks, reserved):
        return DAQmxReadCtrTicksScalar(self.task_handle, timeout, highTicks, lowTicks, reserved)

    def ReadRaw(
        self,
        numSampsPerChan,
        timeout,
        readArray,
        arraySizeInBytes,
        sampsRead,
        numBytesPerSamp,
        reserved,
    ):
        return DAQmxReadRaw(
            self.task_handle,
            numSampsPerChan,
            timeout,
            readArray,
            arraySizeInBytes,
            sampsRead,
            numBytesPerSamp,
            reserved,
        )

    def GetNthTaskReadChannel(self, index, buffer, bufferSize):
        return DAQmxGetNthTaskReadChannel(self.task_handle, index, buffer, bufferSize)

    def GetReadAttribute(self, attribute, value):
        return DAQmxGetReadAttribute(self.task_handle, attribute, value)

    def SetReadAttribute(self, attribute):
        return DAQmxSetReadAttribute(self.task_handle, attribute)

    def ResetReadAttribute(self, attribute):
        return DAQmxResetReadAttribute(self.task_handle, attribute)

    def ConfigureLogging(self, filePath, loggingMode, groupName, operation):
        return DAQmxConfigureLogging(self.task_handle, filePath, loggingMode, groupName, operation)

    def StartNewFile(self, filePath):
        return DAQmxStartNewFile(self.task_handle, filePath)

    def WriteAnalogF64(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteAnalogF64(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteAnalogScalarF64(self, autoStart, timeout, value, reserved):
        return DAQmxWriteAnalogScalarF64(self.task_handle, autoStart, timeout, value, reserved)

    def WriteBinaryI16(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteBinaryI16(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteBinaryU16(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteBinaryU16(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteBinaryI32(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteBinaryI32(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteBinaryU32(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteBinaryU32(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteDigitalU8(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteDigitalU8(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteDigitalU16(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteDigitalU16(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteDigitalU32(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteDigitalU32(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteDigitalScalarU32(self, autoStart, timeout, value, reserved):
        return DAQmxWriteDigitalScalarU32(self.task_handle, autoStart, timeout, value, reserved)

    def WriteDigitalLines(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        writeArray,
        sampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteDigitalLines(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def WriteCtrFreq(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        frequency,
        dutyCycle,
        numSampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteCtrFreq(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            frequency,
            dutyCycle,
            numSampsPerChanWritten,
            reserved,
        )

    def WriteCtrFreqScalar(self, autoStart, timeout, frequency, dutyCycle, reserved):
        return DAQmxWriteCtrFreqScalar(self.task_handle, autoStart, timeout, frequency, dutyCycle, reserved)

    def WriteCtrTime(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        highTime,
        lowTime,
        numSampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteCtrTime(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            highTime,
            lowTime,
            numSampsPerChanWritten,
            reserved,
        )

    def WriteCtrTimeScalar(self, autoStart, timeout, highTime, lowTime, reserved):
        return DAQmxWriteCtrTimeScalar(self.task_handle, autoStart, timeout, highTime, lowTime, reserved)

    def WriteCtrTicks(
        self,
        numSampsPerChan,
        autoStart,
        timeout,
        dataLayout,
        highTicks,
        lowTicks,
        numSampsPerChanWritten,
        reserved,
    ):
        return DAQmxWriteCtrTicks(
            self.task_handle,
            numSampsPerChan,
            autoStart,
            timeout,
            dataLayout,
            highTicks,
            lowTicks,
            numSampsPerChanWritten,
            reserved,
        )

    def WriteCtrTicksScalar(self, autoStart, timeout, highTicks, lowTicks, reserved):
        return DAQmxWriteCtrTicksScalar(self.task_handle, autoStart, timeout, highTicks, lowTicks, reserved)

    def WriteRaw(self, numSamps, autoStart, timeout, writeArray, sampsPerChanWritten, reserved):
        return DAQmxWriteRaw(
            self.task_handle,
            numSamps,
            autoStart,
            timeout,
            writeArray,
            sampsPerChanWritten,
            reserved,
        )

    def GetWriteAttribute(self, attribute, value):
        return DAQmxGetWriteAttribute(self.task_handle, attribute, value)

    def SetWriteAttribute(self, attribute):
        return DAQmxSetWriteAttribute(self.task_handle, attribute)

    def ResetWriteAttribute(self, attribute):
        return DAQmxResetWriteAttribute(self.task_handle, attribute)

    def ExportSignal(self, signalID, outputTerminal):
        return DAQmxExportSignal(self.task_handle, signalID, outputTerminal)

    def GetExportedSignalAttribute(self, attribute, value):
        return DAQmxGetExportedSignalAttribute(self.task_handle, attribute, value)

    def SetExportedSignalAttribute(self, attribute):
        return DAQmxSetExportedSignalAttribute(self.task_handle, attribute)

    def ResetExportedSignalAttribute(self, attribute):
        return DAQmxResetExportedSignalAttribute(self.task_handle, attribute)

    def CfgInputBuffer(self, numSampsPerChan):
        return DAQmxCfgInputBuffer(self.task_handle, numSampsPerChan)

    def CfgOutputBuffer(self, numSampsPerChan):
        return DAQmxCfgOutputBuffer(self.task_handle, numSampsPerChan)

    def GetBufferAttribute(self, attribute, value):
        return DAQmxGetBufferAttribute(self.task_handle, attribute, value)

    def SetBufferAttribute(self, attribute):
        return DAQmxSetBufferAttribute(self.task_handle, attribute)

    def ResetBufferAttribute(self, attribute):
        return DAQmxResetBufferAttribute(self.task_handle, attribute)

    def GetSwitchScanAttribute(self, attribute, value):
        return DAQmxGetSwitchScanAttribute(self.task_handle, attribute, value)

    def SetSwitchScanAttribute(self, attribute):
        return DAQmxSetSwitchScanAttribute(self.task_handle, attribute)

    def ResetSwitchScanAttribute(self, attribute):
        return DAQmxResetSwitchScanAttribute(self.task_handle, attribute)

    def ControlWatchdogTask(self, action):
        return DAQmxControlWatchdogTask(self.task_handle, action)

    def CfgWatchdogAOExpirStates(self, channelNames, expirStateArray, outputTypeArray, arraySize):
        return DAQmxCfgWatchdogAOExpirStates(
            self.task_handle, channelNames, expirStateArray, outputTypeArray, arraySize
        )

    def CfgWatchdogCOExpirStates(self, channelNames, expirStateArray, arraySize):
        return DAQmxCfgWatchdogCOExpirStates(self.task_handle, channelNames, expirStateArray, arraySize)

    def CfgWatchdogDOExpirStates(self, channelNames, expirStateArray, arraySize):
        return DAQmxCfgWatchdogDOExpirStates(self.task_handle, channelNames, expirStateArray, arraySize)

    def GetWatchdogAttribute(self, lines, attribute, value):
        return DAQmxGetWatchdogAttribute(self.task_handle, lines, attribute, value)

    def SetWatchdogAttribute(self, lines, attribute):
        return DAQmxSetWatchdogAttribute(self.task_handle, lines, attribute)

    def ResetWatchdogAttribute(self, lines, attribute):
        return DAQmxResetWatchdogAttribute(self.task_handle, lines, attribute)

    def PerformBridgeOffsetNullingCal(self, channel):
        return DAQmxPerformBridgeOffsetNullingCal(self.task_handle, channel)

    def PerformBridgeOffsetNullingCalEx(self, channel, skipUnsupportedChannels):
        return DAQmxPerformBridgeOffsetNullingCalEx(self.task_handle, channel, skipUnsupportedChannels)

    def PerformThrmcplLeadOffsetNullingCal(self, channel, skipUnsupportedChannels):
        return DAQmxPerformThrmcplLeadOffsetNullingCal(self.task_handle, channel, skipUnsupportedChannels)

    def PerformStrainShuntCal(
        self,
        channel,
        shuntResistorValue,
        shuntResistorLocation,
        skipUnsupportedChannels,
    ):
        return DAQmxPerformStrainShuntCal(
            self.task_handle,
            channel,
            shuntResistorValue,
            shuntResistorLocation,
            skipUnsupportedChannels,
        )

    def PerformBridgeShuntCal(
        self,
        channel,
        shuntResistorValue,
        shuntResistorLocation,
        bridgeResistance,
        skipUnsupportedChannels,
    ):
        return DAQmxPerformBridgeShuntCal(
            self.task_handle,
            channel,
            shuntResistorValue,
            shuntResistorLocation,
            bridgeResistance,
            skipUnsupportedChannels,
        )

    def WaitForNextSampleClock(self, timeout, isLate):
        return DAQmxWaitForNextSampleClock(self.task_handle, timeout, isLate)

    def GetRealTimeAttribute(self, attribute, value):
        return DAQmxGetRealTimeAttribute(self.task_handle, attribute, value)

    def SetRealTimeAttribute(self, attribute):
        return DAQmxSetRealTimeAttribute(self.task_handle, attribute)

    def ResetRealTimeAttribute(self, attribute):
        return DAQmxResetRealTimeAttribute(self.task_handle, attribute)

    def SaveTask(self, saveAs, author, options):
        return DAQmxSaveTask(self.task_handle, saveAs, author, options)

    def SaveGlobalChan(self, channelName, saveAs, author, options):
        return DAQmxSaveGlobalChan(self.task_handle, channelName, saveAs, author, options)

    def GetBufInputBufSize(self, data):
        return DAQmxGetBufInputBufSize(self.task_handle, data)

    def SetBufInputBufSize(self, data):
        return DAQmxSetBufInputBufSize(self.task_handle, data)

    def ResetBufInputBufSize(self):
        return DAQmxResetBufInputBufSize(self.task_handle)

    def GetBufInputOnbrdBufSize(self, data):
        return DAQmxGetBufInputOnbrdBufSize(self.task_handle, data)

    def GetBufOutputBufSize(self, data):
        return DAQmxGetBufOutputBufSize(self.task_handle, data)

    def SetBufOutputBufSize(self, data):
        return DAQmxSetBufOutputBufSize(self.task_handle, data)

    def ResetBufOutputBufSize(self):
        return DAQmxResetBufOutputBufSize(self.task_handle)

    def GetBufOutputOnbrdBufSize(self, data):
        return DAQmxGetBufOutputOnbrdBufSize(self.task_handle, data)

    def SetBufOutputOnbrdBufSize(self, data):
        return DAQmxSetBufOutputOnbrdBufSize(self.task_handle, data)

    def ResetBufOutputOnbrdBufSize(self):
        return DAQmxResetBufOutputOnbrdBufSize(self.task_handle)

    def GetAIMax(self, channel, data):
        return DAQmxGetAIMax(self.task_handle, channel, data)

    def SetAIMax(self, channel, data):
        return DAQmxSetAIMax(self.task_handle, channel, data)

    def ResetAIMax(self, channel):
        return DAQmxResetAIMax(self.task_handle, channel)

    def GetAIMin(self, channel, data):
        return DAQmxGetAIMin(self.task_handle, channel, data)

    def SetAIMin(self, channel, data):
        return DAQmxSetAIMin(self.task_handle, channel, data)

    def ResetAIMin(self, channel):
        return DAQmxResetAIMin(self.task_handle, channel)

    def GetAICustomScaleName(self, channel, data, bufferSize):
        return DAQmxGetAICustomScaleName(self.task_handle, channel, data, bufferSize)

    def SetAICustomScaleName(self, channel, data):
        return DAQmxSetAICustomScaleName(self.task_handle, channel, data)

    def ResetAICustomScaleName(self, channel):
        return DAQmxResetAICustomScaleName(self.task_handle, channel)

    def GetAIMeasType(self, channel, data):
        return DAQmxGetAIMeasType(self.task_handle, channel, data)

    def GetAIVoltageUnits(self, channel, data):
        return DAQmxGetAIVoltageUnits(self.task_handle, channel, data)

    def SetAIVoltageUnits(self, channel, data):
        return DAQmxSetAIVoltageUnits(self.task_handle, channel, data)

    def ResetAIVoltageUnits(self, channel):
        return DAQmxResetAIVoltageUnits(self.task_handle, channel)

    def GetAIVoltagedBRef(self, channel, data):
        return DAQmxGetAIVoltagedBRef(self.task_handle, channel, data)

    def SetAIVoltagedBRef(self, channel, data):
        return DAQmxSetAIVoltagedBRef(self.task_handle, channel, data)

    def ResetAIVoltagedBRef(self, channel):
        return DAQmxResetAIVoltagedBRef(self.task_handle, channel)

    def GetAIVoltageACRMSUnits(self, channel, data):
        return DAQmxGetAIVoltageACRMSUnits(self.task_handle, channel, data)

    def SetAIVoltageACRMSUnits(self, channel, data):
        return DAQmxSetAIVoltageACRMSUnits(self.task_handle, channel, data)

    def ResetAIVoltageACRMSUnits(self, channel):
        return DAQmxResetAIVoltageACRMSUnits(self.task_handle, channel)

    def GetAITempUnits(self, channel, data):
        return DAQmxGetAITempUnits(self.task_handle, channel, data)

    def SetAITempUnits(self, channel, data):
        return DAQmxSetAITempUnits(self.task_handle, channel, data)

    def ResetAITempUnits(self, channel):
        return DAQmxResetAITempUnits(self.task_handle, channel)

    def GetAIThrmcplType(self, channel, data):
        return DAQmxGetAIThrmcplType(self.task_handle, channel, data)

    def SetAIThrmcplType(self, channel, data):
        return DAQmxSetAIThrmcplType(self.task_handle, channel, data)

    def ResetAIThrmcplType(self, channel):
        return DAQmxResetAIThrmcplType(self.task_handle, channel)

    def GetAIThrmcplScaleType(self, channel, data):
        return DAQmxGetAIThrmcplScaleType(self.task_handle, channel, data)

    def SetAIThrmcplScaleType(self, channel, data):
        return DAQmxSetAIThrmcplScaleType(self.task_handle, channel, data)

    def ResetAIThrmcplScaleType(self, channel):
        return DAQmxResetAIThrmcplScaleType(self.task_handle, channel)

    def GetAIThrmcplCJCSrc(self, channel, data):
        return DAQmxGetAIThrmcplCJCSrc(self.task_handle, channel, data)

    def GetAIThrmcplCJCVal(self, channel, data):
        return DAQmxGetAIThrmcplCJCVal(self.task_handle, channel, data)

    def SetAIThrmcplCJCVal(self, channel, data):
        return DAQmxSetAIThrmcplCJCVal(self.task_handle, channel, data)

    def ResetAIThrmcplCJCVal(self, channel):
        return DAQmxResetAIThrmcplCJCVal(self.task_handle, channel)

    def GetAIThrmcplCJCChan(self, channel, data, bufferSize):
        return DAQmxGetAIThrmcplCJCChan(self.task_handle, channel, data, bufferSize)

    def GetAIRTDType(self, channel, data):
        return DAQmxGetAIRTDType(self.task_handle, channel, data)

    def SetAIRTDType(self, channel, data):
        return DAQmxSetAIRTDType(self.task_handle, channel, data)

    def ResetAIRTDType(self, channel):
        return DAQmxResetAIRTDType(self.task_handle, channel)

    def GetAIRTDR0(self, channel, data):
        return DAQmxGetAIRTDR0(self.task_handle, channel, data)

    def SetAIRTDR0(self, channel, data):
        return DAQmxSetAIRTDR0(self.task_handle, channel, data)

    def ResetAIRTDR0(self, channel):
        return DAQmxResetAIRTDR0(self.task_handle, channel)

    def GetAIRTDA(self, channel, data):
        return DAQmxGetAIRTDA(self.task_handle, channel, data)

    def SetAIRTDA(self, channel, data):
        return DAQmxSetAIRTDA(self.task_handle, channel, data)

    def ResetAIRTDA(self, channel):
        return DAQmxResetAIRTDA(self.task_handle, channel)

    def GetAIRTDB(self, channel, data):
        return DAQmxGetAIRTDB(self.task_handle, channel, data)

    def SetAIRTDB(self, channel, data):
        return DAQmxSetAIRTDB(self.task_handle, channel, data)

    def ResetAIRTDB(self, channel):
        return DAQmxResetAIRTDB(self.task_handle, channel)

    def GetAIRTDC(self, channel, data):
        return DAQmxGetAIRTDC(self.task_handle, channel, data)

    def SetAIRTDC(self, channel, data):
        return DAQmxSetAIRTDC(self.task_handle, channel, data)

    def ResetAIRTDC(self, channel):
        return DAQmxResetAIRTDC(self.task_handle, channel)

    def GetAIThrmstrA(self, channel, data):
        return DAQmxGetAIThrmstrA(self.task_handle, channel, data)

    def SetAIThrmstrA(self, channel, data):
        return DAQmxSetAIThrmstrA(self.task_handle, channel, data)

    def ResetAIThrmstrA(self, channel):
        return DAQmxResetAIThrmstrA(self.task_handle, channel)

    def GetAIThrmstrB(self, channel, data):
        return DAQmxGetAIThrmstrB(self.task_handle, channel, data)

    def SetAIThrmstrB(self, channel, data):
        return DAQmxSetAIThrmstrB(self.task_handle, channel, data)

    def ResetAIThrmstrB(self, channel):
        return DAQmxResetAIThrmstrB(self.task_handle, channel)

    def GetAIThrmstrC(self, channel, data):
        return DAQmxGetAIThrmstrC(self.task_handle, channel, data)

    def SetAIThrmstrC(self, channel, data):
        return DAQmxSetAIThrmstrC(self.task_handle, channel, data)

    def ResetAIThrmstrC(self, channel):
        return DAQmxResetAIThrmstrC(self.task_handle, channel)

    def GetAIThrmstrR1(self, channel, data):
        return DAQmxGetAIThrmstrR1(self.task_handle, channel, data)

    def SetAIThrmstrR1(self, channel, data):
        return DAQmxSetAIThrmstrR1(self.task_handle, channel, data)

    def ResetAIThrmstrR1(self, channel):
        return DAQmxResetAIThrmstrR1(self.task_handle, channel)

    def GetAIForceReadFromChan(self, channel, data):
        return DAQmxGetAIForceReadFromChan(self.task_handle, channel, data)

    def SetAIForceReadFromChan(self, channel, data):
        return DAQmxSetAIForceReadFromChan(self.task_handle, channel, data)

    def ResetAIForceReadFromChan(self, channel):
        return DAQmxResetAIForceReadFromChan(self.task_handle, channel)

    def GetAICurrentUnits(self, channel, data):
        return DAQmxGetAICurrentUnits(self.task_handle, channel, data)

    def SetAICurrentUnits(self, channel, data):
        return DAQmxSetAICurrentUnits(self.task_handle, channel, data)

    def ResetAICurrentUnits(self, channel):
        return DAQmxResetAICurrentUnits(self.task_handle, channel)

    def GetAICurrentACRMSUnits(self, channel, data):
        return DAQmxGetAICurrentACRMSUnits(self.task_handle, channel, data)

    def SetAICurrentACRMSUnits(self, channel, data):
        return DAQmxSetAICurrentACRMSUnits(self.task_handle, channel, data)

    def ResetAICurrentACRMSUnits(self, channel):
        return DAQmxResetAICurrentACRMSUnits(self.task_handle, channel)

    def GetAIStrainUnits(self, channel, data):
        return DAQmxGetAIStrainUnits(self.task_handle, channel, data)

    def SetAIStrainUnits(self, channel, data):
        return DAQmxSetAIStrainUnits(self.task_handle, channel, data)

    def ResetAIStrainUnits(self, channel):
        return DAQmxResetAIStrainUnits(self.task_handle, channel)

    def GetAIStrainGageForceReadFromChan(self, channel, data):
        return DAQmxGetAIStrainGageForceReadFromChan(self.task_handle, channel, data)

    def SetAIStrainGageForceReadFromChan(self, channel, data):
        return DAQmxSetAIStrainGageForceReadFromChan(self.task_handle, channel, data)

    def ResetAIStrainGageForceReadFromChan(self, channel):
        return DAQmxResetAIStrainGageForceReadFromChan(self.task_handle, channel)

    def GetAIStrainGageGageFactor(self, channel, data):
        return DAQmxGetAIStrainGageGageFactor(self.task_handle, channel, data)

    def SetAIStrainGageGageFactor(self, channel, data):
        return DAQmxSetAIStrainGageGageFactor(self.task_handle, channel, data)

    def ResetAIStrainGageGageFactor(self, channel):
        return DAQmxResetAIStrainGageGageFactor(self.task_handle, channel)

    def GetAIStrainGagePoissonRatio(self, channel, data):
        return DAQmxGetAIStrainGagePoissonRatio(self.task_handle, channel, data)

    def SetAIStrainGagePoissonRatio(self, channel, data):
        return DAQmxSetAIStrainGagePoissonRatio(self.task_handle, channel, data)

    def ResetAIStrainGagePoissonRatio(self, channel):
        return DAQmxResetAIStrainGagePoissonRatio(self.task_handle, channel)

    def GetAIStrainGageCfg(self, channel, data):
        return DAQmxGetAIStrainGageCfg(self.task_handle, channel, data)

    def SetAIStrainGageCfg(self, channel, data):
        return DAQmxSetAIStrainGageCfg(self.task_handle, channel, data)

    def ResetAIStrainGageCfg(self, channel):
        return DAQmxResetAIStrainGageCfg(self.task_handle, channel)

    def GetAIRosetteStrainGageRosetteType(self, channel, data):
        return DAQmxGetAIRosetteStrainGageRosetteType(self.task_handle, channel, data)

    def GetAIRosetteStrainGageOrientation(self, channel, data):
        return DAQmxGetAIRosetteStrainGageOrientation(self.task_handle, channel, data)

    def SetAIRosetteStrainGageOrientation(self, channel, data):
        return DAQmxSetAIRosetteStrainGageOrientation(self.task_handle, channel, data)

    def ResetAIRosetteStrainGageOrientation(self, channel):
        return DAQmxResetAIRosetteStrainGageOrientation(self.task_handle, channel)

    def GetAIRosetteStrainGageStrainChans(self, channel, data, bufferSize):
        return DAQmxGetAIRosetteStrainGageStrainChans(self.task_handle, channel, data, bufferSize)

    def GetAIRosetteStrainGageRosetteMeasType(self, channel, data):
        return DAQmxGetAIRosetteStrainGageRosetteMeasType(self.task_handle, channel, data)

    def SetAIRosetteStrainGageRosetteMeasType(self, channel, data):
        return DAQmxSetAIRosetteStrainGageRosetteMeasType(self.task_handle, channel, data)

    def ResetAIRosetteStrainGageRosetteMeasType(self, channel):
        return DAQmxResetAIRosetteStrainGageRosetteMeasType(self.task_handle, channel)

    def GetAIResistanceUnits(self, channel, data):
        return DAQmxGetAIResistanceUnits(self.task_handle, channel, data)

    def SetAIResistanceUnits(self, channel, data):
        return DAQmxSetAIResistanceUnits(self.task_handle, channel, data)

    def ResetAIResistanceUnits(self, channel):
        return DAQmxResetAIResistanceUnits(self.task_handle, channel)

    def GetAIFreqUnits(self, channel, data):
        return DAQmxGetAIFreqUnits(self.task_handle, channel, data)

    def SetAIFreqUnits(self, channel, data):
        return DAQmxSetAIFreqUnits(self.task_handle, channel, data)

    def ResetAIFreqUnits(self, channel):
        return DAQmxResetAIFreqUnits(self.task_handle, channel)

    def GetAIFreqThreshVoltage(self, channel, data):
        return DAQmxGetAIFreqThreshVoltage(self.task_handle, channel, data)

    def SetAIFreqThreshVoltage(self, channel, data):
        return DAQmxSetAIFreqThreshVoltage(self.task_handle, channel, data)

    def ResetAIFreqThreshVoltage(self, channel):
        return DAQmxResetAIFreqThreshVoltage(self.task_handle, channel)

    def GetAIFreqHyst(self, channel, data):
        return DAQmxGetAIFreqHyst(self.task_handle, channel, data)

    def SetAIFreqHyst(self, channel, data):
        return DAQmxSetAIFreqHyst(self.task_handle, channel, data)

    def ResetAIFreqHyst(self, channel):
        return DAQmxResetAIFreqHyst(self.task_handle, channel)

    def GetAILVDTUnits(self, channel, data):
        return DAQmxGetAILVDTUnits(self.task_handle, channel, data)

    def SetAILVDTUnits(self, channel, data):
        return DAQmxSetAILVDTUnits(self.task_handle, channel, data)

    def ResetAILVDTUnits(self, channel):
        return DAQmxResetAILVDTUnits(self.task_handle, channel)

    def GetAILVDTSensitivity(self, channel, data):
        return DAQmxGetAILVDTSensitivity(self.task_handle, channel, data)

    def SetAILVDTSensitivity(self, channel, data):
        return DAQmxSetAILVDTSensitivity(self.task_handle, channel, data)

    def ResetAILVDTSensitivity(self, channel):
        return DAQmxResetAILVDTSensitivity(self.task_handle, channel)

    def GetAILVDTSensitivityUnits(self, channel, data):
        return DAQmxGetAILVDTSensitivityUnits(self.task_handle, channel, data)

    def SetAILVDTSensitivityUnits(self, channel, data):
        return DAQmxSetAILVDTSensitivityUnits(self.task_handle, channel, data)

    def ResetAILVDTSensitivityUnits(self, channel):
        return DAQmxResetAILVDTSensitivityUnits(self.task_handle, channel)

    def GetAIRVDTUnits(self, channel, data):
        return DAQmxGetAIRVDTUnits(self.task_handle, channel, data)

    def SetAIRVDTUnits(self, channel, data):
        return DAQmxSetAIRVDTUnits(self.task_handle, channel, data)

    def ResetAIRVDTUnits(self, channel):
        return DAQmxResetAIRVDTUnits(self.task_handle, channel)

    def GetAIRVDTSensitivity(self, channel, data):
        return DAQmxGetAIRVDTSensitivity(self.task_handle, channel, data)

    def SetAIRVDTSensitivity(self, channel, data):
        return DAQmxSetAIRVDTSensitivity(self.task_handle, channel, data)

    def ResetAIRVDTSensitivity(self, channel):
        return DAQmxResetAIRVDTSensitivity(self.task_handle, channel)

    def GetAIRVDTSensitivityUnits(self, channel, data):
        return DAQmxGetAIRVDTSensitivityUnits(self.task_handle, channel, data)

    def SetAIRVDTSensitivityUnits(self, channel, data):
        return DAQmxSetAIRVDTSensitivityUnits(self.task_handle, channel, data)

    def ResetAIRVDTSensitivityUnits(self, channel):
        return DAQmxResetAIRVDTSensitivityUnits(self.task_handle, channel)

    def GetAIEddyCurrentProxProbeUnits(self, channel, data):
        return DAQmxGetAIEddyCurrentProxProbeUnits(self.task_handle, channel, data)

    def SetAIEddyCurrentProxProbeUnits(self, channel, data):
        return DAQmxSetAIEddyCurrentProxProbeUnits(self.task_handle, channel, data)

    def ResetAIEddyCurrentProxProbeUnits(self, channel):
        return DAQmxResetAIEddyCurrentProxProbeUnits(self.task_handle, channel)

    def GetAIEddyCurrentProxProbeSensitivity(self, channel, data):
        return DAQmxGetAIEddyCurrentProxProbeSensitivity(self.task_handle, channel, data)

    def SetAIEddyCurrentProxProbeSensitivity(self, channel, data):
        return DAQmxSetAIEddyCurrentProxProbeSensitivity(self.task_handle, channel, data)

    def ResetAIEddyCurrentProxProbeSensitivity(self, channel):
        return DAQmxResetAIEddyCurrentProxProbeSensitivity(self.task_handle, channel)

    def GetAIEddyCurrentProxProbeSensitivityUnits(self, channel, data):
        return DAQmxGetAIEddyCurrentProxProbeSensitivityUnits(self.task_handle, channel, data)

    def SetAIEddyCurrentProxProbeSensitivityUnits(self, channel, data):
        return DAQmxSetAIEddyCurrentProxProbeSensitivityUnits(self.task_handle, channel, data)

    def ResetAIEddyCurrentProxProbeSensitivityUnits(self, channel):
        return DAQmxResetAIEddyCurrentProxProbeSensitivityUnits(self.task_handle, channel)

    def GetAISoundPressureMaxSoundPressureLvl(self, channel, data):
        return DAQmxGetAISoundPressureMaxSoundPressureLvl(self.task_handle, channel, data)

    def SetAISoundPressureMaxSoundPressureLvl(self, channel, data):
        return DAQmxSetAISoundPressureMaxSoundPressureLvl(self.task_handle, channel, data)

    def ResetAISoundPressureMaxSoundPressureLvl(self, channel):
        return DAQmxResetAISoundPressureMaxSoundPressureLvl(self.task_handle, channel)

    def GetAISoundPressureUnits(self, channel, data):
        return DAQmxGetAISoundPressureUnits(self.task_handle, channel, data)

    def SetAISoundPressureUnits(self, channel, data):
        return DAQmxSetAISoundPressureUnits(self.task_handle, channel, data)

    def ResetAISoundPressureUnits(self, channel):
        return DAQmxResetAISoundPressureUnits(self.task_handle, channel)

    def GetAISoundPressuredBRef(self, channel, data):
        return DAQmxGetAISoundPressuredBRef(self.task_handle, channel, data)

    def SetAISoundPressuredBRef(self, channel, data):
        return DAQmxSetAISoundPressuredBRef(self.task_handle, channel, data)

    def ResetAISoundPressuredBRef(self, channel):
        return DAQmxResetAISoundPressuredBRef(self.task_handle, channel)

    def GetAIMicrophoneSensitivity(self, channel, data):
        return DAQmxGetAIMicrophoneSensitivity(self.task_handle, channel, data)

    def SetAIMicrophoneSensitivity(self, channel, data):
        return DAQmxSetAIMicrophoneSensitivity(self.task_handle, channel, data)

    def ResetAIMicrophoneSensitivity(self, channel):
        return DAQmxResetAIMicrophoneSensitivity(self.task_handle, channel)

    def GetAIAccelUnits(self, channel, data):
        return DAQmxGetAIAccelUnits(self.task_handle, channel, data)

    def SetAIAccelUnits(self, channel, data):
        return DAQmxSetAIAccelUnits(self.task_handle, channel, data)

    def ResetAIAccelUnits(self, channel):
        return DAQmxResetAIAccelUnits(self.task_handle, channel)

    def GetAIAcceldBRef(self, channel, data):
        return DAQmxGetAIAcceldBRef(self.task_handle, channel, data)

    def SetAIAcceldBRef(self, channel, data):
        return DAQmxSetAIAcceldBRef(self.task_handle, channel, data)

    def ResetAIAcceldBRef(self, channel):
        return DAQmxResetAIAcceldBRef(self.task_handle, channel)

    def GetAIAccelSensitivity(self, channel, data):
        return DAQmxGetAIAccelSensitivity(self.task_handle, channel, data)

    def SetAIAccelSensitivity(self, channel, data):
        return DAQmxSetAIAccelSensitivity(self.task_handle, channel, data)

    def ResetAIAccelSensitivity(self, channel):
        return DAQmxResetAIAccelSensitivity(self.task_handle, channel)

    def GetAIAccelSensitivityUnits(self, channel, data):
        return DAQmxGetAIAccelSensitivityUnits(self.task_handle, channel, data)

    def SetAIAccelSensitivityUnits(self, channel, data):
        return DAQmxSetAIAccelSensitivityUnits(self.task_handle, channel, data)

    def ResetAIAccelSensitivityUnits(self, channel):
        return DAQmxResetAIAccelSensitivityUnits(self.task_handle, channel)

    def GetAIVelocityUnits(self, channel, data):
        return DAQmxGetAIVelocityUnits(self.task_handle, channel, data)

    def SetAIVelocityUnits(self, channel, data):
        return DAQmxSetAIVelocityUnits(self.task_handle, channel, data)

    def ResetAIVelocityUnits(self, channel):
        return DAQmxResetAIVelocityUnits(self.task_handle, channel)

    def GetAIVelocityIEPESensordBRef(self, channel, data):
        return DAQmxGetAIVelocityIEPESensordBRef(self.task_handle, channel, data)

    def SetAIVelocityIEPESensordBRef(self, channel, data):
        return DAQmxSetAIVelocityIEPESensordBRef(self.task_handle, channel, data)

    def ResetAIVelocityIEPESensordBRef(self, channel):
        return DAQmxResetAIVelocityIEPESensordBRef(self.task_handle, channel)

    def GetAIVelocityIEPESensorSensitivity(self, channel, data):
        return DAQmxGetAIVelocityIEPESensorSensitivity(self.task_handle, channel, data)

    def SetAIVelocityIEPESensorSensitivity(self, channel, data):
        return DAQmxSetAIVelocityIEPESensorSensitivity(self.task_handle, channel, data)

    def ResetAIVelocityIEPESensorSensitivity(self, channel):
        return DAQmxResetAIVelocityIEPESensorSensitivity(self.task_handle, channel)

    def GetAIVelocityIEPESensorSensitivityUnits(self, channel, data):
        return DAQmxGetAIVelocityIEPESensorSensitivityUnits(self.task_handle, channel, data)

    def SetAIVelocityIEPESensorSensitivityUnits(self, channel, data):
        return DAQmxSetAIVelocityIEPESensorSensitivityUnits(self.task_handle, channel, data)

    def ResetAIVelocityIEPESensorSensitivityUnits(self, channel):
        return DAQmxResetAIVelocityIEPESensorSensitivityUnits(self.task_handle, channel)

    def GetAIForceUnits(self, channel, data):
        return DAQmxGetAIForceUnits(self.task_handle, channel, data)

    def SetAIForceUnits(self, channel, data):
        return DAQmxSetAIForceUnits(self.task_handle, channel, data)

    def ResetAIForceUnits(self, channel):
        return DAQmxResetAIForceUnits(self.task_handle, channel)

    def GetAIForceIEPESensorSensitivity(self, channel, data):
        return DAQmxGetAIForceIEPESensorSensitivity(self.task_handle, channel, data)

    def SetAIForceIEPESensorSensitivity(self, channel, data):
        return DAQmxSetAIForceIEPESensorSensitivity(self.task_handle, channel, data)

    def ResetAIForceIEPESensorSensitivity(self, channel):
        return DAQmxResetAIForceIEPESensorSensitivity(self.task_handle, channel)

    def GetAIForceIEPESensorSensitivityUnits(self, channel, data):
        return DAQmxGetAIForceIEPESensorSensitivityUnits(self.task_handle, channel, data)

    def SetAIForceIEPESensorSensitivityUnits(self, channel, data):
        return DAQmxSetAIForceIEPESensorSensitivityUnits(self.task_handle, channel, data)

    def ResetAIForceIEPESensorSensitivityUnits(self, channel):
        return DAQmxResetAIForceIEPESensorSensitivityUnits(self.task_handle, channel)

    def GetAIPressureUnits(self, channel, data):
        return DAQmxGetAIPressureUnits(self.task_handle, channel, data)

    def SetAIPressureUnits(self, channel, data):
        return DAQmxSetAIPressureUnits(self.task_handle, channel, data)

    def ResetAIPressureUnits(self, channel):
        return DAQmxResetAIPressureUnits(self.task_handle, channel)

    def GetAITorqueUnits(self, channel, data):
        return DAQmxGetAITorqueUnits(self.task_handle, channel, data)

    def SetAITorqueUnits(self, channel, data):
        return DAQmxSetAITorqueUnits(self.task_handle, channel, data)

    def ResetAITorqueUnits(self, channel):
        return DAQmxResetAITorqueUnits(self.task_handle, channel)

    def GetAIBridgeUnits(self, channel, data):
        return DAQmxGetAIBridgeUnits(self.task_handle, channel, data)

    def SetAIBridgeUnits(self, channel, data):
        return DAQmxSetAIBridgeUnits(self.task_handle, channel, data)

    def ResetAIBridgeUnits(self, channel):
        return DAQmxResetAIBridgeUnits(self.task_handle, channel)

    def GetAIBridgeElectricalUnits(self, channel, data):
        return DAQmxGetAIBridgeElectricalUnits(self.task_handle, channel, data)

    def SetAIBridgeElectricalUnits(self, channel, data):
        return DAQmxSetAIBridgeElectricalUnits(self.task_handle, channel, data)

    def ResetAIBridgeElectricalUnits(self, channel):
        return DAQmxResetAIBridgeElectricalUnits(self.task_handle, channel)

    def GetAIBridgePhysicalUnits(self, channel, data):
        return DAQmxGetAIBridgePhysicalUnits(self.task_handle, channel, data)

    def SetAIBridgePhysicalUnits(self, channel, data):
        return DAQmxSetAIBridgePhysicalUnits(self.task_handle, channel, data)

    def ResetAIBridgePhysicalUnits(self, channel):
        return DAQmxResetAIBridgePhysicalUnits(self.task_handle, channel)

    def GetAIBridgeScaleType(self, channel, data):
        return DAQmxGetAIBridgeScaleType(self.task_handle, channel, data)

    def SetAIBridgeScaleType(self, channel, data):
        return DAQmxSetAIBridgeScaleType(self.task_handle, channel, data)

    def ResetAIBridgeScaleType(self, channel):
        return DAQmxResetAIBridgeScaleType(self.task_handle, channel)

    def GetAIBridgeTwoPointLinFirstElectricalVal(self, channel, data):
        return DAQmxGetAIBridgeTwoPointLinFirstElectricalVal(self.task_handle, channel, data)

    def SetAIBridgeTwoPointLinFirstElectricalVal(self, channel, data):
        return DAQmxSetAIBridgeTwoPointLinFirstElectricalVal(self.task_handle, channel, data)

    def ResetAIBridgeTwoPointLinFirstElectricalVal(self, channel):
        return DAQmxResetAIBridgeTwoPointLinFirstElectricalVal(self.task_handle, channel)

    def GetAIBridgeTwoPointLinFirstPhysicalVal(self, channel, data):
        return DAQmxGetAIBridgeTwoPointLinFirstPhysicalVal(self.task_handle, channel, data)

    def SetAIBridgeTwoPointLinFirstPhysicalVal(self, channel, data):
        return DAQmxSetAIBridgeTwoPointLinFirstPhysicalVal(self.task_handle, channel, data)

    def ResetAIBridgeTwoPointLinFirstPhysicalVal(self, channel):
        return DAQmxResetAIBridgeTwoPointLinFirstPhysicalVal(self.task_handle, channel)

    def GetAIBridgeTwoPointLinSecondElectricalVal(self, channel, data):
        return DAQmxGetAIBridgeTwoPointLinSecondElectricalVal(self.task_handle, channel, data)

    def SetAIBridgeTwoPointLinSecondElectricalVal(self, channel, data):
        return DAQmxSetAIBridgeTwoPointLinSecondElectricalVal(self.task_handle, channel, data)

    def ResetAIBridgeTwoPointLinSecondElectricalVal(self, channel):
        return DAQmxResetAIBridgeTwoPointLinSecondElectricalVal(self.task_handle, channel)

    def GetAIBridgeTwoPointLinSecondPhysicalVal(self, channel, data):
        return DAQmxGetAIBridgeTwoPointLinSecondPhysicalVal(self.task_handle, channel, data)

    def SetAIBridgeTwoPointLinSecondPhysicalVal(self, channel, data):
        return DAQmxSetAIBridgeTwoPointLinSecondPhysicalVal(self.task_handle, channel, data)

    def ResetAIBridgeTwoPointLinSecondPhysicalVal(self, channel):
        return DAQmxResetAIBridgeTwoPointLinSecondPhysicalVal(self.task_handle, channel)

    def GetAIBridgeTableElectricalVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIBridgeTableElectricalVals(self.task_handle, channel, data, arraySizeInElements)

    def SetAIBridgeTableElectricalVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIBridgeTableElectricalVals(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIBridgeTableElectricalVals(self, channel):
        return DAQmxResetAIBridgeTableElectricalVals(self.task_handle, channel)

    def GetAIBridgeTablePhysicalVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIBridgeTablePhysicalVals(self.task_handle, channel, data, arraySizeInElements)

    def SetAIBridgeTablePhysicalVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIBridgeTablePhysicalVals(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIBridgeTablePhysicalVals(self, channel):
        return DAQmxResetAIBridgeTablePhysicalVals(self.task_handle, channel)

    def GetAIBridgePolyForwardCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIBridgePolyForwardCoeff(self.task_handle, channel, data, arraySizeInElements)

    def SetAIBridgePolyForwardCoeff(self, channel, data, arraySizeInElements):
        return DAQmxSetAIBridgePolyForwardCoeff(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIBridgePolyForwardCoeff(self, channel):
        return DAQmxResetAIBridgePolyForwardCoeff(self.task_handle, channel)

    def GetAIBridgePolyReverseCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIBridgePolyReverseCoeff(self.task_handle, channel, data, arraySizeInElements)

    def SetAIBridgePolyReverseCoeff(self, channel, data, arraySizeInElements):
        return DAQmxSetAIBridgePolyReverseCoeff(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIBridgePolyReverseCoeff(self, channel):
        return DAQmxResetAIBridgePolyReverseCoeff(self.task_handle, channel)

    def GetAIIsTEDS(self, channel, data):
        return DAQmxGetAIIsTEDS(self.task_handle, channel, data)

    def GetAITEDSUnits(self, channel, data, bufferSize):
        return DAQmxGetAITEDSUnits(self.task_handle, channel, data, bufferSize)

    def GetAICoupling(self, channel, data):
        return DAQmxGetAICoupling(self.task_handle, channel, data)

    def SetAICoupling(self, channel, data):
        return DAQmxSetAICoupling(self.task_handle, channel, data)

    def ResetAICoupling(self, channel):
        return DAQmxResetAICoupling(self.task_handle, channel)

    def GetAIImpedance(self, channel, data):
        return DAQmxGetAIImpedance(self.task_handle, channel, data)

    def SetAIImpedance(self, channel, data):
        return DAQmxSetAIImpedance(self.task_handle, channel, data)

    def ResetAIImpedance(self, channel):
        return DAQmxResetAIImpedance(self.task_handle, channel)

    def GetAITermCfg(self, channel, data):
        return DAQmxGetAITermCfg(self.task_handle, channel, data)

    def SetAITermCfg(self, channel, data):
        return DAQmxSetAITermCfg(self.task_handle, channel, data)

    def ResetAITermCfg(self, channel):
        return DAQmxResetAITermCfg(self.task_handle, channel)

    def GetAIInputSrc(self, channel, data, bufferSize):
        return DAQmxGetAIInputSrc(self.task_handle, channel, data, bufferSize)

    def SetAIInputSrc(self, channel, data):
        return DAQmxSetAIInputSrc(self.task_handle, channel, data)

    def ResetAIInputSrc(self, channel):
        return DAQmxResetAIInputSrc(self.task_handle, channel)

    def GetAIResistanceCfg(self, channel, data):
        return DAQmxGetAIResistanceCfg(self.task_handle, channel, data)

    def SetAIResistanceCfg(self, channel, data):
        return DAQmxSetAIResistanceCfg(self.task_handle, channel, data)

    def ResetAIResistanceCfg(self, channel):
        return DAQmxResetAIResistanceCfg(self.task_handle, channel)

    def GetAILeadWireResistance(self, channel, data):
        return DAQmxGetAILeadWireResistance(self.task_handle, channel, data)

    def SetAILeadWireResistance(self, channel, data):
        return DAQmxSetAILeadWireResistance(self.task_handle, channel, data)

    def ResetAILeadWireResistance(self, channel):
        return DAQmxResetAILeadWireResistance(self.task_handle, channel)

    def GetAIBridgeCfg(self, channel, data):
        return DAQmxGetAIBridgeCfg(self.task_handle, channel, data)

    def SetAIBridgeCfg(self, channel, data):
        return DAQmxSetAIBridgeCfg(self.task_handle, channel, data)

    def ResetAIBridgeCfg(self, channel):
        return DAQmxResetAIBridgeCfg(self.task_handle, channel)

    def GetAIBridgeNomResistance(self, channel, data):
        return DAQmxGetAIBridgeNomResistance(self.task_handle, channel, data)

    def SetAIBridgeNomResistance(self, channel, data):
        return DAQmxSetAIBridgeNomResistance(self.task_handle, channel, data)

    def ResetAIBridgeNomResistance(self, channel):
        return DAQmxResetAIBridgeNomResistance(self.task_handle, channel)

    def GetAIBridgeInitialVoltage(self, channel, data):
        return DAQmxGetAIBridgeInitialVoltage(self.task_handle, channel, data)

    def SetAIBridgeInitialVoltage(self, channel, data):
        return DAQmxSetAIBridgeInitialVoltage(self.task_handle, channel, data)

    def ResetAIBridgeInitialVoltage(self, channel):
        return DAQmxResetAIBridgeInitialVoltage(self.task_handle, channel)

    def GetAIBridgeInitialRatio(self, channel, data):
        return DAQmxGetAIBridgeInitialRatio(self.task_handle, channel, data)

    def SetAIBridgeInitialRatio(self, channel, data):
        return DAQmxSetAIBridgeInitialRatio(self.task_handle, channel, data)

    def ResetAIBridgeInitialRatio(self, channel):
        return DAQmxResetAIBridgeInitialRatio(self.task_handle, channel)

    def GetAIBridgeShuntCalEnable(self, channel, data):
        return DAQmxGetAIBridgeShuntCalEnable(self.task_handle, channel, data)

    def SetAIBridgeShuntCalEnable(self, channel, data):
        return DAQmxSetAIBridgeShuntCalEnable(self.task_handle, channel, data)

    def ResetAIBridgeShuntCalEnable(self, channel):
        return DAQmxResetAIBridgeShuntCalEnable(self.task_handle, channel)

    def GetAIBridgeShuntCalSelect(self, channel, data):
        return DAQmxGetAIBridgeShuntCalSelect(self.task_handle, channel, data)

    def SetAIBridgeShuntCalSelect(self, channel, data):
        return DAQmxSetAIBridgeShuntCalSelect(self.task_handle, channel, data)

    def ResetAIBridgeShuntCalSelect(self, channel):
        return DAQmxResetAIBridgeShuntCalSelect(self.task_handle, channel)

    def GetAIBridgeShuntCalGainAdjust(self, channel, data):
        return DAQmxGetAIBridgeShuntCalGainAdjust(self.task_handle, channel, data)

    def SetAIBridgeShuntCalGainAdjust(self, channel, data):
        return DAQmxSetAIBridgeShuntCalGainAdjust(self.task_handle, channel, data)

    def ResetAIBridgeShuntCalGainAdjust(self, channel):
        return DAQmxResetAIBridgeShuntCalGainAdjust(self.task_handle, channel)

    def GetAIBridgeShuntCalShuntCalAResistance(self, channel, data):
        return DAQmxGetAIBridgeShuntCalShuntCalAResistance(self.task_handle, channel, data)

    def SetAIBridgeShuntCalShuntCalAResistance(self, channel, data):
        return DAQmxSetAIBridgeShuntCalShuntCalAResistance(self.task_handle, channel, data)

    def ResetAIBridgeShuntCalShuntCalAResistance(self, channel):
        return DAQmxResetAIBridgeShuntCalShuntCalAResistance(self.task_handle, channel)

    def GetAIBridgeShuntCalShuntCalAActualResistance(self, channel, data):
        return DAQmxGetAIBridgeShuntCalShuntCalAActualResistance(self.task_handle, channel, data)

    def SetAIBridgeShuntCalShuntCalAActualResistance(self, channel, data):
        return DAQmxSetAIBridgeShuntCalShuntCalAActualResistance(self.task_handle, channel, data)

    def ResetAIBridgeShuntCalShuntCalAActualResistance(self, channel):
        return DAQmxResetAIBridgeShuntCalShuntCalAActualResistance(self.task_handle, channel)

    def GetAIBridgeBalanceCoarsePot(self, channel, data):
        return DAQmxGetAIBridgeBalanceCoarsePot(self.task_handle, channel, data)

    def SetAIBridgeBalanceCoarsePot(self, channel, data):
        return DAQmxSetAIBridgeBalanceCoarsePot(self.task_handle, channel, data)

    def ResetAIBridgeBalanceCoarsePot(self, channel):
        return DAQmxResetAIBridgeBalanceCoarsePot(self.task_handle, channel)

    def GetAIBridgeBalanceFinePot(self, channel, data):
        return DAQmxGetAIBridgeBalanceFinePot(self.task_handle, channel, data)

    def SetAIBridgeBalanceFinePot(self, channel, data):
        return DAQmxSetAIBridgeBalanceFinePot(self.task_handle, channel, data)

    def ResetAIBridgeBalanceFinePot(self, channel):
        return DAQmxResetAIBridgeBalanceFinePot(self.task_handle, channel)

    def GetAICurrentShuntLoc(self, channel, data):
        return DAQmxGetAICurrentShuntLoc(self.task_handle, channel, data)

    def SetAICurrentShuntLoc(self, channel, data):
        return DAQmxSetAICurrentShuntLoc(self.task_handle, channel, data)

    def ResetAICurrentShuntLoc(self, channel):
        return DAQmxResetAICurrentShuntLoc(self.task_handle, channel)

    def GetAICurrentShuntResistance(self, channel, data):
        return DAQmxGetAICurrentShuntResistance(self.task_handle, channel, data)

    def SetAICurrentShuntResistance(self, channel, data):
        return DAQmxSetAICurrentShuntResistance(self.task_handle, channel, data)

    def ResetAICurrentShuntResistance(self, channel):
        return DAQmxResetAICurrentShuntResistance(self.task_handle, channel)

    def GetAIExcitSrc(self, channel, data):
        return DAQmxGetAIExcitSrc(self.task_handle, channel, data)

    def SetAIExcitSrc(self, channel, data):
        return DAQmxSetAIExcitSrc(self.task_handle, channel, data)

    def ResetAIExcitSrc(self, channel):
        return DAQmxResetAIExcitSrc(self.task_handle, channel)

    def GetAIExcitVal(self, channel, data):
        return DAQmxGetAIExcitVal(self.task_handle, channel, data)

    def SetAIExcitVal(self, channel, data):
        return DAQmxSetAIExcitVal(self.task_handle, channel, data)

    def ResetAIExcitVal(self, channel):
        return DAQmxResetAIExcitVal(self.task_handle, channel)

    def GetAIExcitUseForScaling(self, channel, data):
        return DAQmxGetAIExcitUseForScaling(self.task_handle, channel, data)

    def SetAIExcitUseForScaling(self, channel, data):
        return DAQmxSetAIExcitUseForScaling(self.task_handle, channel, data)

    def ResetAIExcitUseForScaling(self, channel):
        return DAQmxResetAIExcitUseForScaling(self.task_handle, channel)

    def GetAIExcitUseMultiplexed(self, channel, data):
        return DAQmxGetAIExcitUseMultiplexed(self.task_handle, channel, data)

    def SetAIExcitUseMultiplexed(self, channel, data):
        return DAQmxSetAIExcitUseMultiplexed(self.task_handle, channel, data)

    def ResetAIExcitUseMultiplexed(self, channel):
        return DAQmxResetAIExcitUseMultiplexed(self.task_handle, channel)

    def GetAIExcitActualVal(self, channel, data):
        return DAQmxGetAIExcitActualVal(self.task_handle, channel, data)

    def SetAIExcitActualVal(self, channel, data):
        return DAQmxSetAIExcitActualVal(self.task_handle, channel, data)

    def ResetAIExcitActualVal(self, channel):
        return DAQmxResetAIExcitActualVal(self.task_handle, channel)

    def GetAIExcitDCorAC(self, channel, data):
        return DAQmxGetAIExcitDCorAC(self.task_handle, channel, data)

    def SetAIExcitDCorAC(self, channel, data):
        return DAQmxSetAIExcitDCorAC(self.task_handle, channel, data)

    def ResetAIExcitDCorAC(self, channel):
        return DAQmxResetAIExcitDCorAC(self.task_handle, channel)

    def GetAIExcitVoltageOrCurrent(self, channel, data):
        return DAQmxGetAIExcitVoltageOrCurrent(self.task_handle, channel, data)

    def SetAIExcitVoltageOrCurrent(self, channel, data):
        return DAQmxSetAIExcitVoltageOrCurrent(self.task_handle, channel, data)

    def ResetAIExcitVoltageOrCurrent(self, channel):
        return DAQmxResetAIExcitVoltageOrCurrent(self.task_handle, channel)

    def GetAIACExcitFreq(self, channel, data):
        return DAQmxGetAIACExcitFreq(self.task_handle, channel, data)

    def SetAIACExcitFreq(self, channel, data):
        return DAQmxSetAIACExcitFreq(self.task_handle, channel, data)

    def ResetAIACExcitFreq(self, channel):
        return DAQmxResetAIACExcitFreq(self.task_handle, channel)

    def GetAIACExcitSyncEnable(self, channel, data):
        return DAQmxGetAIACExcitSyncEnable(self.task_handle, channel, data)

    def SetAIACExcitSyncEnable(self, channel, data):
        return DAQmxSetAIACExcitSyncEnable(self.task_handle, channel, data)

    def ResetAIACExcitSyncEnable(self, channel):
        return DAQmxResetAIACExcitSyncEnable(self.task_handle, channel)

    def GetAIACExcitWireMode(self, channel, data):
        return DAQmxGetAIACExcitWireMode(self.task_handle, channel, data)

    def SetAIACExcitWireMode(self, channel, data):
        return DAQmxSetAIACExcitWireMode(self.task_handle, channel, data)

    def ResetAIACExcitWireMode(self, channel):
        return DAQmxResetAIACExcitWireMode(self.task_handle, channel)

    def GetAIOpenThrmcplDetectEnable(self, channel, data):
        return DAQmxGetAIOpenThrmcplDetectEnable(self.task_handle, channel, data)

    def SetAIOpenThrmcplDetectEnable(self, channel, data):
        return DAQmxSetAIOpenThrmcplDetectEnable(self.task_handle, channel, data)

    def ResetAIOpenThrmcplDetectEnable(self, channel):
        return DAQmxResetAIOpenThrmcplDetectEnable(self.task_handle, channel)

    def GetAIThrmcplLeadOffsetVoltage(self, channel, data):
        return DAQmxGetAIThrmcplLeadOffsetVoltage(self.task_handle, channel, data)

    def SetAIThrmcplLeadOffsetVoltage(self, channel, data):
        return DAQmxSetAIThrmcplLeadOffsetVoltage(self.task_handle, channel, data)

    def ResetAIThrmcplLeadOffsetVoltage(self, channel):
        return DAQmxResetAIThrmcplLeadOffsetVoltage(self.task_handle, channel)

    def GetAIAtten(self, channel, data):
        return DAQmxGetAIAtten(self.task_handle, channel, data)

    def SetAIAtten(self, channel, data):
        return DAQmxSetAIAtten(self.task_handle, channel, data)

    def ResetAIAtten(self, channel):
        return DAQmxResetAIAtten(self.task_handle, channel)

    def GetAIProbeAtten(self, channel, data):
        return DAQmxGetAIProbeAtten(self.task_handle, channel, data)

    def SetAIProbeAtten(self, channel, data):
        return DAQmxSetAIProbeAtten(self.task_handle, channel, data)

    def ResetAIProbeAtten(self, channel):
        return DAQmxResetAIProbeAtten(self.task_handle, channel)

    def GetAILowpassEnable(self, channel, data):
        return DAQmxGetAILowpassEnable(self.task_handle, channel, data)

    def SetAILowpassEnable(self, channel, data):
        return DAQmxSetAILowpassEnable(self.task_handle, channel, data)

    def ResetAILowpassEnable(self, channel):
        return DAQmxResetAILowpassEnable(self.task_handle, channel)

    def GetAILowpassCutoffFreq(self, channel, data):
        return DAQmxGetAILowpassCutoffFreq(self.task_handle, channel, data)

    def SetAILowpassCutoffFreq(self, channel, data):
        return DAQmxSetAILowpassCutoffFreq(self.task_handle, channel, data)

    def ResetAILowpassCutoffFreq(self, channel):
        return DAQmxResetAILowpassCutoffFreq(self.task_handle, channel)

    def GetAILowpassSwitchCapClkSrc(self, channel, data):
        return DAQmxGetAILowpassSwitchCapClkSrc(self.task_handle, channel, data)

    def SetAILowpassSwitchCapClkSrc(self, channel, data):
        return DAQmxSetAILowpassSwitchCapClkSrc(self.task_handle, channel, data)

    def ResetAILowpassSwitchCapClkSrc(self, channel):
        return DAQmxResetAILowpassSwitchCapClkSrc(self.task_handle, channel)

    def GetAILowpassSwitchCapExtClkFreq(self, channel, data):
        return DAQmxGetAILowpassSwitchCapExtClkFreq(self.task_handle, channel, data)

    def SetAILowpassSwitchCapExtClkFreq(self, channel, data):
        return DAQmxSetAILowpassSwitchCapExtClkFreq(self.task_handle, channel, data)

    def ResetAILowpassSwitchCapExtClkFreq(self, channel):
        return DAQmxResetAILowpassSwitchCapExtClkFreq(self.task_handle, channel)

    def GetAILowpassSwitchCapExtClkDiv(self, channel, data):
        return DAQmxGetAILowpassSwitchCapExtClkDiv(self.task_handle, channel, data)

    def SetAILowpassSwitchCapExtClkDiv(self, channel, data):
        return DAQmxSetAILowpassSwitchCapExtClkDiv(self.task_handle, channel, data)

    def ResetAILowpassSwitchCapExtClkDiv(self, channel):
        return DAQmxResetAILowpassSwitchCapExtClkDiv(self.task_handle, channel)

    def GetAILowpassSwitchCapOutClkDiv(self, channel, data):
        return DAQmxGetAILowpassSwitchCapOutClkDiv(self.task_handle, channel, data)

    def SetAILowpassSwitchCapOutClkDiv(self, channel, data):
        return DAQmxSetAILowpassSwitchCapOutClkDiv(self.task_handle, channel, data)

    def ResetAILowpassSwitchCapOutClkDiv(self, channel):
        return DAQmxResetAILowpassSwitchCapOutClkDiv(self.task_handle, channel)

    def GetAIFilterDelay(self, channel, data):
        return DAQmxGetAIFilterDelay(self.task_handle, channel, data)

    def GetAIRemoveFilterDelay(self, channel, data):
        return DAQmxGetAIRemoveFilterDelay(self.task_handle, channel, data)

    def SetAIRemoveFilterDelay(self, channel, data):
        return DAQmxSetAIRemoveFilterDelay(self.task_handle, channel, data)

    def ResetAIRemoveFilterDelay(self, channel):
        return DAQmxResetAIRemoveFilterDelay(self.task_handle, channel)

    def GetAIAveragingWinSize(self, channel, data):
        return DAQmxGetAIAveragingWinSize(self.task_handle, channel, data)

    def SetAIAveragingWinSize(self, channel, data):
        return DAQmxSetAIAveragingWinSize(self.task_handle, channel, data)

    def ResetAIAveragingWinSize(self, channel):
        return DAQmxResetAIAveragingWinSize(self.task_handle, channel)

    def GetAIResolutionUnits(self, channel, data):
        return DAQmxGetAIResolutionUnits(self.task_handle, channel, data)

    def GetAIResolution(self, channel, data):
        return DAQmxGetAIResolution(self.task_handle, channel, data)

    def GetAIRawSampSize(self, channel, data):
        return DAQmxGetAIRawSampSize(self.task_handle, channel, data)

    def GetAIRawSampJustification(self, channel, data):
        return DAQmxGetAIRawSampJustification(self.task_handle, channel, data)

    def GetAIADCTimingMode(self, channel, data):
        return DAQmxGetAIADCTimingMode(self.task_handle, channel, data)

    def SetAIADCTimingMode(self, channel, data):
        return DAQmxSetAIADCTimingMode(self.task_handle, channel, data)

    def ResetAIADCTimingMode(self, channel):
        return DAQmxResetAIADCTimingMode(self.task_handle, channel)

    def GetAIADCCustomTimingMode(self, channel, data):
        return DAQmxGetAIADCCustomTimingMode(self.task_handle, channel, data)

    def SetAIADCCustomTimingMode(self, channel, data):
        return DAQmxSetAIADCCustomTimingMode(self.task_handle, channel, data)

    def ResetAIADCCustomTimingMode(self, channel):
        return DAQmxResetAIADCCustomTimingMode(self.task_handle, channel)

    def GetAIDitherEnable(self, channel, data):
        return DAQmxGetAIDitherEnable(self.task_handle, channel, data)

    def SetAIDitherEnable(self, channel, data):
        return DAQmxSetAIDitherEnable(self.task_handle, channel, data)

    def ResetAIDitherEnable(self, channel):
        return DAQmxResetAIDitherEnable(self.task_handle, channel)

    def GetAIChanCalHasValidCalInfo(self, channel, data):
        return DAQmxGetAIChanCalHasValidCalInfo(self.task_handle, channel, data)

    def GetAIChanCalEnableCal(self, channel, data):
        return DAQmxGetAIChanCalEnableCal(self.task_handle, channel, data)

    def SetAIChanCalEnableCal(self, channel, data):
        return DAQmxSetAIChanCalEnableCal(self.task_handle, channel, data)

    def ResetAIChanCalEnableCal(self, channel):
        return DAQmxResetAIChanCalEnableCal(self.task_handle, channel)

    def GetAIChanCalApplyCalIfExp(self, channel, data):
        return DAQmxGetAIChanCalApplyCalIfExp(self.task_handle, channel, data)

    def SetAIChanCalApplyCalIfExp(self, channel, data):
        return DAQmxSetAIChanCalApplyCalIfExp(self.task_handle, channel, data)

    def ResetAIChanCalApplyCalIfExp(self, channel):
        return DAQmxResetAIChanCalApplyCalIfExp(self.task_handle, channel)

    def GetAIChanCalScaleType(self, channel, data):
        return DAQmxGetAIChanCalScaleType(self.task_handle, channel, data)

    def SetAIChanCalScaleType(self, channel, data):
        return DAQmxSetAIChanCalScaleType(self.task_handle, channel, data)

    def ResetAIChanCalScaleType(self, channel):
        return DAQmxResetAIChanCalScaleType(self.task_handle, channel)

    def GetAIChanCalTablePreScaledVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalTablePreScaledVals(self.task_handle, channel, data, arraySizeInElements)

    def SetAIChanCalTablePreScaledVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalTablePreScaledVals(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIChanCalTablePreScaledVals(self, channel):
        return DAQmxResetAIChanCalTablePreScaledVals(self.task_handle, channel)

    def GetAIChanCalTableScaledVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalTableScaledVals(self.task_handle, channel, data, arraySizeInElements)

    def SetAIChanCalTableScaledVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalTableScaledVals(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIChanCalTableScaledVals(self, channel):
        return DAQmxResetAIChanCalTableScaledVals(self.task_handle, channel)

    def GetAIChanCalPolyForwardCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalPolyForwardCoeff(self.task_handle, channel, data, arraySizeInElements)

    def SetAIChanCalPolyForwardCoeff(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalPolyForwardCoeff(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIChanCalPolyForwardCoeff(self, channel):
        return DAQmxResetAIChanCalPolyForwardCoeff(self.task_handle, channel)

    def GetAIChanCalPolyReverseCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalPolyReverseCoeff(self.task_handle, channel, data, arraySizeInElements)

    def SetAIChanCalPolyReverseCoeff(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalPolyReverseCoeff(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIChanCalPolyReverseCoeff(self, channel):
        return DAQmxResetAIChanCalPolyReverseCoeff(self.task_handle, channel)

    def GetAIChanCalOperatorName(self, channel, data, bufferSize):
        return DAQmxGetAIChanCalOperatorName(self.task_handle, channel, data, bufferSize)

    def SetAIChanCalOperatorName(self, channel, data):
        return DAQmxSetAIChanCalOperatorName(self.task_handle, channel, data)

    def ResetAIChanCalOperatorName(self, channel):
        return DAQmxResetAIChanCalOperatorName(self.task_handle, channel)

    def GetAIChanCalDesc(self, channel, data, bufferSize):
        return DAQmxGetAIChanCalDesc(self.task_handle, channel, data, bufferSize)

    def SetAIChanCalDesc(self, channel, data):
        return DAQmxSetAIChanCalDesc(self.task_handle, channel, data)

    def ResetAIChanCalDesc(self, channel):
        return DAQmxResetAIChanCalDesc(self.task_handle, channel)

    def GetAIChanCalVerifRefVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalVerifRefVals(self.task_handle, channel, data, arraySizeInElements)

    def SetAIChanCalVerifRefVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalVerifRefVals(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIChanCalVerifRefVals(self, channel):
        return DAQmxResetAIChanCalVerifRefVals(self.task_handle, channel)

    def GetAIChanCalVerifAcqVals(self, channel, data, arraySizeInElements):
        return DAQmxGetAIChanCalVerifAcqVals(self.task_handle, channel, data, arraySizeInElements)

    def SetAIChanCalVerifAcqVals(self, channel, data, arraySizeInElements):
        return DAQmxSetAIChanCalVerifAcqVals(self.task_handle, channel, data, arraySizeInElements)

    def ResetAIChanCalVerifAcqVals(self, channel):
        return DAQmxResetAIChanCalVerifAcqVals(self.task_handle, channel)

    def GetAIRngHigh(self, channel, data):
        return DAQmxGetAIRngHigh(self.task_handle, channel, data)

    def SetAIRngHigh(self, channel, data):
        return DAQmxSetAIRngHigh(self.task_handle, channel, data)

    def ResetAIRngHigh(self, channel):
        return DAQmxResetAIRngHigh(self.task_handle, channel)

    def GetAIRngLow(self, channel, data):
        return DAQmxGetAIRngLow(self.task_handle, channel, data)

    def SetAIRngLow(self, channel, data):
        return DAQmxSetAIRngLow(self.task_handle, channel, data)

    def ResetAIRngLow(self, channel):
        return DAQmxResetAIRngLow(self.task_handle, channel)

    def GetAIDCOffset(self, channel, data):
        return DAQmxGetAIDCOffset(self.task_handle, channel, data)

    def SetAIDCOffset(self, channel, data):
        return DAQmxSetAIDCOffset(self.task_handle, channel, data)

    def ResetAIDCOffset(self, channel):
        return DAQmxResetAIDCOffset(self.task_handle, channel)

    def GetAIGain(self, channel, data):
        return DAQmxGetAIGain(self.task_handle, channel, data)

    def SetAIGain(self, channel, data):
        return DAQmxSetAIGain(self.task_handle, channel, data)

    def ResetAIGain(self, channel):
        return DAQmxResetAIGain(self.task_handle, channel)

    def GetAISampAndHoldEnable(self, channel, data):
        return DAQmxGetAISampAndHoldEnable(self.task_handle, channel, data)

    def SetAISampAndHoldEnable(self, channel, data):
        return DAQmxSetAISampAndHoldEnable(self.task_handle, channel, data)

    def ResetAISampAndHoldEnable(self, channel):
        return DAQmxResetAISampAndHoldEnable(self.task_handle, channel)

    def GetAIAutoZeroMode(self, channel, data):
        return DAQmxGetAIAutoZeroMode(self.task_handle, channel, data)

    def SetAIAutoZeroMode(self, channel, data):
        return DAQmxSetAIAutoZeroMode(self.task_handle, channel, data)

    def ResetAIAutoZeroMode(self, channel):
        return DAQmxResetAIAutoZeroMode(self.task_handle, channel)

    def GetAIDataXferMech(self, channel, data):
        return DAQmxGetAIDataXferMech(self.task_handle, channel, data)

    def SetAIDataXferMech(self, channel, data):
        return DAQmxSetAIDataXferMech(self.task_handle, channel, data)

    def ResetAIDataXferMech(self, channel):
        return DAQmxResetAIDataXferMech(self.task_handle, channel)

    def GetAIDataXferReqCond(self, channel, data):
        return DAQmxGetAIDataXferReqCond(self.task_handle, channel, data)

    def SetAIDataXferReqCond(self, channel, data):
        return DAQmxSetAIDataXferReqCond(self.task_handle, channel, data)

    def ResetAIDataXferReqCond(self, channel):
        return DAQmxResetAIDataXferReqCond(self.task_handle, channel)

    def GetAIDataXferCustomThreshold(self, channel, data):
        return DAQmxGetAIDataXferCustomThreshold(self.task_handle, channel, data)

    def SetAIDataXferCustomThreshold(self, channel, data):
        return DAQmxSetAIDataXferCustomThreshold(self.task_handle, channel, data)

    def ResetAIDataXferCustomThreshold(self, channel):
        return DAQmxResetAIDataXferCustomThreshold(self.task_handle, channel)

    def GetAIUsbXferReqSize(self, channel, data):
        return DAQmxGetAIUsbXferReqSize(self.task_handle, channel, data)

    def SetAIUsbXferReqSize(self, channel, data):
        return DAQmxSetAIUsbXferReqSize(self.task_handle, channel, data)

    def ResetAIUsbXferReqSize(self, channel):
        return DAQmxResetAIUsbXferReqSize(self.task_handle, channel)

    def GetAIUsbXferReqCount(self, channel, data):
        return DAQmxGetAIUsbXferReqCount(self.task_handle, channel, data)

    def SetAIUsbXferReqCount(self, channel, data):
        return DAQmxSetAIUsbXferReqCount(self.task_handle, channel, data)

    def ResetAIUsbXferReqCount(self, channel):
        return DAQmxResetAIUsbXferReqCount(self.task_handle, channel)

    def GetAIMemMapEnable(self, channel, data):
        return DAQmxGetAIMemMapEnable(self.task_handle, channel, data)

    def SetAIMemMapEnable(self, channel, data):
        return DAQmxSetAIMemMapEnable(self.task_handle, channel, data)

    def ResetAIMemMapEnable(self, channel):
        return DAQmxResetAIMemMapEnable(self.task_handle, channel)

    def GetAIRawDataCompressionType(self, channel, data):
        return DAQmxGetAIRawDataCompressionType(self.task_handle, channel, data)

    def SetAIRawDataCompressionType(self, channel, data):
        return DAQmxSetAIRawDataCompressionType(self.task_handle, channel, data)

    def ResetAIRawDataCompressionType(self, channel):
        return DAQmxResetAIRawDataCompressionType(self.task_handle, channel)

    def GetAILossyLSBRemovalCompressedSampSize(self, channel, data):
        return DAQmxGetAILossyLSBRemovalCompressedSampSize(self.task_handle, channel, data)

    def SetAILossyLSBRemovalCompressedSampSize(self, channel, data):
        return DAQmxSetAILossyLSBRemovalCompressedSampSize(self.task_handle, channel, data)

    def ResetAILossyLSBRemovalCompressedSampSize(self, channel):
        return DAQmxResetAILossyLSBRemovalCompressedSampSize(self.task_handle, channel)

    def GetAIDevScalingCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAIDevScalingCoeff(self.task_handle, channel, data, arraySizeInElements)

    def GetAIEnhancedAliasRejectionEnable(self, channel, data):
        return DAQmxGetAIEnhancedAliasRejectionEnable(self.task_handle, channel, data)

    def SetAIEnhancedAliasRejectionEnable(self, channel, data):
        return DAQmxSetAIEnhancedAliasRejectionEnable(self.task_handle, channel, data)

    def ResetAIEnhancedAliasRejectionEnable(self, channel):
        return DAQmxResetAIEnhancedAliasRejectionEnable(self.task_handle, channel)

    def GetAOMax(self, channel, data):
        return DAQmxGetAOMax(self.task_handle, channel, data)

    def SetAOMax(self, channel, data):
        return DAQmxSetAOMax(self.task_handle, channel, data)

    def ResetAOMax(self, channel):
        return DAQmxResetAOMax(self.task_handle, channel)

    def GetAOMin(self, channel, data):
        return DAQmxGetAOMin(self.task_handle, channel, data)

    def SetAOMin(self, channel, data):
        return DAQmxSetAOMin(self.task_handle, channel, data)

    def ResetAOMin(self, channel):
        return DAQmxResetAOMin(self.task_handle, channel)

    def GetAOCustomScaleName(self, channel, data, bufferSize):
        return DAQmxGetAOCustomScaleName(self.task_handle, channel, data, bufferSize)

    def SetAOCustomScaleName(self, channel, data):
        return DAQmxSetAOCustomScaleName(self.task_handle, channel, data)

    def ResetAOCustomScaleName(self, channel):
        return DAQmxResetAOCustomScaleName(self.task_handle, channel)

    def GetAOOutputType(self, channel, data):
        return DAQmxGetAOOutputType(self.task_handle, channel, data)

    def GetAOVoltageUnits(self, channel, data):
        return DAQmxGetAOVoltageUnits(self.task_handle, channel, data)

    def SetAOVoltageUnits(self, channel, data):
        return DAQmxSetAOVoltageUnits(self.task_handle, channel, data)

    def ResetAOVoltageUnits(self, channel):
        return DAQmxResetAOVoltageUnits(self.task_handle, channel)

    def GetAOVoltageCurrentLimit(self, channel, data):
        return DAQmxGetAOVoltageCurrentLimit(self.task_handle, channel, data)

    def SetAOVoltageCurrentLimit(self, channel, data):
        return DAQmxSetAOVoltageCurrentLimit(self.task_handle, channel, data)

    def ResetAOVoltageCurrentLimit(self, channel):
        return DAQmxResetAOVoltageCurrentLimit(self.task_handle, channel)

    def GetAOCurrentUnits(self, channel, data):
        return DAQmxGetAOCurrentUnits(self.task_handle, channel, data)

    def SetAOCurrentUnits(self, channel, data):
        return DAQmxSetAOCurrentUnits(self.task_handle, channel, data)

    def ResetAOCurrentUnits(self, channel):
        return DAQmxResetAOCurrentUnits(self.task_handle, channel)

    def GetAOFuncGenType(self, channel, data):
        return DAQmxGetAOFuncGenType(self.task_handle, channel, data)

    def SetAOFuncGenType(self, channel, data):
        return DAQmxSetAOFuncGenType(self.task_handle, channel, data)

    def ResetAOFuncGenType(self, channel):
        return DAQmxResetAOFuncGenType(self.task_handle, channel)

    def GetAOFuncGenFreq(self, channel, data):
        return DAQmxGetAOFuncGenFreq(self.task_handle, channel, data)

    def SetAOFuncGenFreq(self, channel, data):
        return DAQmxSetAOFuncGenFreq(self.task_handle, channel, data)

    def ResetAOFuncGenFreq(self, channel):
        return DAQmxResetAOFuncGenFreq(self.task_handle, channel)

    def GetAOFuncGenAmplitude(self, channel, data):
        return DAQmxGetAOFuncGenAmplitude(self.task_handle, channel, data)

    def SetAOFuncGenAmplitude(self, channel, data):
        return DAQmxSetAOFuncGenAmplitude(self.task_handle, channel, data)

    def ResetAOFuncGenAmplitude(self, channel):
        return DAQmxResetAOFuncGenAmplitude(self.task_handle, channel)

    def GetAOFuncGenOffset(self, channel, data):
        return DAQmxGetAOFuncGenOffset(self.task_handle, channel, data)

    def SetAOFuncGenOffset(self, channel, data):
        return DAQmxSetAOFuncGenOffset(self.task_handle, channel, data)

    def ResetAOFuncGenOffset(self, channel):
        return DAQmxResetAOFuncGenOffset(self.task_handle, channel)

    def GetAOFuncGenSquareDutyCycle(self, channel, data):
        return DAQmxGetAOFuncGenSquareDutyCycle(self.task_handle, channel, data)

    def SetAOFuncGenSquareDutyCycle(self, channel, data):
        return DAQmxSetAOFuncGenSquareDutyCycle(self.task_handle, channel, data)

    def ResetAOFuncGenSquareDutyCycle(self, channel):
        return DAQmxResetAOFuncGenSquareDutyCycle(self.task_handle, channel)

    def GetAOFuncGenModulationType(self, channel, data):
        return DAQmxGetAOFuncGenModulationType(self.task_handle, channel, data)

    def SetAOFuncGenModulationType(self, channel, data):
        return DAQmxSetAOFuncGenModulationType(self.task_handle, channel, data)

    def ResetAOFuncGenModulationType(self, channel):
        return DAQmxResetAOFuncGenModulationType(self.task_handle, channel)

    def GetAOFuncGenFMDeviation(self, channel, data):
        return DAQmxGetAOFuncGenFMDeviation(self.task_handle, channel, data)

    def SetAOFuncGenFMDeviation(self, channel, data):
        return DAQmxSetAOFuncGenFMDeviation(self.task_handle, channel, data)

    def ResetAOFuncGenFMDeviation(self, channel):
        return DAQmxResetAOFuncGenFMDeviation(self.task_handle, channel)

    def GetAOOutputImpedance(self, channel, data):
        return DAQmxGetAOOutputImpedance(self.task_handle, channel, data)

    def SetAOOutputImpedance(self, channel, data):
        return DAQmxSetAOOutputImpedance(self.task_handle, channel, data)

    def ResetAOOutputImpedance(self, channel):
        return DAQmxResetAOOutputImpedance(self.task_handle, channel)

    def GetAOLoadImpedance(self, channel, data):
        return DAQmxGetAOLoadImpedance(self.task_handle, channel, data)

    def SetAOLoadImpedance(self, channel, data):
        return DAQmxSetAOLoadImpedance(self.task_handle, channel, data)

    def ResetAOLoadImpedance(self, channel):
        return DAQmxResetAOLoadImpedance(self.task_handle, channel)

    def GetAOIdleOutputBehavior(self, channel, data):
        return DAQmxGetAOIdleOutputBehavior(self.task_handle, channel, data)

    def SetAOIdleOutputBehavior(self, channel, data):
        return DAQmxSetAOIdleOutputBehavior(self.task_handle, channel, data)

    def ResetAOIdleOutputBehavior(self, channel):
        return DAQmxResetAOIdleOutputBehavior(self.task_handle, channel)

    def GetAOTermCfg(self, channel, data):
        return DAQmxGetAOTermCfg(self.task_handle, channel, data)

    def SetAOTermCfg(self, channel, data):
        return DAQmxSetAOTermCfg(self.task_handle, channel, data)

    def ResetAOTermCfg(self, channel):
        return DAQmxResetAOTermCfg(self.task_handle, channel)

    def GetAOResolutionUnits(self, channel, data):
        return DAQmxGetAOResolutionUnits(self.task_handle, channel, data)

    def SetAOResolutionUnits(self, channel, data):
        return DAQmxSetAOResolutionUnits(self.task_handle, channel, data)

    def ResetAOResolutionUnits(self, channel):
        return DAQmxResetAOResolutionUnits(self.task_handle, channel)

    def GetAOResolution(self, channel, data):
        return DAQmxGetAOResolution(self.task_handle, channel, data)

    def GetAODACRngHigh(self, channel, data):
        return DAQmxGetAODACRngHigh(self.task_handle, channel, data)

    def SetAODACRngHigh(self, channel, data):
        return DAQmxSetAODACRngHigh(self.task_handle, channel, data)

    def ResetAODACRngHigh(self, channel):
        return DAQmxResetAODACRngHigh(self.task_handle, channel)

    def GetAODACRngLow(self, channel, data):
        return DAQmxGetAODACRngLow(self.task_handle, channel, data)

    def SetAODACRngLow(self, channel, data):
        return DAQmxSetAODACRngLow(self.task_handle, channel, data)

    def ResetAODACRngLow(self, channel):
        return DAQmxResetAODACRngLow(self.task_handle, channel)

    def GetAODACRefConnToGnd(self, channel, data):
        return DAQmxGetAODACRefConnToGnd(self.task_handle, channel, data)

    def SetAODACRefConnToGnd(self, channel, data):
        return DAQmxSetAODACRefConnToGnd(self.task_handle, channel, data)

    def ResetAODACRefConnToGnd(self, channel):
        return DAQmxResetAODACRefConnToGnd(self.task_handle, channel)

    def GetAODACRefAllowConnToGnd(self, channel, data):
        return DAQmxGetAODACRefAllowConnToGnd(self.task_handle, channel, data)

    def SetAODACRefAllowConnToGnd(self, channel, data):
        return DAQmxSetAODACRefAllowConnToGnd(self.task_handle, channel, data)

    def ResetAODACRefAllowConnToGnd(self, channel):
        return DAQmxResetAODACRefAllowConnToGnd(self.task_handle, channel)

    def GetAODACRefSrc(self, channel, data):
        return DAQmxGetAODACRefSrc(self.task_handle, channel, data)

    def SetAODACRefSrc(self, channel, data):
        return DAQmxSetAODACRefSrc(self.task_handle, channel, data)

    def ResetAODACRefSrc(self, channel):
        return DAQmxResetAODACRefSrc(self.task_handle, channel)

    def GetAODACRefExtSrc(self, channel, data, bufferSize):
        return DAQmxGetAODACRefExtSrc(self.task_handle, channel, data, bufferSize)

    def SetAODACRefExtSrc(self, channel, data):
        return DAQmxSetAODACRefExtSrc(self.task_handle, channel, data)

    def ResetAODACRefExtSrc(self, channel):
        return DAQmxResetAODACRefExtSrc(self.task_handle, channel)

    def GetAODACRefVal(self, channel, data):
        return DAQmxGetAODACRefVal(self.task_handle, channel, data)

    def SetAODACRefVal(self, channel, data):
        return DAQmxSetAODACRefVal(self.task_handle, channel, data)

    def ResetAODACRefVal(self, channel):
        return DAQmxResetAODACRefVal(self.task_handle, channel)

    def GetAODACOffsetSrc(self, channel, data):
        return DAQmxGetAODACOffsetSrc(self.task_handle, channel, data)

    def SetAODACOffsetSrc(self, channel, data):
        return DAQmxSetAODACOffsetSrc(self.task_handle, channel, data)

    def ResetAODACOffsetSrc(self, channel):
        return DAQmxResetAODACOffsetSrc(self.task_handle, channel)

    def GetAODACOffsetExtSrc(self, channel, data, bufferSize):
        return DAQmxGetAODACOffsetExtSrc(self.task_handle, channel, data, bufferSize)

    def SetAODACOffsetExtSrc(self, channel, data):
        return DAQmxSetAODACOffsetExtSrc(self.task_handle, channel, data)

    def ResetAODACOffsetExtSrc(self, channel):
        return DAQmxResetAODACOffsetExtSrc(self.task_handle, channel)

    def GetAODACOffsetVal(self, channel, data):
        return DAQmxGetAODACOffsetVal(self.task_handle, channel, data)

    def SetAODACOffsetVal(self, channel, data):
        return DAQmxSetAODACOffsetVal(self.task_handle, channel, data)

    def ResetAODACOffsetVal(self, channel):
        return DAQmxResetAODACOffsetVal(self.task_handle, channel)

    def GetAOReglitchEnable(self, channel, data):
        return DAQmxGetAOReglitchEnable(self.task_handle, channel, data)

    def SetAOReglitchEnable(self, channel, data):
        return DAQmxSetAOReglitchEnable(self.task_handle, channel, data)

    def ResetAOReglitchEnable(self, channel):
        return DAQmxResetAOReglitchEnable(self.task_handle, channel)

    def GetAOGain(self, channel, data):
        return DAQmxGetAOGain(self.task_handle, channel, data)

    def SetAOGain(self, channel, data):
        return DAQmxSetAOGain(self.task_handle, channel, data)

    def ResetAOGain(self, channel):
        return DAQmxResetAOGain(self.task_handle, channel)

    def GetAOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxGetAOUseOnlyOnBrdMem(self.task_handle, channel, data)

    def SetAOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxSetAOUseOnlyOnBrdMem(self.task_handle, channel, data)

    def ResetAOUseOnlyOnBrdMem(self, channel):
        return DAQmxResetAOUseOnlyOnBrdMem(self.task_handle, channel)

    def GetAODataXferMech(self, channel, data):
        return DAQmxGetAODataXferMech(self.task_handle, channel, data)

    def SetAODataXferMech(self, channel, data):
        return DAQmxSetAODataXferMech(self.task_handle, channel, data)

    def ResetAODataXferMech(self, channel):
        return DAQmxResetAODataXferMech(self.task_handle, channel)

    def GetAODataXferReqCond(self, channel, data):
        return DAQmxGetAODataXferReqCond(self.task_handle, channel, data)

    def SetAODataXferReqCond(self, channel, data):
        return DAQmxSetAODataXferReqCond(self.task_handle, channel, data)

    def ResetAODataXferReqCond(self, channel):
        return DAQmxResetAODataXferReqCond(self.task_handle, channel)

    def GetAOUsbXferReqSize(self, channel, data):
        return DAQmxGetAOUsbXferReqSize(self.task_handle, channel, data)

    def SetAOUsbXferReqSize(self, channel, data):
        return DAQmxSetAOUsbXferReqSize(self.task_handle, channel, data)

    def ResetAOUsbXferReqSize(self, channel):
        return DAQmxResetAOUsbXferReqSize(self.task_handle, channel)

    def GetAOUsbXferReqCount(self, channel, data):
        return DAQmxGetAOUsbXferReqCount(self.task_handle, channel, data)

    def SetAOUsbXferReqCount(self, channel, data):
        return DAQmxSetAOUsbXferReqCount(self.task_handle, channel, data)

    def ResetAOUsbXferReqCount(self, channel):
        return DAQmxResetAOUsbXferReqCount(self.task_handle, channel)

    def GetAOMemMapEnable(self, channel, data):
        return DAQmxGetAOMemMapEnable(self.task_handle, channel, data)

    def SetAOMemMapEnable(self, channel, data):
        return DAQmxSetAOMemMapEnable(self.task_handle, channel, data)

    def ResetAOMemMapEnable(self, channel):
        return DAQmxResetAOMemMapEnable(self.task_handle, channel)

    def GetAODevScalingCoeff(self, channel, data, arraySizeInElements):
        return DAQmxGetAODevScalingCoeff(self.task_handle, channel, data, arraySizeInElements)

    def GetAOEnhancedImageRejectionEnable(self, channel, data):
        return DAQmxGetAOEnhancedImageRejectionEnable(self.task_handle, channel, data)

    def SetAOEnhancedImageRejectionEnable(self, channel, data):
        return DAQmxSetAOEnhancedImageRejectionEnable(self.task_handle, channel, data)

    def ResetAOEnhancedImageRejectionEnable(self, channel):
        return DAQmxResetAOEnhancedImageRejectionEnable(self.task_handle, channel)

    def GetDIInvertLines(self, channel, data):
        return DAQmxGetDIInvertLines(self.task_handle, channel, data)

    def SetDIInvertLines(self, channel, data):
        return DAQmxSetDIInvertLines(self.task_handle, channel, data)

    def ResetDIInvertLines(self, channel):
        return DAQmxResetDIInvertLines(self.task_handle, channel)

    def GetDINumLines(self, channel, data):
        return DAQmxGetDINumLines(self.task_handle, channel, data)

    def GetDIDigFltrEnable(self, channel, data):
        return DAQmxGetDIDigFltrEnable(self.task_handle, channel, data)

    def SetDIDigFltrEnable(self, channel, data):
        return DAQmxSetDIDigFltrEnable(self.task_handle, channel, data)

    def ResetDIDigFltrEnable(self, channel):
        return DAQmxResetDIDigFltrEnable(self.task_handle, channel)

    def GetDIDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetDIDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetDIDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetDIDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetDIDigFltrMinPulseWidth(self, channel):
        return DAQmxResetDIDigFltrMinPulseWidth(self.task_handle, channel)

    def GetDIDigFltrEnableBusMode(self, channel, data):
        return DAQmxGetDIDigFltrEnableBusMode(self.task_handle, channel, data)

    def SetDIDigFltrEnableBusMode(self, channel, data):
        return DAQmxSetDIDigFltrEnableBusMode(self.task_handle, channel, data)

    def ResetDIDigFltrEnableBusMode(self, channel):
        return DAQmxResetDIDigFltrEnableBusMode(self.task_handle, channel)

    def GetDIDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetDIDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetDIDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetDIDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetDIDigFltrTimebaseSrc(self, channel):
        return DAQmxResetDIDigFltrTimebaseSrc(self.task_handle, channel)

    def GetDIDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetDIDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetDIDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetDIDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetDIDigFltrTimebaseRate(self, channel):
        return DAQmxResetDIDigFltrTimebaseRate(self.task_handle, channel)

    def GetDIDigSyncEnable(self, channel, data):
        return DAQmxGetDIDigSyncEnable(self.task_handle, channel, data)

    def SetDIDigSyncEnable(self, channel, data):
        return DAQmxSetDIDigSyncEnable(self.task_handle, channel, data)

    def ResetDIDigSyncEnable(self, channel):
        return DAQmxResetDIDigSyncEnable(self.task_handle, channel)

    def GetDITristate(self, channel, data):
        return DAQmxGetDITristate(self.task_handle, channel, data)

    def SetDITristate(self, channel, data):
        return DAQmxSetDITristate(self.task_handle, channel, data)

    def ResetDITristate(self, channel):
        return DAQmxResetDITristate(self.task_handle, channel)

    def GetDILogicFamily(self, channel, data):
        return DAQmxGetDILogicFamily(self.task_handle, channel, data)

    def SetDILogicFamily(self, channel, data):
        return DAQmxSetDILogicFamily(self.task_handle, channel, data)

    def ResetDILogicFamily(self, channel):
        return DAQmxResetDILogicFamily(self.task_handle, channel)

    def GetDIDataXferMech(self, channel, data):
        return DAQmxGetDIDataXferMech(self.task_handle, channel, data)

    def SetDIDataXferMech(self, channel, data):
        return DAQmxSetDIDataXferMech(self.task_handle, channel, data)

    def ResetDIDataXferMech(self, channel):
        return DAQmxResetDIDataXferMech(self.task_handle, channel)

    def GetDIDataXferReqCond(self, channel, data):
        return DAQmxGetDIDataXferReqCond(self.task_handle, channel, data)

    def SetDIDataXferReqCond(self, channel, data):
        return DAQmxSetDIDataXferReqCond(self.task_handle, channel, data)

    def ResetDIDataXferReqCond(self, channel):
        return DAQmxResetDIDataXferReqCond(self.task_handle, channel)

    def GetDIUsbXferReqSize(self, channel, data):
        return DAQmxGetDIUsbXferReqSize(self.task_handle, channel, data)

    def SetDIUsbXferReqSize(self, channel, data):
        return DAQmxSetDIUsbXferReqSize(self.task_handle, channel, data)

    def ResetDIUsbXferReqSize(self, channel):
        return DAQmxResetDIUsbXferReqSize(self.task_handle, channel)

    def GetDIUsbXferReqCount(self, channel, data):
        return DAQmxGetDIUsbXferReqCount(self.task_handle, channel, data)

    def SetDIUsbXferReqCount(self, channel, data):
        return DAQmxSetDIUsbXferReqCount(self.task_handle, channel, data)

    def ResetDIUsbXferReqCount(self, channel):
        return DAQmxResetDIUsbXferReqCount(self.task_handle, channel)

    def GetDIMemMapEnable(self, channel, data):
        return DAQmxGetDIMemMapEnable(self.task_handle, channel, data)

    def SetDIMemMapEnable(self, channel, data):
        return DAQmxSetDIMemMapEnable(self.task_handle, channel, data)

    def ResetDIMemMapEnable(self, channel):
        return DAQmxResetDIMemMapEnable(self.task_handle, channel)

    def GetDIAcquireOn(self, channel, data):
        return DAQmxGetDIAcquireOn(self.task_handle, channel, data)

    def SetDIAcquireOn(self, channel, data):
        return DAQmxSetDIAcquireOn(self.task_handle, channel, data)

    def ResetDIAcquireOn(self, channel):
        return DAQmxResetDIAcquireOn(self.task_handle, channel)

    def GetDOOutputDriveType(self, channel, data):
        return DAQmxGetDOOutputDriveType(self.task_handle, channel, data)

    def SetDOOutputDriveType(self, channel, data):
        return DAQmxSetDOOutputDriveType(self.task_handle, channel, data)

    def ResetDOOutputDriveType(self, channel):
        return DAQmxResetDOOutputDriveType(self.task_handle, channel)

    def GetDOInvertLines(self, channel, data):
        return DAQmxGetDOInvertLines(self.task_handle, channel, data)

    def SetDOInvertLines(self, channel, data):
        return DAQmxSetDOInvertLines(self.task_handle, channel, data)

    def ResetDOInvertLines(self, channel):
        return DAQmxResetDOInvertLines(self.task_handle, channel)

    def GetDONumLines(self, channel, data):
        return DAQmxGetDONumLines(self.task_handle, channel, data)

    def GetDOTristate(self, channel, data):
        return DAQmxGetDOTristate(self.task_handle, channel, data)

    def SetDOTristate(self, channel, data):
        return DAQmxSetDOTristate(self.task_handle, channel, data)

    def ResetDOTristate(self, channel):
        return DAQmxResetDOTristate(self.task_handle, channel)

    def GetDOLineStatesStartState(self, channel, data):
        return DAQmxGetDOLineStatesStartState(self.task_handle, channel, data)

    def SetDOLineStatesStartState(self, channel, data):
        return DAQmxSetDOLineStatesStartState(self.task_handle, channel, data)

    def ResetDOLineStatesStartState(self, channel):
        return DAQmxResetDOLineStatesStartState(self.task_handle, channel)

    def GetDOLineStatesPausedState(self, channel, data):
        return DAQmxGetDOLineStatesPausedState(self.task_handle, channel, data)

    def SetDOLineStatesPausedState(self, channel, data):
        return DAQmxSetDOLineStatesPausedState(self.task_handle, channel, data)

    def ResetDOLineStatesPausedState(self, channel):
        return DAQmxResetDOLineStatesPausedState(self.task_handle, channel)

    def GetDOLineStatesDoneState(self, channel, data):
        return DAQmxGetDOLineStatesDoneState(self.task_handle, channel, data)

    def SetDOLineStatesDoneState(self, channel, data):
        return DAQmxSetDOLineStatesDoneState(self.task_handle, channel, data)

    def ResetDOLineStatesDoneState(self, channel):
        return DAQmxResetDOLineStatesDoneState(self.task_handle, channel)

    def GetDOLogicFamily(self, channel, data):
        return DAQmxGetDOLogicFamily(self.task_handle, channel, data)

    def SetDOLogicFamily(self, channel, data):
        return DAQmxSetDOLogicFamily(self.task_handle, channel, data)

    def ResetDOLogicFamily(self, channel):
        return DAQmxResetDOLogicFamily(self.task_handle, channel)

    def GetDOOvercurrentLimit(self, channel, data):
        return DAQmxGetDOOvercurrentLimit(self.task_handle, channel, data)

    def SetDOOvercurrentLimit(self, channel, data):
        return DAQmxSetDOOvercurrentLimit(self.task_handle, channel, data)

    def ResetDOOvercurrentLimit(self, channel):
        return DAQmxResetDOOvercurrentLimit(self.task_handle, channel)

    def GetDOOvercurrentAutoReenable(self, channel, data):
        return DAQmxGetDOOvercurrentAutoReenable(self.task_handle, channel, data)

    def SetDOOvercurrentAutoReenable(self, channel, data):
        return DAQmxSetDOOvercurrentAutoReenable(self.task_handle, channel, data)

    def ResetDOOvercurrentAutoReenable(self, channel):
        return DAQmxResetDOOvercurrentAutoReenable(self.task_handle, channel)

    def GetDOOvercurrentReenablePeriod(self, channel, data):
        return DAQmxGetDOOvercurrentReenablePeriod(self.task_handle, channel, data)

    def SetDOOvercurrentReenablePeriod(self, channel, data):
        return DAQmxSetDOOvercurrentReenablePeriod(self.task_handle, channel, data)

    def ResetDOOvercurrentReenablePeriod(self, channel):
        return DAQmxResetDOOvercurrentReenablePeriod(self.task_handle, channel)

    def GetDOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxGetDOUseOnlyOnBrdMem(self.task_handle, channel, data)

    def SetDOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxSetDOUseOnlyOnBrdMem(self.task_handle, channel, data)

    def ResetDOUseOnlyOnBrdMem(self, channel):
        return DAQmxResetDOUseOnlyOnBrdMem(self.task_handle, channel)

    def GetDODataXferMech(self, channel, data):
        return DAQmxGetDODataXferMech(self.task_handle, channel, data)

    def SetDODataXferMech(self, channel, data):
        return DAQmxSetDODataXferMech(self.task_handle, channel, data)

    def ResetDODataXferMech(self, channel):
        return DAQmxResetDODataXferMech(self.task_handle, channel)

    def GetDODataXferReqCond(self, channel, data):
        return DAQmxGetDODataXferReqCond(self.task_handle, channel, data)

    def SetDODataXferReqCond(self, channel, data):
        return DAQmxSetDODataXferReqCond(self.task_handle, channel, data)

    def ResetDODataXferReqCond(self, channel):
        return DAQmxResetDODataXferReqCond(self.task_handle, channel)

    def GetDOUsbXferReqSize(self, channel, data):
        return DAQmxGetDOUsbXferReqSize(self.task_handle, channel, data)

    def SetDOUsbXferReqSize(self, channel, data):
        return DAQmxSetDOUsbXferReqSize(self.task_handle, channel, data)

    def ResetDOUsbXferReqSize(self, channel):
        return DAQmxResetDOUsbXferReqSize(self.task_handle, channel)

    def GetDOUsbXferReqCount(self, channel, data):
        return DAQmxGetDOUsbXferReqCount(self.task_handle, channel, data)

    def SetDOUsbXferReqCount(self, channel, data):
        return DAQmxSetDOUsbXferReqCount(self.task_handle, channel, data)

    def ResetDOUsbXferReqCount(self, channel):
        return DAQmxResetDOUsbXferReqCount(self.task_handle, channel)

    def GetDOMemMapEnable(self, channel, data):
        return DAQmxGetDOMemMapEnable(self.task_handle, channel, data)

    def SetDOMemMapEnable(self, channel, data):
        return DAQmxSetDOMemMapEnable(self.task_handle, channel, data)

    def ResetDOMemMapEnable(self, channel):
        return DAQmxResetDOMemMapEnable(self.task_handle, channel)

    def GetDOGenerateOn(self, channel, data):
        return DAQmxGetDOGenerateOn(self.task_handle, channel, data)

    def SetDOGenerateOn(self, channel, data):
        return DAQmxSetDOGenerateOn(self.task_handle, channel, data)

    def ResetDOGenerateOn(self, channel):
        return DAQmxResetDOGenerateOn(self.task_handle, channel)

    def GetCIMax(self, channel, data):
        return DAQmxGetCIMax(self.task_handle, channel, data)

    def SetCIMax(self, channel, data):
        return DAQmxSetCIMax(self.task_handle, channel, data)

    def ResetCIMax(self, channel):
        return DAQmxResetCIMax(self.task_handle, channel)

    def GetCIMin(self, channel, data):
        return DAQmxGetCIMin(self.task_handle, channel, data)

    def SetCIMin(self, channel, data):
        return DAQmxSetCIMin(self.task_handle, channel, data)

    def ResetCIMin(self, channel):
        return DAQmxResetCIMin(self.task_handle, channel)

    def GetCICustomScaleName(self, channel, data, bufferSize):
        return DAQmxGetCICustomScaleName(self.task_handle, channel, data, bufferSize)

    def SetCICustomScaleName(self, channel, data):
        return DAQmxSetCICustomScaleName(self.task_handle, channel, data)

    def ResetCICustomScaleName(self, channel):
        return DAQmxResetCICustomScaleName(self.task_handle, channel)

    def GetCIMeasType(self, channel, data):
        return DAQmxGetCIMeasType(self.task_handle, channel, data)

    def GetCIFreqUnits(self, channel, data):
        return DAQmxGetCIFreqUnits(self.task_handle, channel, data)

    def SetCIFreqUnits(self, channel, data):
        return DAQmxSetCIFreqUnits(self.task_handle, channel, data)

    def ResetCIFreqUnits(self, channel):
        return DAQmxResetCIFreqUnits(self.task_handle, channel)

    def GetCIFreqTerm(self, channel, data, bufferSize):
        return DAQmxGetCIFreqTerm(self.task_handle, channel, data, bufferSize)

    def SetCIFreqTerm(self, channel, data):
        return DAQmxSetCIFreqTerm(self.task_handle, channel, data)

    def ResetCIFreqTerm(self, channel):
        return DAQmxResetCIFreqTerm(self.task_handle, channel)

    def GetCIFreqStartingEdge(self, channel, data):
        return DAQmxGetCIFreqStartingEdge(self.task_handle, channel, data)

    def SetCIFreqStartingEdge(self, channel, data):
        return DAQmxSetCIFreqStartingEdge(self.task_handle, channel, data)

    def ResetCIFreqStartingEdge(self, channel):
        return DAQmxResetCIFreqStartingEdge(self.task_handle, channel)

    def GetCIFreqMeasMeth(self, channel, data):
        return DAQmxGetCIFreqMeasMeth(self.task_handle, channel, data)

    def SetCIFreqMeasMeth(self, channel, data):
        return DAQmxSetCIFreqMeasMeth(self.task_handle, channel, data)

    def ResetCIFreqMeasMeth(self, channel):
        return DAQmxResetCIFreqMeasMeth(self.task_handle, channel)

    def GetCIFreqEnableAveraging(self, channel, data):
        return DAQmxGetCIFreqEnableAveraging(self.task_handle, channel, data)

    def SetCIFreqEnableAveraging(self, channel, data):
        return DAQmxSetCIFreqEnableAveraging(self.task_handle, channel, data)

    def ResetCIFreqEnableAveraging(self, channel):
        return DAQmxResetCIFreqEnableAveraging(self.task_handle, channel)

    def GetCIFreqMeasTime(self, channel, data):
        return DAQmxGetCIFreqMeasTime(self.task_handle, channel, data)

    def SetCIFreqMeasTime(self, channel, data):
        return DAQmxSetCIFreqMeasTime(self.task_handle, channel, data)

    def ResetCIFreqMeasTime(self, channel):
        return DAQmxResetCIFreqMeasTime(self.task_handle, channel)

    def GetCIFreqDiv(self, channel, data):
        return DAQmxGetCIFreqDiv(self.task_handle, channel, data)

    def SetCIFreqDiv(self, channel, data):
        return DAQmxSetCIFreqDiv(self.task_handle, channel, data)

    def ResetCIFreqDiv(self, channel):
        return DAQmxResetCIFreqDiv(self.task_handle, channel)

    def GetCIFreqDigFltrEnable(self, channel, data):
        return DAQmxGetCIFreqDigFltrEnable(self.task_handle, channel, data)

    def SetCIFreqDigFltrEnable(self, channel, data):
        return DAQmxSetCIFreqDigFltrEnable(self.task_handle, channel, data)

    def ResetCIFreqDigFltrEnable(self, channel):
        return DAQmxResetCIFreqDigFltrEnable(self.task_handle, channel)

    def GetCIFreqDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIFreqDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIFreqDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIFreqDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIFreqDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIFreqDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIFreqDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIFreqDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIFreqDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIFreqDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIFreqDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIFreqDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIFreqDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIFreqDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIFreqDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIFreqDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIFreqDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIFreqDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIFreqDigSyncEnable(self, channel, data):
        return DAQmxGetCIFreqDigSyncEnable(self.task_handle, channel, data)

    def SetCIFreqDigSyncEnable(self, channel, data):
        return DAQmxSetCIFreqDigSyncEnable(self.task_handle, channel, data)

    def ResetCIFreqDigSyncEnable(self, channel):
        return DAQmxResetCIFreqDigSyncEnable(self.task_handle, channel)

    def GetCIPeriodUnits(self, channel, data):
        return DAQmxGetCIPeriodUnits(self.task_handle, channel, data)

    def SetCIPeriodUnits(self, channel, data):
        return DAQmxSetCIPeriodUnits(self.task_handle, channel, data)

    def ResetCIPeriodUnits(self, channel):
        return DAQmxResetCIPeriodUnits(self.task_handle, channel)

    def GetCIPeriodTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPeriodTerm(self.task_handle, channel, data, bufferSize)

    def SetCIPeriodTerm(self, channel, data):
        return DAQmxSetCIPeriodTerm(self.task_handle, channel, data)

    def ResetCIPeriodTerm(self, channel):
        return DAQmxResetCIPeriodTerm(self.task_handle, channel)

    def GetCIPeriodStartingEdge(self, channel, data):
        return DAQmxGetCIPeriodStartingEdge(self.task_handle, channel, data)

    def SetCIPeriodStartingEdge(self, channel, data):
        return DAQmxSetCIPeriodStartingEdge(self.task_handle, channel, data)

    def ResetCIPeriodStartingEdge(self, channel):
        return DAQmxResetCIPeriodStartingEdge(self.task_handle, channel)

    def GetCIPeriodMeasMeth(self, channel, data):
        return DAQmxGetCIPeriodMeasMeth(self.task_handle, channel, data)

    def SetCIPeriodMeasMeth(self, channel, data):
        return DAQmxSetCIPeriodMeasMeth(self.task_handle, channel, data)

    def ResetCIPeriodMeasMeth(self, channel):
        return DAQmxResetCIPeriodMeasMeth(self.task_handle, channel)

    def GetCIPeriodEnableAveraging(self, channel, data):
        return DAQmxGetCIPeriodEnableAveraging(self.task_handle, channel, data)

    def SetCIPeriodEnableAveraging(self, channel, data):
        return DAQmxSetCIPeriodEnableAveraging(self.task_handle, channel, data)

    def ResetCIPeriodEnableAveraging(self, channel):
        return DAQmxResetCIPeriodEnableAveraging(self.task_handle, channel)

    def GetCIPeriodMeasTime(self, channel, data):
        return DAQmxGetCIPeriodMeasTime(self.task_handle, channel, data)

    def SetCIPeriodMeasTime(self, channel, data):
        return DAQmxSetCIPeriodMeasTime(self.task_handle, channel, data)

    def ResetCIPeriodMeasTime(self, channel):
        return DAQmxResetCIPeriodMeasTime(self.task_handle, channel)

    def GetCIPeriodDiv(self, channel, data):
        return DAQmxGetCIPeriodDiv(self.task_handle, channel, data)

    def SetCIPeriodDiv(self, channel, data):
        return DAQmxSetCIPeriodDiv(self.task_handle, channel, data)

    def ResetCIPeriodDiv(self, channel):
        return DAQmxResetCIPeriodDiv(self.task_handle, channel)

    def GetCIPeriodDigFltrEnable(self, channel, data):
        return DAQmxGetCIPeriodDigFltrEnable(self.task_handle, channel, data)

    def SetCIPeriodDigFltrEnable(self, channel, data):
        return DAQmxSetCIPeriodDigFltrEnable(self.task_handle, channel, data)

    def ResetCIPeriodDigFltrEnable(self, channel):
        return DAQmxResetCIPeriodDigFltrEnable(self.task_handle, channel)

    def GetCIPeriodDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPeriodDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIPeriodDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPeriodDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIPeriodDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPeriodDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIPeriodDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPeriodDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIPeriodDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPeriodDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIPeriodDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPeriodDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIPeriodDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPeriodDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIPeriodDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPeriodDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIPeriodDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPeriodDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIPeriodDigSyncEnable(self, channel, data):
        return DAQmxGetCIPeriodDigSyncEnable(self.task_handle, channel, data)

    def SetCIPeriodDigSyncEnable(self, channel, data):
        return DAQmxSetCIPeriodDigSyncEnable(self.task_handle, channel, data)

    def ResetCIPeriodDigSyncEnable(self, channel):
        return DAQmxResetCIPeriodDigSyncEnable(self.task_handle, channel)

    def GetCICountEdgesTerm(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesTerm(self.task_handle, channel, data, bufferSize)

    def SetCICountEdgesTerm(self, channel, data):
        return DAQmxSetCICountEdgesTerm(self.task_handle, channel, data)

    def ResetCICountEdgesTerm(self, channel):
        return DAQmxResetCICountEdgesTerm(self.task_handle, channel)

    def GetCICountEdgesDir(self, channel, data):
        return DAQmxGetCICountEdgesDir(self.task_handle, channel, data)

    def SetCICountEdgesDir(self, channel, data):
        return DAQmxSetCICountEdgesDir(self.task_handle, channel, data)

    def ResetCICountEdgesDir(self, channel):
        return DAQmxResetCICountEdgesDir(self.task_handle, channel)

    def GetCICountEdgesDirTerm(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesDirTerm(self.task_handle, channel, data, bufferSize)

    def SetCICountEdgesDirTerm(self, channel, data):
        return DAQmxSetCICountEdgesDirTerm(self.task_handle, channel, data)

    def ResetCICountEdgesDirTerm(self, channel):
        return DAQmxResetCICountEdgesDirTerm(self.task_handle, channel)

    def GetCICountEdgesCountDirDigFltrEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountDirDigFltrEnable(self.task_handle, channel, data)

    def SetCICountEdgesCountDirDigFltrEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigFltrEnable(self.task_handle, channel, data)

    def ResetCICountEdgesCountDirDigFltrEnable(self, channel):
        return DAQmxResetCICountEdgesCountDirDigFltrEnable(self.task_handle, channel)

    def GetCICountEdgesCountDirDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCICountEdgesCountDirDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCICountEdgesCountDirDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCICountEdgesCountDirDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCICountEdgesCountDirDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCICountEdgesCountDirDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesCountDirDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCICountEdgesCountDirDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCICountEdgesCountDirDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCICountEdgesCountDirDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCICountEdgesCountDirDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCICountEdgesCountDirDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCICountEdgesCountDirDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCICountEdgesCountDirDigFltrTimebaseRate(self, channel):
        return DAQmxResetCICountEdgesCountDirDigFltrTimebaseRate(self.task_handle, channel)

    def GetCICountEdgesCountDirDigSyncEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountDirDigSyncEnable(self.task_handle, channel, data)

    def SetCICountEdgesCountDirDigSyncEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountDirDigSyncEnable(self.task_handle, channel, data)

    def ResetCICountEdgesCountDirDigSyncEnable(self, channel):
        return DAQmxResetCICountEdgesCountDirDigSyncEnable(self.task_handle, channel)

    def GetCICountEdgesInitialCnt(self, channel, data):
        return DAQmxGetCICountEdgesInitialCnt(self.task_handle, channel, data)

    def SetCICountEdgesInitialCnt(self, channel, data):
        return DAQmxSetCICountEdgesInitialCnt(self.task_handle, channel, data)

    def ResetCICountEdgesInitialCnt(self, channel):
        return DAQmxResetCICountEdgesInitialCnt(self.task_handle, channel)

    def GetCICountEdgesActiveEdge(self, channel, data):
        return DAQmxGetCICountEdgesActiveEdge(self.task_handle, channel, data)

    def SetCICountEdgesActiveEdge(self, channel, data):
        return DAQmxSetCICountEdgesActiveEdge(self.task_handle, channel, data)

    def ResetCICountEdgesActiveEdge(self, channel):
        return DAQmxResetCICountEdgesActiveEdge(self.task_handle, channel)

    def GetCICountEdgesCountResetEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountResetEnable(self.task_handle, channel, data)

    def SetCICountEdgesCountResetEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountResetEnable(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetEnable(self, channel):
        return DAQmxResetCICountEdgesCountResetEnable(self.task_handle, channel)

    def GetCICountEdgesCountResetResetCount(self, channel, data):
        return DAQmxGetCICountEdgesCountResetResetCount(self.task_handle, channel, data)

    def SetCICountEdgesCountResetResetCount(self, channel, data):
        return DAQmxSetCICountEdgesCountResetResetCount(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetResetCount(self, channel):
        return DAQmxResetCICountEdgesCountResetResetCount(self.task_handle, channel)

    def GetCICountEdgesCountResetTerm(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesCountResetTerm(self.task_handle, channel, data, bufferSize)

    def SetCICountEdgesCountResetTerm(self, channel, data):
        return DAQmxSetCICountEdgesCountResetTerm(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetTerm(self, channel):
        return DAQmxResetCICountEdgesCountResetTerm(self.task_handle, channel)

    def GetCICountEdgesCountResetActiveEdge(self, channel, data):
        return DAQmxGetCICountEdgesCountResetActiveEdge(self.task_handle, channel, data)

    def SetCICountEdgesCountResetActiveEdge(self, channel, data):
        return DAQmxSetCICountEdgesCountResetActiveEdge(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetActiveEdge(self, channel):
        return DAQmxResetCICountEdgesCountResetActiveEdge(self.task_handle, channel)

    def GetCICountEdgesCountResetDigFltrEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountResetDigFltrEnable(self.task_handle, channel, data)

    def SetCICountEdgesCountResetDigFltrEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigFltrEnable(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetDigFltrEnable(self, channel):
        return DAQmxResetCICountEdgesCountResetDigFltrEnable(self.task_handle, channel)

    def GetCICountEdgesCountResetDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCICountEdgesCountResetDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCICountEdgesCountResetDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCICountEdgesCountResetDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCICountEdgesCountResetDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesCountResetDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCICountEdgesCountResetDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCICountEdgesCountResetDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCICountEdgesCountResetDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCICountEdgesCountResetDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCICountEdgesCountResetDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetDigFltrTimebaseRate(self, channel):
        return DAQmxResetCICountEdgesCountResetDigFltrTimebaseRate(self.task_handle, channel)

    def GetCICountEdgesCountResetDigSyncEnable(self, channel, data):
        return DAQmxGetCICountEdgesCountResetDigSyncEnable(self.task_handle, channel, data)

    def SetCICountEdgesCountResetDigSyncEnable(self, channel, data):
        return DAQmxSetCICountEdgesCountResetDigSyncEnable(self.task_handle, channel, data)

    def ResetCICountEdgesCountResetDigSyncEnable(self, channel):
        return DAQmxResetCICountEdgesCountResetDigSyncEnable(self.task_handle, channel)

    def GetCICountEdgesDigFltrEnable(self, channel, data):
        return DAQmxGetCICountEdgesDigFltrEnable(self.task_handle, channel, data)

    def SetCICountEdgesDigFltrEnable(self, channel, data):
        return DAQmxSetCICountEdgesDigFltrEnable(self.task_handle, channel, data)

    def ResetCICountEdgesDigFltrEnable(self, channel):
        return DAQmxResetCICountEdgesDigFltrEnable(self.task_handle, channel)

    def GetCICountEdgesDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCICountEdgesDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCICountEdgesDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCICountEdgesDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCICountEdgesDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCICountEdgesDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCICountEdgesDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICountEdgesDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCICountEdgesDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCICountEdgesDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCICountEdgesDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCICountEdgesDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCICountEdgesDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCICountEdgesDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCICountEdgesDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCICountEdgesDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCICountEdgesDigFltrTimebaseRate(self, channel):
        return DAQmxResetCICountEdgesDigFltrTimebaseRate(self.task_handle, channel)

    def GetCICountEdgesDigSyncEnable(self, channel, data):
        return DAQmxGetCICountEdgesDigSyncEnable(self.task_handle, channel, data)

    def SetCICountEdgesDigSyncEnable(self, channel, data):
        return DAQmxSetCICountEdgesDigSyncEnable(self.task_handle, channel, data)

    def ResetCICountEdgesDigSyncEnable(self, channel):
        return DAQmxResetCICountEdgesDigSyncEnable(self.task_handle, channel)

    def GetCIAngEncoderUnits(self, channel, data):
        return DAQmxGetCIAngEncoderUnits(self.task_handle, channel, data)

    def SetCIAngEncoderUnits(self, channel, data):
        return DAQmxSetCIAngEncoderUnits(self.task_handle, channel, data)

    def ResetCIAngEncoderUnits(self, channel):
        return DAQmxResetCIAngEncoderUnits(self.task_handle, channel)

    def GetCIAngEncoderPulsesPerRev(self, channel, data):
        return DAQmxGetCIAngEncoderPulsesPerRev(self.task_handle, channel, data)

    def SetCIAngEncoderPulsesPerRev(self, channel, data):
        return DAQmxSetCIAngEncoderPulsesPerRev(self.task_handle, channel, data)

    def ResetCIAngEncoderPulsesPerRev(self, channel):
        return DAQmxResetCIAngEncoderPulsesPerRev(self.task_handle, channel)

    def GetCIAngEncoderInitialAngle(self, channel, data):
        return DAQmxGetCIAngEncoderInitialAngle(self.task_handle, channel, data)

    def SetCIAngEncoderInitialAngle(self, channel, data):
        return DAQmxSetCIAngEncoderInitialAngle(self.task_handle, channel, data)

    def ResetCIAngEncoderInitialAngle(self, channel):
        return DAQmxResetCIAngEncoderInitialAngle(self.task_handle, channel)

    def GetCILinEncoderUnits(self, channel, data):
        return DAQmxGetCILinEncoderUnits(self.task_handle, channel, data)

    def SetCILinEncoderUnits(self, channel, data):
        return DAQmxSetCILinEncoderUnits(self.task_handle, channel, data)

    def ResetCILinEncoderUnits(self, channel):
        return DAQmxResetCILinEncoderUnits(self.task_handle, channel)

    def GetCILinEncoderDistPerPulse(self, channel, data):
        return DAQmxGetCILinEncoderDistPerPulse(self.task_handle, channel, data)

    def SetCILinEncoderDistPerPulse(self, channel, data):
        return DAQmxSetCILinEncoderDistPerPulse(self.task_handle, channel, data)

    def ResetCILinEncoderDistPerPulse(self, channel):
        return DAQmxResetCILinEncoderDistPerPulse(self.task_handle, channel)

    def GetCILinEncoderInitialPos(self, channel, data):
        return DAQmxGetCILinEncoderInitialPos(self.task_handle, channel, data)

    def SetCILinEncoderInitialPos(self, channel, data):
        return DAQmxSetCILinEncoderInitialPos(self.task_handle, channel, data)

    def ResetCILinEncoderInitialPos(self, channel):
        return DAQmxResetCILinEncoderInitialPos(self.task_handle, channel)

    def GetCIEncoderDecodingType(self, channel, data):
        return DAQmxGetCIEncoderDecodingType(self.task_handle, channel, data)

    def SetCIEncoderDecodingType(self, channel, data):
        return DAQmxSetCIEncoderDecodingType(self.task_handle, channel, data)

    def ResetCIEncoderDecodingType(self, channel):
        return DAQmxResetCIEncoderDecodingType(self.task_handle, channel)

    def GetCIEncoderAInputTerm(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderAInputTerm(self.task_handle, channel, data, bufferSize)

    def SetCIEncoderAInputTerm(self, channel, data):
        return DAQmxSetCIEncoderAInputTerm(self.task_handle, channel, data)

    def ResetCIEncoderAInputTerm(self, channel):
        return DAQmxResetCIEncoderAInputTerm(self.task_handle, channel)

    def GetCIEncoderAInputDigFltrEnable(self, channel, data):
        return DAQmxGetCIEncoderAInputDigFltrEnable(self.task_handle, channel, data)

    def SetCIEncoderAInputDigFltrEnable(self, channel, data):
        return DAQmxSetCIEncoderAInputDigFltrEnable(self.task_handle, channel, data)

    def ResetCIEncoderAInputDigFltrEnable(self, channel):
        return DAQmxResetCIEncoderAInputDigFltrEnable(self.task_handle, channel)

    def GetCIEncoderAInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIEncoderAInputDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIEncoderAInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIEncoderAInputDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIEncoderAInputDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIEncoderAInputDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIEncoderAInputDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderAInputDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIEncoderAInputDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIEncoderAInputDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIEncoderAInputDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIEncoderAInputDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIEncoderAInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIEncoderAInputDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIEncoderAInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIEncoderAInputDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIEncoderAInputDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIEncoderAInputDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIEncoderAInputDigSyncEnable(self, channel, data):
        return DAQmxGetCIEncoderAInputDigSyncEnable(self.task_handle, channel, data)

    def SetCIEncoderAInputDigSyncEnable(self, channel, data):
        return DAQmxSetCIEncoderAInputDigSyncEnable(self.task_handle, channel, data)

    def ResetCIEncoderAInputDigSyncEnable(self, channel):
        return DAQmxResetCIEncoderAInputDigSyncEnable(self.task_handle, channel)

    def GetCIEncoderBInputTerm(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderBInputTerm(self.task_handle, channel, data, bufferSize)

    def SetCIEncoderBInputTerm(self, channel, data):
        return DAQmxSetCIEncoderBInputTerm(self.task_handle, channel, data)

    def ResetCIEncoderBInputTerm(self, channel):
        return DAQmxResetCIEncoderBInputTerm(self.task_handle, channel)

    def GetCIEncoderBInputDigFltrEnable(self, channel, data):
        return DAQmxGetCIEncoderBInputDigFltrEnable(self.task_handle, channel, data)

    def SetCIEncoderBInputDigFltrEnable(self, channel, data):
        return DAQmxSetCIEncoderBInputDigFltrEnable(self.task_handle, channel, data)

    def ResetCIEncoderBInputDigFltrEnable(self, channel):
        return DAQmxResetCIEncoderBInputDigFltrEnable(self.task_handle, channel)

    def GetCIEncoderBInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIEncoderBInputDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIEncoderBInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIEncoderBInputDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIEncoderBInputDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIEncoderBInputDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIEncoderBInputDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderBInputDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIEncoderBInputDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIEncoderBInputDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIEncoderBInputDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIEncoderBInputDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIEncoderBInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIEncoderBInputDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIEncoderBInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIEncoderBInputDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIEncoderBInputDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIEncoderBInputDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIEncoderBInputDigSyncEnable(self, channel, data):
        return DAQmxGetCIEncoderBInputDigSyncEnable(self.task_handle, channel, data)

    def SetCIEncoderBInputDigSyncEnable(self, channel, data):
        return DAQmxSetCIEncoderBInputDigSyncEnable(self.task_handle, channel, data)

    def ResetCIEncoderBInputDigSyncEnable(self, channel):
        return DAQmxResetCIEncoderBInputDigSyncEnable(self.task_handle, channel)

    def GetCIEncoderZInputTerm(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderZInputTerm(self.task_handle, channel, data, bufferSize)

    def SetCIEncoderZInputTerm(self, channel, data):
        return DAQmxSetCIEncoderZInputTerm(self.task_handle, channel, data)

    def ResetCIEncoderZInputTerm(self, channel):
        return DAQmxResetCIEncoderZInputTerm(self.task_handle, channel)

    def GetCIEncoderZInputDigFltrEnable(self, channel, data):
        return DAQmxGetCIEncoderZInputDigFltrEnable(self.task_handle, channel, data)

    def SetCIEncoderZInputDigFltrEnable(self, channel, data):
        return DAQmxSetCIEncoderZInputDigFltrEnable(self.task_handle, channel, data)

    def ResetCIEncoderZInputDigFltrEnable(self, channel):
        return DAQmxResetCIEncoderZInputDigFltrEnable(self.task_handle, channel)

    def GetCIEncoderZInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIEncoderZInputDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIEncoderZInputDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIEncoderZInputDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIEncoderZInputDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIEncoderZInputDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIEncoderZInputDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIEncoderZInputDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIEncoderZInputDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIEncoderZInputDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIEncoderZInputDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIEncoderZInputDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIEncoderZInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIEncoderZInputDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIEncoderZInputDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIEncoderZInputDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIEncoderZInputDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIEncoderZInputDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIEncoderZInputDigSyncEnable(self, channel, data):
        return DAQmxGetCIEncoderZInputDigSyncEnable(self.task_handle, channel, data)

    def SetCIEncoderZInputDigSyncEnable(self, channel, data):
        return DAQmxSetCIEncoderZInputDigSyncEnable(self.task_handle, channel, data)

    def ResetCIEncoderZInputDigSyncEnable(self, channel):
        return DAQmxResetCIEncoderZInputDigSyncEnable(self.task_handle, channel)

    def GetCIEncoderZIndexEnable(self, channel, data):
        return DAQmxGetCIEncoderZIndexEnable(self.task_handle, channel, data)

    def SetCIEncoderZIndexEnable(self, channel, data):
        return DAQmxSetCIEncoderZIndexEnable(self.task_handle, channel, data)

    def ResetCIEncoderZIndexEnable(self, channel):
        return DAQmxResetCIEncoderZIndexEnable(self.task_handle, channel)

    def GetCIEncoderZIndexVal(self, channel, data):
        return DAQmxGetCIEncoderZIndexVal(self.task_handle, channel, data)

    def SetCIEncoderZIndexVal(self, channel, data):
        return DAQmxSetCIEncoderZIndexVal(self.task_handle, channel, data)

    def ResetCIEncoderZIndexVal(self, channel):
        return DAQmxResetCIEncoderZIndexVal(self.task_handle, channel)

    def GetCIEncoderZIndexPhase(self, channel, data):
        return DAQmxGetCIEncoderZIndexPhase(self.task_handle, channel, data)

    def SetCIEncoderZIndexPhase(self, channel, data):
        return DAQmxSetCIEncoderZIndexPhase(self.task_handle, channel, data)

    def ResetCIEncoderZIndexPhase(self, channel):
        return DAQmxResetCIEncoderZIndexPhase(self.task_handle, channel)

    def GetCIPulseWidthUnits(self, channel, data):
        return DAQmxGetCIPulseWidthUnits(self.task_handle, channel, data)

    def SetCIPulseWidthUnits(self, channel, data):
        return DAQmxSetCIPulseWidthUnits(self.task_handle, channel, data)

    def ResetCIPulseWidthUnits(self, channel):
        return DAQmxResetCIPulseWidthUnits(self.task_handle, channel)

    def GetCIPulseWidthTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPulseWidthTerm(self.task_handle, channel, data, bufferSize)

    def SetCIPulseWidthTerm(self, channel, data):
        return DAQmxSetCIPulseWidthTerm(self.task_handle, channel, data)

    def ResetCIPulseWidthTerm(self, channel):
        return DAQmxResetCIPulseWidthTerm(self.task_handle, channel)

    def GetCIPulseWidthStartingEdge(self, channel, data):
        return DAQmxGetCIPulseWidthStartingEdge(self.task_handle, channel, data)

    def SetCIPulseWidthStartingEdge(self, channel, data):
        return DAQmxSetCIPulseWidthStartingEdge(self.task_handle, channel, data)

    def ResetCIPulseWidthStartingEdge(self, channel):
        return DAQmxResetCIPulseWidthStartingEdge(self.task_handle, channel)

    def GetCIPulseWidthDigFltrEnable(self, channel, data):
        return DAQmxGetCIPulseWidthDigFltrEnable(self.task_handle, channel, data)

    def SetCIPulseWidthDigFltrEnable(self, channel, data):
        return DAQmxSetCIPulseWidthDigFltrEnable(self.task_handle, channel, data)

    def ResetCIPulseWidthDigFltrEnable(self, channel):
        return DAQmxResetCIPulseWidthDigFltrEnable(self.task_handle, channel)

    def GetCIPulseWidthDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPulseWidthDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIPulseWidthDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPulseWidthDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIPulseWidthDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPulseWidthDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIPulseWidthDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPulseWidthDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIPulseWidthDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPulseWidthDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIPulseWidthDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPulseWidthDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIPulseWidthDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPulseWidthDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIPulseWidthDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPulseWidthDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIPulseWidthDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPulseWidthDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIPulseWidthDigSyncEnable(self, channel, data):
        return DAQmxGetCIPulseWidthDigSyncEnable(self.task_handle, channel, data)

    def SetCIPulseWidthDigSyncEnable(self, channel, data):
        return DAQmxSetCIPulseWidthDigSyncEnable(self.task_handle, channel, data)

    def ResetCIPulseWidthDigSyncEnable(self, channel):
        return DAQmxResetCIPulseWidthDigSyncEnable(self.task_handle, channel)

    def GetCITimestampUnits(self, channel, data):
        return DAQmxGetCITimestampUnits(self.task_handle, channel, data)

    def SetCITimestampUnits(self, channel, data):
        return DAQmxSetCITimestampUnits(self.task_handle, channel, data)

    def ResetCITimestampUnits(self, channel):
        return DAQmxResetCITimestampUnits(self.task_handle, channel)

    def GetCITwoEdgeSepUnits(self, channel, data):
        return DAQmxGetCITwoEdgeSepUnits(self.task_handle, channel, data)

    def SetCITwoEdgeSepUnits(self, channel, data):
        return DAQmxSetCITwoEdgeSepUnits(self.task_handle, channel, data)

    def ResetCITwoEdgeSepUnits(self, channel):
        return DAQmxResetCITwoEdgeSepUnits(self.task_handle, channel)

    def GetCITwoEdgeSepFirstTerm(self, channel, data, bufferSize):
        return DAQmxGetCITwoEdgeSepFirstTerm(self.task_handle, channel, data, bufferSize)

    def SetCITwoEdgeSepFirstTerm(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstTerm(self.task_handle, channel, data)

    def ResetCITwoEdgeSepFirstTerm(self, channel):
        return DAQmxResetCITwoEdgeSepFirstTerm(self.task_handle, channel)

    def GetCITwoEdgeSepFirstEdge(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstEdge(self.task_handle, channel, data)

    def SetCITwoEdgeSepFirstEdge(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstEdge(self.task_handle, channel, data)

    def ResetCITwoEdgeSepFirstEdge(self, channel):
        return DAQmxResetCITwoEdgeSepFirstEdge(self.task_handle, channel)

    def GetCITwoEdgeSepFirstDigFltrEnable(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstDigFltrEnable(self.task_handle, channel, data)

    def SetCITwoEdgeSepFirstDigFltrEnable(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigFltrEnable(self.task_handle, channel, data)

    def ResetCITwoEdgeSepFirstDigFltrEnable(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigFltrEnable(self.task_handle, channel)

    def GetCITwoEdgeSepFirstDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCITwoEdgeSepFirstDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCITwoEdgeSepFirstDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCITwoEdgeSepFirstDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCITwoEdgeSepFirstDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCITwoEdgeSepFirstDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCITwoEdgeSepFirstDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCITwoEdgeSepFirstDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCITwoEdgeSepFirstDigFltrTimebaseRate(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigFltrTimebaseRate(self.task_handle, channel)

    def GetCITwoEdgeSepFirstDigSyncEnable(self, channel, data):
        return DAQmxGetCITwoEdgeSepFirstDigSyncEnable(self.task_handle, channel, data)

    def SetCITwoEdgeSepFirstDigSyncEnable(self, channel, data):
        return DAQmxSetCITwoEdgeSepFirstDigSyncEnable(self.task_handle, channel, data)

    def ResetCITwoEdgeSepFirstDigSyncEnable(self, channel):
        return DAQmxResetCITwoEdgeSepFirstDigSyncEnable(self.task_handle, channel)

    def GetCITwoEdgeSepSecondTerm(self, channel, data, bufferSize):
        return DAQmxGetCITwoEdgeSepSecondTerm(self.task_handle, channel, data, bufferSize)

    def SetCITwoEdgeSepSecondTerm(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondTerm(self.task_handle, channel, data)

    def ResetCITwoEdgeSepSecondTerm(self, channel):
        return DAQmxResetCITwoEdgeSepSecondTerm(self.task_handle, channel)

    def GetCITwoEdgeSepSecondEdge(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondEdge(self.task_handle, channel, data)

    def SetCITwoEdgeSepSecondEdge(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondEdge(self.task_handle, channel, data)

    def ResetCITwoEdgeSepSecondEdge(self, channel):
        return DAQmxResetCITwoEdgeSepSecondEdge(self.task_handle, channel)

    def GetCITwoEdgeSepSecondDigFltrEnable(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondDigFltrEnable(self.task_handle, channel, data)

    def SetCITwoEdgeSepSecondDigFltrEnable(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigFltrEnable(self.task_handle, channel, data)

    def ResetCITwoEdgeSepSecondDigFltrEnable(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigFltrEnable(self.task_handle, channel)

    def GetCITwoEdgeSepSecondDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCITwoEdgeSepSecondDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCITwoEdgeSepSecondDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCITwoEdgeSepSecondDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCITwoEdgeSepSecondDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCITwoEdgeSepSecondDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCITwoEdgeSepSecondDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCITwoEdgeSepSecondDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCITwoEdgeSepSecondDigFltrTimebaseRate(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigFltrTimebaseRate(self.task_handle, channel)

    def GetCITwoEdgeSepSecondDigSyncEnable(self, channel, data):
        return DAQmxGetCITwoEdgeSepSecondDigSyncEnable(self.task_handle, channel, data)

    def SetCITwoEdgeSepSecondDigSyncEnable(self, channel, data):
        return DAQmxSetCITwoEdgeSepSecondDigSyncEnable(self.task_handle, channel, data)

    def ResetCITwoEdgeSepSecondDigSyncEnable(self, channel):
        return DAQmxResetCITwoEdgeSepSecondDigSyncEnable(self.task_handle, channel)

    def GetCISemiPeriodUnits(self, channel, data):
        return DAQmxGetCISemiPeriodUnits(self.task_handle, channel, data)

    def SetCISemiPeriodUnits(self, channel, data):
        return DAQmxSetCISemiPeriodUnits(self.task_handle, channel, data)

    def ResetCISemiPeriodUnits(self, channel):
        return DAQmxResetCISemiPeriodUnits(self.task_handle, channel)

    def GetCISemiPeriodTerm(self, channel, data, bufferSize):
        return DAQmxGetCISemiPeriodTerm(self.task_handle, channel, data, bufferSize)

    def SetCISemiPeriodTerm(self, channel, data):
        return DAQmxSetCISemiPeriodTerm(self.task_handle, channel, data)

    def ResetCISemiPeriodTerm(self, channel):
        return DAQmxResetCISemiPeriodTerm(self.task_handle, channel)

    def GetCISemiPeriodStartingEdge(self, channel, data):
        return DAQmxGetCISemiPeriodStartingEdge(self.task_handle, channel, data)

    def SetCISemiPeriodStartingEdge(self, channel, data):
        return DAQmxSetCISemiPeriodStartingEdge(self.task_handle, channel, data)

    def ResetCISemiPeriodStartingEdge(self, channel):
        return DAQmxResetCISemiPeriodStartingEdge(self.task_handle, channel)

    def GetCISemiPeriodDigFltrEnable(self, channel, data):
        return DAQmxGetCISemiPeriodDigFltrEnable(self.task_handle, channel, data)

    def SetCISemiPeriodDigFltrEnable(self, channel, data):
        return DAQmxSetCISemiPeriodDigFltrEnable(self.task_handle, channel, data)

    def ResetCISemiPeriodDigFltrEnable(self, channel):
        return DAQmxResetCISemiPeriodDigFltrEnable(self.task_handle, channel)

    def GetCISemiPeriodDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCISemiPeriodDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCISemiPeriodDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCISemiPeriodDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCISemiPeriodDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCISemiPeriodDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCISemiPeriodDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCISemiPeriodDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCISemiPeriodDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCISemiPeriodDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCISemiPeriodDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCISemiPeriodDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCISemiPeriodDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCISemiPeriodDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCISemiPeriodDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCISemiPeriodDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCISemiPeriodDigFltrTimebaseRate(self, channel):
        return DAQmxResetCISemiPeriodDigFltrTimebaseRate(self.task_handle, channel)

    def GetCISemiPeriodDigSyncEnable(self, channel, data):
        return DAQmxGetCISemiPeriodDigSyncEnable(self.task_handle, channel, data)

    def SetCISemiPeriodDigSyncEnable(self, channel, data):
        return DAQmxSetCISemiPeriodDigSyncEnable(self.task_handle, channel, data)

    def ResetCISemiPeriodDigSyncEnable(self, channel):
        return DAQmxResetCISemiPeriodDigSyncEnable(self.task_handle, channel)

    def GetCIPulseFreqUnits(self, channel, data):
        return DAQmxGetCIPulseFreqUnits(self.task_handle, channel, data)

    def SetCIPulseFreqUnits(self, channel, data):
        return DAQmxSetCIPulseFreqUnits(self.task_handle, channel, data)

    def ResetCIPulseFreqUnits(self, channel):
        return DAQmxResetCIPulseFreqUnits(self.task_handle, channel)

    def GetCIPulseFreqTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPulseFreqTerm(self.task_handle, channel, data, bufferSize)

    def SetCIPulseFreqTerm(self, channel, data):
        return DAQmxSetCIPulseFreqTerm(self.task_handle, channel, data)

    def ResetCIPulseFreqTerm(self, channel):
        return DAQmxResetCIPulseFreqTerm(self.task_handle, channel)

    def GetCIPulseFreqStartEdge(self, channel, data):
        return DAQmxGetCIPulseFreqStartEdge(self.task_handle, channel, data)

    def SetCIPulseFreqStartEdge(self, channel, data):
        return DAQmxSetCIPulseFreqStartEdge(self.task_handle, channel, data)

    def ResetCIPulseFreqStartEdge(self, channel):
        return DAQmxResetCIPulseFreqStartEdge(self.task_handle, channel)

    def GetCIPulseFreqDigFltrEnable(self, channel, data):
        return DAQmxGetCIPulseFreqDigFltrEnable(self.task_handle, channel, data)

    def SetCIPulseFreqDigFltrEnable(self, channel, data):
        return DAQmxSetCIPulseFreqDigFltrEnable(self.task_handle, channel, data)

    def ResetCIPulseFreqDigFltrEnable(self, channel):
        return DAQmxResetCIPulseFreqDigFltrEnable(self.task_handle, channel)

    def GetCIPulseFreqDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPulseFreqDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIPulseFreqDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPulseFreqDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIPulseFreqDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPulseFreqDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIPulseFreqDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPulseFreqDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIPulseFreqDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPulseFreqDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIPulseFreqDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPulseFreqDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIPulseFreqDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPulseFreqDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIPulseFreqDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPulseFreqDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIPulseFreqDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPulseFreqDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIPulseFreqDigSyncEnable(self, channel, data):
        return DAQmxGetCIPulseFreqDigSyncEnable(self.task_handle, channel, data)

    def SetCIPulseFreqDigSyncEnable(self, channel, data):
        return DAQmxSetCIPulseFreqDigSyncEnable(self.task_handle, channel, data)

    def ResetCIPulseFreqDigSyncEnable(self, channel):
        return DAQmxResetCIPulseFreqDigSyncEnable(self.task_handle, channel)

    def GetCIPulseTimeUnits(self, channel, data):
        return DAQmxGetCIPulseTimeUnits(self.task_handle, channel, data)

    def SetCIPulseTimeUnits(self, channel, data):
        return DAQmxSetCIPulseTimeUnits(self.task_handle, channel, data)

    def ResetCIPulseTimeUnits(self, channel):
        return DAQmxResetCIPulseTimeUnits(self.task_handle, channel)

    def GetCIPulseTimeTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPulseTimeTerm(self.task_handle, channel, data, bufferSize)

    def SetCIPulseTimeTerm(self, channel, data):
        return DAQmxSetCIPulseTimeTerm(self.task_handle, channel, data)

    def ResetCIPulseTimeTerm(self, channel):
        return DAQmxResetCIPulseTimeTerm(self.task_handle, channel)

    def GetCIPulseTimeStartEdge(self, channel, data):
        return DAQmxGetCIPulseTimeStartEdge(self.task_handle, channel, data)

    def SetCIPulseTimeStartEdge(self, channel, data):
        return DAQmxSetCIPulseTimeStartEdge(self.task_handle, channel, data)

    def ResetCIPulseTimeStartEdge(self, channel):
        return DAQmxResetCIPulseTimeStartEdge(self.task_handle, channel)

    def GetCIPulseTimeDigFltrEnable(self, channel, data):
        return DAQmxGetCIPulseTimeDigFltrEnable(self.task_handle, channel, data)

    def SetCIPulseTimeDigFltrEnable(self, channel, data):
        return DAQmxSetCIPulseTimeDigFltrEnable(self.task_handle, channel, data)

    def ResetCIPulseTimeDigFltrEnable(self, channel):
        return DAQmxResetCIPulseTimeDigFltrEnable(self.task_handle, channel)

    def GetCIPulseTimeDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPulseTimeDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIPulseTimeDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPulseTimeDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIPulseTimeDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPulseTimeDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIPulseTimeDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPulseTimeDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIPulseTimeDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPulseTimeDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIPulseTimeDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPulseTimeDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIPulseTimeDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPulseTimeDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIPulseTimeDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPulseTimeDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIPulseTimeDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPulseTimeDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIPulseTimeDigSyncEnable(self, channel, data):
        return DAQmxGetCIPulseTimeDigSyncEnable(self.task_handle, channel, data)

    def SetCIPulseTimeDigSyncEnable(self, channel, data):
        return DAQmxSetCIPulseTimeDigSyncEnable(self.task_handle, channel, data)

    def ResetCIPulseTimeDigSyncEnable(self, channel):
        return DAQmxResetCIPulseTimeDigSyncEnable(self.task_handle, channel)

    def GetCIPulseTicksTerm(self, channel, data, bufferSize):
        return DAQmxGetCIPulseTicksTerm(self.task_handle, channel, data, bufferSize)

    def SetCIPulseTicksTerm(self, channel, data):
        return DAQmxSetCIPulseTicksTerm(self.task_handle, channel, data)

    def ResetCIPulseTicksTerm(self, channel):
        return DAQmxResetCIPulseTicksTerm(self.task_handle, channel)

    def GetCIPulseTicksStartEdge(self, channel, data):
        return DAQmxGetCIPulseTicksStartEdge(self.task_handle, channel, data)

    def SetCIPulseTicksStartEdge(self, channel, data):
        return DAQmxSetCIPulseTicksStartEdge(self.task_handle, channel, data)

    def ResetCIPulseTicksStartEdge(self, channel):
        return DAQmxResetCIPulseTicksStartEdge(self.task_handle, channel)

    def GetCIPulseTicksDigFltrEnable(self, channel, data):
        return DAQmxGetCIPulseTicksDigFltrEnable(self.task_handle, channel, data)

    def SetCIPulseTicksDigFltrEnable(self, channel, data):
        return DAQmxSetCIPulseTicksDigFltrEnable(self.task_handle, channel, data)

    def ResetCIPulseTicksDigFltrEnable(self, channel):
        return DAQmxResetCIPulseTicksDigFltrEnable(self.task_handle, channel)

    def GetCIPulseTicksDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCIPulseTicksDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCIPulseTicksDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCIPulseTicksDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCIPulseTicksDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCIPulseTicksDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCIPulseTicksDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCIPulseTicksDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCIPulseTicksDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCIPulseTicksDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCIPulseTicksDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCIPulseTicksDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCIPulseTicksDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCIPulseTicksDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCIPulseTicksDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCIPulseTicksDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCIPulseTicksDigFltrTimebaseRate(self, channel):
        return DAQmxResetCIPulseTicksDigFltrTimebaseRate(self.task_handle, channel)

    def GetCIPulseTicksDigSyncEnable(self, channel, data):
        return DAQmxGetCIPulseTicksDigSyncEnable(self.task_handle, channel, data)

    def SetCIPulseTicksDigSyncEnable(self, channel, data):
        return DAQmxSetCIPulseTicksDigSyncEnable(self.task_handle, channel, data)

    def ResetCIPulseTicksDigSyncEnable(self, channel):
        return DAQmxResetCIPulseTicksDigSyncEnable(self.task_handle, channel)

    def GetCITimestampInitialSeconds(self, channel, data):
        return DAQmxGetCITimestampInitialSeconds(self.task_handle, channel, data)

    def SetCITimestampInitialSeconds(self, channel, data):
        return DAQmxSetCITimestampInitialSeconds(self.task_handle, channel, data)

    def ResetCITimestampInitialSeconds(self, channel):
        return DAQmxResetCITimestampInitialSeconds(self.task_handle, channel)

    def GetCIGPSSyncMethod(self, channel, data):
        return DAQmxGetCIGPSSyncMethod(self.task_handle, channel, data)

    def SetCIGPSSyncMethod(self, channel, data):
        return DAQmxSetCIGPSSyncMethod(self.task_handle, channel, data)

    def ResetCIGPSSyncMethod(self, channel):
        return DAQmxResetCIGPSSyncMethod(self.task_handle, channel)

    def GetCIGPSSyncSrc(self, channel, data, bufferSize):
        return DAQmxGetCIGPSSyncSrc(self.task_handle, channel, data, bufferSize)

    def SetCIGPSSyncSrc(self, channel, data):
        return DAQmxSetCIGPSSyncSrc(self.task_handle, channel, data)

    def ResetCIGPSSyncSrc(self, channel):
        return DAQmxResetCIGPSSyncSrc(self.task_handle, channel)

    def GetCICtrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICtrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCICtrTimebaseSrc(self, channel, data):
        return DAQmxSetCICtrTimebaseSrc(self.task_handle, channel, data)

    def ResetCICtrTimebaseSrc(self, channel):
        return DAQmxResetCICtrTimebaseSrc(self.task_handle, channel)

    def GetCICtrTimebaseRate(self, channel, data):
        return DAQmxGetCICtrTimebaseRate(self.task_handle, channel, data)

    def SetCICtrTimebaseRate(self, channel, data):
        return DAQmxSetCICtrTimebaseRate(self.task_handle, channel, data)

    def ResetCICtrTimebaseRate(self, channel):
        return DAQmxResetCICtrTimebaseRate(self.task_handle, channel)

    def GetCICtrTimebaseActiveEdge(self, channel, data):
        return DAQmxGetCICtrTimebaseActiveEdge(self.task_handle, channel, data)

    def SetCICtrTimebaseActiveEdge(self, channel, data):
        return DAQmxSetCICtrTimebaseActiveEdge(self.task_handle, channel, data)

    def ResetCICtrTimebaseActiveEdge(self, channel):
        return DAQmxResetCICtrTimebaseActiveEdge(self.task_handle, channel)

    def GetCICtrTimebaseDigFltrEnable(self, channel, data):
        return DAQmxGetCICtrTimebaseDigFltrEnable(self.task_handle, channel, data)

    def SetCICtrTimebaseDigFltrEnable(self, channel, data):
        return DAQmxSetCICtrTimebaseDigFltrEnable(self.task_handle, channel, data)

    def ResetCICtrTimebaseDigFltrEnable(self, channel):
        return DAQmxResetCICtrTimebaseDigFltrEnable(self.task_handle, channel)

    def GetCICtrTimebaseDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCICtrTimebaseDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCICtrTimebaseDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCICtrTimebaseDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCICtrTimebaseDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCICtrTimebaseDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCICtrTimebaseDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCICtrTimebaseDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCICtrTimebaseDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCICtrTimebaseDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCICtrTimebaseDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCICtrTimebaseDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCICtrTimebaseDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCICtrTimebaseDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCICtrTimebaseDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCICtrTimebaseDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCICtrTimebaseDigFltrTimebaseRate(self, channel):
        return DAQmxResetCICtrTimebaseDigFltrTimebaseRate(self.task_handle, channel)

    def GetCICtrTimebaseDigSyncEnable(self, channel, data):
        return DAQmxGetCICtrTimebaseDigSyncEnable(self.task_handle, channel, data)

    def SetCICtrTimebaseDigSyncEnable(self, channel, data):
        return DAQmxSetCICtrTimebaseDigSyncEnable(self.task_handle, channel, data)

    def ResetCICtrTimebaseDigSyncEnable(self, channel):
        return DAQmxResetCICtrTimebaseDigSyncEnable(self.task_handle, channel)

    def GetCICount(self, channel, data):
        return DAQmxGetCICount(self.task_handle, channel, data)

    def GetCIOutputState(self, channel, data):
        return DAQmxGetCIOutputState(self.task_handle, channel, data)

    def GetCITCReached(self, channel, data):
        return DAQmxGetCITCReached(self.task_handle, channel, data)

    def GetCICtrTimebaseMasterTimebaseDiv(self, channel, data):
        return DAQmxGetCICtrTimebaseMasterTimebaseDiv(self.task_handle, channel, data)

    def SetCICtrTimebaseMasterTimebaseDiv(self, channel, data):
        return DAQmxSetCICtrTimebaseMasterTimebaseDiv(self.task_handle, channel, data)

    def ResetCICtrTimebaseMasterTimebaseDiv(self, channel):
        return DAQmxResetCICtrTimebaseMasterTimebaseDiv(self.task_handle, channel)

    def GetCIDataXferMech(self, channel, data):
        return DAQmxGetCIDataXferMech(self.task_handle, channel, data)

    def SetCIDataXferMech(self, channel, data):
        return DAQmxSetCIDataXferMech(self.task_handle, channel, data)

    def ResetCIDataXferMech(self, channel):
        return DAQmxResetCIDataXferMech(self.task_handle, channel)

    def GetCIDataXferReqCond(self, channel, data):
        return DAQmxGetCIDataXferReqCond(self.task_handle, channel, data)

    def SetCIDataXferReqCond(self, channel, data):
        return DAQmxSetCIDataXferReqCond(self.task_handle, channel, data)

    def ResetCIDataXferReqCond(self, channel):
        return DAQmxResetCIDataXferReqCond(self.task_handle, channel)

    def GetCIUsbXferReqSize(self, channel, data):
        return DAQmxGetCIUsbXferReqSize(self.task_handle, channel, data)

    def SetCIUsbXferReqSize(self, channel, data):
        return DAQmxSetCIUsbXferReqSize(self.task_handle, channel, data)

    def ResetCIUsbXferReqSize(self, channel):
        return DAQmxResetCIUsbXferReqSize(self.task_handle, channel)

    def GetCIUsbXferReqCount(self, channel, data):
        return DAQmxGetCIUsbXferReqCount(self.task_handle, channel, data)

    def SetCIUsbXferReqCount(self, channel, data):
        return DAQmxSetCIUsbXferReqCount(self.task_handle, channel, data)

    def ResetCIUsbXferReqCount(self, channel):
        return DAQmxResetCIUsbXferReqCount(self.task_handle, channel)

    def GetCIMemMapEnable(self, channel, data):
        return DAQmxGetCIMemMapEnable(self.task_handle, channel, data)

    def SetCIMemMapEnable(self, channel, data):
        return DAQmxSetCIMemMapEnable(self.task_handle, channel, data)

    def ResetCIMemMapEnable(self, channel):
        return DAQmxResetCIMemMapEnable(self.task_handle, channel)

    def GetCINumPossiblyInvalidSamps(self, channel, data):
        return DAQmxGetCINumPossiblyInvalidSamps(self.task_handle, channel, data)

    def GetCIDupCountPrevent(self, channel, data):
        return DAQmxGetCIDupCountPrevent(self.task_handle, channel, data)

    def SetCIDupCountPrevent(self, channel, data):
        return DAQmxSetCIDupCountPrevent(self.task_handle, channel, data)

    def ResetCIDupCountPrevent(self, channel):
        return DAQmxResetCIDupCountPrevent(self.task_handle, channel)

    def GetCIPrescaler(self, channel, data):
        return DAQmxGetCIPrescaler(self.task_handle, channel, data)

    def SetCIPrescaler(self, channel, data):
        return DAQmxSetCIPrescaler(self.task_handle, channel, data)

    def ResetCIPrescaler(self, channel):
        return DAQmxResetCIPrescaler(self.task_handle, channel)

    def GetCOOutputType(self, channel, data):
        return DAQmxGetCOOutputType(self.task_handle, channel, data)

    def GetCOPulseIdleState(self, channel, data):
        return DAQmxGetCOPulseIdleState(self.task_handle, channel, data)

    def SetCOPulseIdleState(self, channel, data):
        return DAQmxSetCOPulseIdleState(self.task_handle, channel, data)

    def ResetCOPulseIdleState(self, channel):
        return DAQmxResetCOPulseIdleState(self.task_handle, channel)

    def GetCOPulseTerm(self, channel, data, bufferSize):
        return DAQmxGetCOPulseTerm(self.task_handle, channel, data, bufferSize)

    def SetCOPulseTerm(self, channel, data):
        return DAQmxSetCOPulseTerm(self.task_handle, channel, data)

    def ResetCOPulseTerm(self, channel):
        return DAQmxResetCOPulseTerm(self.task_handle, channel)

    def GetCOPulseTimeUnits(self, channel, data):
        return DAQmxGetCOPulseTimeUnits(self.task_handle, channel, data)

    def SetCOPulseTimeUnits(self, channel, data):
        return DAQmxSetCOPulseTimeUnits(self.task_handle, channel, data)

    def ResetCOPulseTimeUnits(self, channel):
        return DAQmxResetCOPulseTimeUnits(self.task_handle, channel)

    def GetCOPulseHighTime(self, channel, data):
        return DAQmxGetCOPulseHighTime(self.task_handle, channel, data)

    def SetCOPulseHighTime(self, channel, data):
        return DAQmxSetCOPulseHighTime(self.task_handle, channel, data)

    def ResetCOPulseHighTime(self, channel):
        return DAQmxResetCOPulseHighTime(self.task_handle, channel)

    def GetCOPulseLowTime(self, channel, data):
        return DAQmxGetCOPulseLowTime(self.task_handle, channel, data)

    def SetCOPulseLowTime(self, channel, data):
        return DAQmxSetCOPulseLowTime(self.task_handle, channel, data)

    def ResetCOPulseLowTime(self, channel):
        return DAQmxResetCOPulseLowTime(self.task_handle, channel)

    def GetCOPulseTimeInitialDelay(self, channel, data):
        return DAQmxGetCOPulseTimeInitialDelay(self.task_handle, channel, data)

    def SetCOPulseTimeInitialDelay(self, channel, data):
        return DAQmxSetCOPulseTimeInitialDelay(self.task_handle, channel, data)

    def ResetCOPulseTimeInitialDelay(self, channel):
        return DAQmxResetCOPulseTimeInitialDelay(self.task_handle, channel)

    def GetCOPulseDutyCyc(self, channel, data):
        return DAQmxGetCOPulseDutyCyc(self.task_handle, channel, data)

    def SetCOPulseDutyCyc(self, channel, data):
        return DAQmxSetCOPulseDutyCyc(self.task_handle, channel, data)

    def ResetCOPulseDutyCyc(self, channel):
        return DAQmxResetCOPulseDutyCyc(self.task_handle, channel)

    def GetCOPulseFreqUnits(self, channel, data):
        return DAQmxGetCOPulseFreqUnits(self.task_handle, channel, data)

    def SetCOPulseFreqUnits(self, channel, data):
        return DAQmxSetCOPulseFreqUnits(self.task_handle, channel, data)

    def ResetCOPulseFreqUnits(self, channel):
        return DAQmxResetCOPulseFreqUnits(self.task_handle, channel)

    def GetCOPulseFreq(self, channel, data):
        return DAQmxGetCOPulseFreq(self.task_handle, channel, data)

    def SetCOPulseFreq(self, channel, data):
        return DAQmxSetCOPulseFreq(self.task_handle, channel, data)

    def ResetCOPulseFreq(self, channel):
        return DAQmxResetCOPulseFreq(self.task_handle, channel)

    def GetCOPulseFreqInitialDelay(self, channel, data):
        return DAQmxGetCOPulseFreqInitialDelay(self.task_handle, channel, data)

    def SetCOPulseFreqInitialDelay(self, channel, data):
        return DAQmxSetCOPulseFreqInitialDelay(self.task_handle, channel, data)

    def ResetCOPulseFreqInitialDelay(self, channel):
        return DAQmxResetCOPulseFreqInitialDelay(self.task_handle, channel)

    def GetCOPulseHighTicks(self, channel, data):
        return DAQmxGetCOPulseHighTicks(self.task_handle, channel, data)

    def SetCOPulseHighTicks(self, channel, data):
        return DAQmxSetCOPulseHighTicks(self.task_handle, channel, data)

    def ResetCOPulseHighTicks(self, channel):
        return DAQmxResetCOPulseHighTicks(self.task_handle, channel)

    def GetCOPulseLowTicks(self, channel, data):
        return DAQmxGetCOPulseLowTicks(self.task_handle, channel, data)

    def SetCOPulseLowTicks(self, channel, data):
        return DAQmxSetCOPulseLowTicks(self.task_handle, channel, data)

    def ResetCOPulseLowTicks(self, channel):
        return DAQmxResetCOPulseLowTicks(self.task_handle, channel)

    def GetCOPulseTicksInitialDelay(self, channel, data):
        return DAQmxGetCOPulseTicksInitialDelay(self.task_handle, channel, data)

    def SetCOPulseTicksInitialDelay(self, channel, data):
        return DAQmxSetCOPulseTicksInitialDelay(self.task_handle, channel, data)

    def ResetCOPulseTicksInitialDelay(self, channel):
        return DAQmxResetCOPulseTicksInitialDelay(self.task_handle, channel)

    def GetCOCtrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCOCtrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCOCtrTimebaseSrc(self, channel, data):
        return DAQmxSetCOCtrTimebaseSrc(self.task_handle, channel, data)

    def ResetCOCtrTimebaseSrc(self, channel):
        return DAQmxResetCOCtrTimebaseSrc(self.task_handle, channel)

    def GetCOCtrTimebaseRate(self, channel, data):
        return DAQmxGetCOCtrTimebaseRate(self.task_handle, channel, data)

    def SetCOCtrTimebaseRate(self, channel, data):
        return DAQmxSetCOCtrTimebaseRate(self.task_handle, channel, data)

    def ResetCOCtrTimebaseRate(self, channel):
        return DAQmxResetCOCtrTimebaseRate(self.task_handle, channel)

    def GetCOCtrTimebaseActiveEdge(self, channel, data):
        return DAQmxGetCOCtrTimebaseActiveEdge(self.task_handle, channel, data)

    def SetCOCtrTimebaseActiveEdge(self, channel, data):
        return DAQmxSetCOCtrTimebaseActiveEdge(self.task_handle, channel, data)

    def ResetCOCtrTimebaseActiveEdge(self, channel):
        return DAQmxResetCOCtrTimebaseActiveEdge(self.task_handle, channel)

    def GetCOCtrTimebaseDigFltrEnable(self, channel, data):
        return DAQmxGetCOCtrTimebaseDigFltrEnable(self.task_handle, channel, data)

    def SetCOCtrTimebaseDigFltrEnable(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigFltrEnable(self.task_handle, channel, data)

    def ResetCOCtrTimebaseDigFltrEnable(self, channel):
        return DAQmxResetCOCtrTimebaseDigFltrEnable(self.task_handle, channel)

    def GetCOCtrTimebaseDigFltrMinPulseWidth(self, channel, data):
        return DAQmxGetCOCtrTimebaseDigFltrMinPulseWidth(self.task_handle, channel, data)

    def SetCOCtrTimebaseDigFltrMinPulseWidth(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigFltrMinPulseWidth(self.task_handle, channel, data)

    def ResetCOCtrTimebaseDigFltrMinPulseWidth(self, channel):
        return DAQmxResetCOCtrTimebaseDigFltrMinPulseWidth(self.task_handle, channel)

    def GetCOCtrTimebaseDigFltrTimebaseSrc(self, channel, data, bufferSize):
        return DAQmxGetCOCtrTimebaseDigFltrTimebaseSrc(self.task_handle, channel, data, bufferSize)

    def SetCOCtrTimebaseDigFltrTimebaseSrc(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigFltrTimebaseSrc(self.task_handle, channel, data)

    def ResetCOCtrTimebaseDigFltrTimebaseSrc(self, channel):
        return DAQmxResetCOCtrTimebaseDigFltrTimebaseSrc(self.task_handle, channel)

    def GetCOCtrTimebaseDigFltrTimebaseRate(self, channel, data):
        return DAQmxGetCOCtrTimebaseDigFltrTimebaseRate(self.task_handle, channel, data)

    def SetCOCtrTimebaseDigFltrTimebaseRate(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigFltrTimebaseRate(self.task_handle, channel, data)

    def ResetCOCtrTimebaseDigFltrTimebaseRate(self, channel):
        return DAQmxResetCOCtrTimebaseDigFltrTimebaseRate(self.task_handle, channel)

    def GetCOCtrTimebaseDigSyncEnable(self, channel, data):
        return DAQmxGetCOCtrTimebaseDigSyncEnable(self.task_handle, channel, data)

    def SetCOCtrTimebaseDigSyncEnable(self, channel, data):
        return DAQmxSetCOCtrTimebaseDigSyncEnable(self.task_handle, channel, data)

    def ResetCOCtrTimebaseDigSyncEnable(self, channel):
        return DAQmxResetCOCtrTimebaseDigSyncEnable(self.task_handle, channel)

    def GetCOCount(self, channel, data):
        return DAQmxGetCOCount(self.task_handle, channel, data)

    def GetCOOutputState(self, channel, data):
        return DAQmxGetCOOutputState(self.task_handle, channel, data)

    def GetCOAutoIncrCnt(self, channel, data):
        return DAQmxGetCOAutoIncrCnt(self.task_handle, channel, data)

    def SetCOAutoIncrCnt(self, channel, data):
        return DAQmxSetCOAutoIncrCnt(self.task_handle, channel, data)

    def ResetCOAutoIncrCnt(self, channel):
        return DAQmxResetCOAutoIncrCnt(self.task_handle, channel)

    def GetCOCtrTimebaseMasterTimebaseDiv(self, channel, data):
        return DAQmxGetCOCtrTimebaseMasterTimebaseDiv(self.task_handle, channel, data)

    def SetCOCtrTimebaseMasterTimebaseDiv(self, channel, data):
        return DAQmxSetCOCtrTimebaseMasterTimebaseDiv(self.task_handle, channel, data)

    def ResetCOCtrTimebaseMasterTimebaseDiv(self, channel):
        return DAQmxResetCOCtrTimebaseMasterTimebaseDiv(self.task_handle, channel)

    def GetCOPulseDone(self, channel, data):
        return DAQmxGetCOPulseDone(self.task_handle, channel, data)

    def GetCOEnableInitialDelayOnRetrigger(self, channel, data):
        return DAQmxGetCOEnableInitialDelayOnRetrigger(self.task_handle, channel, data)

    def SetCOEnableInitialDelayOnRetrigger(self, channel, data):
        return DAQmxSetCOEnableInitialDelayOnRetrigger(self.task_handle, channel, data)

    def ResetCOEnableInitialDelayOnRetrigger(self, channel):
        return DAQmxResetCOEnableInitialDelayOnRetrigger(self.task_handle, channel)

    def GetCOConstrainedGenMode(self, channel, data):
        return DAQmxGetCOConstrainedGenMode(self.task_handle, channel, data)

    def SetCOConstrainedGenMode(self, channel, data):
        return DAQmxSetCOConstrainedGenMode(self.task_handle, channel, data)

    def ResetCOConstrainedGenMode(self, channel):
        return DAQmxResetCOConstrainedGenMode(self.task_handle, channel)

    def GetCOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxGetCOUseOnlyOnBrdMem(self.task_handle, channel, data)

    def SetCOUseOnlyOnBrdMem(self, channel, data):
        return DAQmxSetCOUseOnlyOnBrdMem(self.task_handle, channel, data)

    def ResetCOUseOnlyOnBrdMem(self, channel):
        return DAQmxResetCOUseOnlyOnBrdMem(self.task_handle, channel)

    def GetCODataXferMech(self, channel, data):
        return DAQmxGetCODataXferMech(self.task_handle, channel, data)

    def SetCODataXferMech(self, channel, data):
        return DAQmxSetCODataXferMech(self.task_handle, channel, data)

    def ResetCODataXferMech(self, channel):
        return DAQmxResetCODataXferMech(self.task_handle, channel)

    def GetCODataXferReqCond(self, channel, data):
        return DAQmxGetCODataXferReqCond(self.task_handle, channel, data)

    def SetCODataXferReqCond(self, channel, data):
        return DAQmxSetCODataXferReqCond(self.task_handle, channel, data)

    def ResetCODataXferReqCond(self, channel):
        return DAQmxResetCODataXferReqCond(self.task_handle, channel)

    def GetCOUsbXferReqSize(self, channel, data):
        return DAQmxGetCOUsbXferReqSize(self.task_handle, channel, data)

    def SetCOUsbXferReqSize(self, channel, data):
        return DAQmxSetCOUsbXferReqSize(self.task_handle, channel, data)

    def ResetCOUsbXferReqSize(self, channel):
        return DAQmxResetCOUsbXferReqSize(self.task_handle, channel)

    def GetCOUsbXferReqCount(self, channel, data):
        return DAQmxGetCOUsbXferReqCount(self.task_handle, channel, data)

    def SetCOUsbXferReqCount(self, channel, data):
        return DAQmxSetCOUsbXferReqCount(self.task_handle, channel, data)

    def ResetCOUsbXferReqCount(self, channel):
        return DAQmxResetCOUsbXferReqCount(self.task_handle, channel)

    def GetCOMemMapEnable(self, channel, data):
        return DAQmxGetCOMemMapEnable(self.task_handle, channel, data)

    def SetCOMemMapEnable(self, channel, data):
        return DAQmxSetCOMemMapEnable(self.task_handle, channel, data)

    def ResetCOMemMapEnable(self, channel):
        return DAQmxResetCOMemMapEnable(self.task_handle, channel)

    def GetCOPrescaler(self, channel, data):
        return DAQmxGetCOPrescaler(self.task_handle, channel, data)

    def SetCOPrescaler(self, channel, data):
        return DAQmxSetCOPrescaler(self.task_handle, channel, data)

    def ResetCOPrescaler(self, channel):
        return DAQmxResetCOPrescaler(self.task_handle, channel)

    def GetCORdyForNewVal(self, channel, data):
        return DAQmxGetCORdyForNewVal(self.task_handle, channel, data)

    def GetChanType(self, channel, data):
        return DAQmxGetChanType(self.task_handle, channel, data)

    def GetPhysicalChanName(self, channel, data, bufferSize):
        return DAQmxGetPhysicalChanName(self.task_handle, channel, data, bufferSize)

    def SetPhysicalChanName(self, channel, data):
        return DAQmxSetPhysicalChanName(self.task_handle, channel, data)

    def GetChanDescr(self, channel, data, bufferSize):
        return DAQmxGetChanDescr(self.task_handle, channel, data, bufferSize)

    def SetChanDescr(self, channel, data):
        return DAQmxSetChanDescr(self.task_handle, channel, data)

    def ResetChanDescr(self, channel):
        return DAQmxResetChanDescr(self.task_handle, channel)

    def GetChanIsGlobal(self, channel, data):
        return DAQmxGetChanIsGlobal(self.task_handle, channel, data)

    def GetExportedAIConvClkOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedAIConvClkOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedAIConvClkOutputTerm(self, data):
        return DAQmxSetExportedAIConvClkOutputTerm(self.task_handle, data)

    def ResetExportedAIConvClkOutputTerm(self):
        return DAQmxResetExportedAIConvClkOutputTerm(self.task_handle)

    def GetExportedAIConvClkPulsePolarity(self, data):
        return DAQmxGetExportedAIConvClkPulsePolarity(self.task_handle, data)

    def GetExported10MHzRefClkOutputTerm(self, data, bufferSize):
        return DAQmxGetExported10MHzRefClkOutputTerm(self.task_handle, data, bufferSize)

    def SetExported10MHzRefClkOutputTerm(self, data):
        return DAQmxSetExported10MHzRefClkOutputTerm(self.task_handle, data)

    def ResetExported10MHzRefClkOutputTerm(self):
        return DAQmxResetExported10MHzRefClkOutputTerm(self.task_handle)

    def GetExported20MHzTimebaseOutputTerm(self, data, bufferSize):
        return DAQmxGetExported20MHzTimebaseOutputTerm(self.task_handle, data, bufferSize)

    def SetExported20MHzTimebaseOutputTerm(self, data):
        return DAQmxSetExported20MHzTimebaseOutputTerm(self.task_handle, data)

    def ResetExported20MHzTimebaseOutputTerm(self):
        return DAQmxResetExported20MHzTimebaseOutputTerm(self.task_handle)

    def GetExportedSampClkOutputBehavior(self, data):
        return DAQmxGetExportedSampClkOutputBehavior(self.task_handle, data)

    def SetExportedSampClkOutputBehavior(self, data):
        return DAQmxSetExportedSampClkOutputBehavior(self.task_handle, data)

    def ResetExportedSampClkOutputBehavior(self):
        return DAQmxResetExportedSampClkOutputBehavior(self.task_handle)

    def GetExportedSampClkOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedSampClkOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedSampClkOutputTerm(self, data):
        return DAQmxSetExportedSampClkOutputTerm(self.task_handle, data)

    def ResetExportedSampClkOutputTerm(self):
        return DAQmxResetExportedSampClkOutputTerm(self.task_handle)

    def GetExportedSampClkDelayOffset(self, data):
        return DAQmxGetExportedSampClkDelayOffset(self.task_handle, data)

    def SetExportedSampClkDelayOffset(self, data):
        return DAQmxSetExportedSampClkDelayOffset(self.task_handle, data)

    def ResetExportedSampClkDelayOffset(self):
        return DAQmxResetExportedSampClkDelayOffset(self.task_handle)

    def GetExportedSampClkPulsePolarity(self, data):
        return DAQmxGetExportedSampClkPulsePolarity(self.task_handle, data)

    def SetExportedSampClkPulsePolarity(self, data):
        return DAQmxSetExportedSampClkPulsePolarity(self.task_handle, data)

    def ResetExportedSampClkPulsePolarity(self):
        return DAQmxResetExportedSampClkPulsePolarity(self.task_handle)

    def GetExportedSampClkTimebaseOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedSampClkTimebaseOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedSampClkTimebaseOutputTerm(self, data):
        return DAQmxSetExportedSampClkTimebaseOutputTerm(self.task_handle, data)

    def ResetExportedSampClkTimebaseOutputTerm(self):
        return DAQmxResetExportedSampClkTimebaseOutputTerm(self.task_handle)

    def GetExportedDividedSampClkTimebaseOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedDividedSampClkTimebaseOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedDividedSampClkTimebaseOutputTerm(self, data):
        return DAQmxSetExportedDividedSampClkTimebaseOutputTerm(self.task_handle, data)

    def ResetExportedDividedSampClkTimebaseOutputTerm(self):
        return DAQmxResetExportedDividedSampClkTimebaseOutputTerm(self.task_handle)

    def GetExportedAdvTrigOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedAdvTrigOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedAdvTrigOutputTerm(self, data):
        return DAQmxSetExportedAdvTrigOutputTerm(self.task_handle, data)

    def ResetExportedAdvTrigOutputTerm(self):
        return DAQmxResetExportedAdvTrigOutputTerm(self.task_handle)

    def GetExportedAdvTrigPulsePolarity(self, data):
        return DAQmxGetExportedAdvTrigPulsePolarity(self.task_handle, data)

    def GetExportedAdvTrigPulseWidthUnits(self, data):
        return DAQmxGetExportedAdvTrigPulseWidthUnits(self.task_handle, data)

    def SetExportedAdvTrigPulseWidthUnits(self, data):
        return DAQmxSetExportedAdvTrigPulseWidthUnits(self.task_handle, data)

    def ResetExportedAdvTrigPulseWidthUnits(self):
        return DAQmxResetExportedAdvTrigPulseWidthUnits(self.task_handle)

    def GetExportedAdvTrigPulseWidth(self, data):
        return DAQmxGetExportedAdvTrigPulseWidth(self.task_handle, data)

    def SetExportedAdvTrigPulseWidth(self, data):
        return DAQmxSetExportedAdvTrigPulseWidth(self.task_handle, data)

    def ResetExportedAdvTrigPulseWidth(self):
        return DAQmxResetExportedAdvTrigPulseWidth(self.task_handle)

    def GetExportedPauseTrigOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedPauseTrigOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedPauseTrigOutputTerm(self, data):
        return DAQmxSetExportedPauseTrigOutputTerm(self.task_handle, data)

    def ResetExportedPauseTrigOutputTerm(self):
        return DAQmxResetExportedPauseTrigOutputTerm(self.task_handle)

    def GetExportedPauseTrigLvlActiveLvl(self, data):
        return DAQmxGetExportedPauseTrigLvlActiveLvl(self.task_handle, data)

    def SetExportedPauseTrigLvlActiveLvl(self, data):
        return DAQmxSetExportedPauseTrigLvlActiveLvl(self.task_handle, data)

    def ResetExportedPauseTrigLvlActiveLvl(self):
        return DAQmxResetExportedPauseTrigLvlActiveLvl(self.task_handle)

    def GetExportedRefTrigOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedRefTrigOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedRefTrigOutputTerm(self, data):
        return DAQmxSetExportedRefTrigOutputTerm(self.task_handle, data)

    def ResetExportedRefTrigOutputTerm(self):
        return DAQmxResetExportedRefTrigOutputTerm(self.task_handle)

    def GetExportedRefTrigPulsePolarity(self, data):
        return DAQmxGetExportedRefTrigPulsePolarity(self.task_handle, data)

    def SetExportedRefTrigPulsePolarity(self, data):
        return DAQmxSetExportedRefTrigPulsePolarity(self.task_handle, data)

    def ResetExportedRefTrigPulsePolarity(self):
        return DAQmxResetExportedRefTrigPulsePolarity(self.task_handle)

    def GetExportedStartTrigOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedStartTrigOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedStartTrigOutputTerm(self, data):
        return DAQmxSetExportedStartTrigOutputTerm(self.task_handle, data)

    def ResetExportedStartTrigOutputTerm(self):
        return DAQmxResetExportedStartTrigOutputTerm(self.task_handle)

    def GetExportedStartTrigPulsePolarity(self, data):
        return DAQmxGetExportedStartTrigPulsePolarity(self.task_handle, data)

    def SetExportedStartTrigPulsePolarity(self, data):
        return DAQmxSetExportedStartTrigPulsePolarity(self.task_handle, data)

    def ResetExportedStartTrigPulsePolarity(self):
        return DAQmxResetExportedStartTrigPulsePolarity(self.task_handle)

    def GetExportedAdvCmpltEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedAdvCmpltEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedAdvCmpltEventOutputTerm(self, data):
        return DAQmxSetExportedAdvCmpltEventOutputTerm(self.task_handle, data)

    def ResetExportedAdvCmpltEventOutputTerm(self):
        return DAQmxResetExportedAdvCmpltEventOutputTerm(self.task_handle)

    def GetExportedAdvCmpltEventDelay(self, data):
        return DAQmxGetExportedAdvCmpltEventDelay(self.task_handle, data)

    def SetExportedAdvCmpltEventDelay(self, data):
        return DAQmxSetExportedAdvCmpltEventDelay(self.task_handle, data)

    def ResetExportedAdvCmpltEventDelay(self):
        return DAQmxResetExportedAdvCmpltEventDelay(self.task_handle)

    def GetExportedAdvCmpltEventPulsePolarity(self, data):
        return DAQmxGetExportedAdvCmpltEventPulsePolarity(self.task_handle, data)

    def SetExportedAdvCmpltEventPulsePolarity(self, data):
        return DAQmxSetExportedAdvCmpltEventPulsePolarity(self.task_handle, data)

    def ResetExportedAdvCmpltEventPulsePolarity(self):
        return DAQmxResetExportedAdvCmpltEventPulsePolarity(self.task_handle)

    def GetExportedAdvCmpltEventPulseWidth(self, data):
        return DAQmxGetExportedAdvCmpltEventPulseWidth(self.task_handle, data)

    def SetExportedAdvCmpltEventPulseWidth(self, data):
        return DAQmxSetExportedAdvCmpltEventPulseWidth(self.task_handle, data)

    def ResetExportedAdvCmpltEventPulseWidth(self):
        return DAQmxResetExportedAdvCmpltEventPulseWidth(self.task_handle)

    def GetExportedAIHoldCmpltEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedAIHoldCmpltEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedAIHoldCmpltEventOutputTerm(self, data):
        return DAQmxSetExportedAIHoldCmpltEventOutputTerm(self.task_handle, data)

    def ResetExportedAIHoldCmpltEventOutputTerm(self):
        return DAQmxResetExportedAIHoldCmpltEventOutputTerm(self.task_handle)

    def GetExportedAIHoldCmpltEventPulsePolarity(self, data):
        return DAQmxGetExportedAIHoldCmpltEventPulsePolarity(self.task_handle, data)

    def SetExportedAIHoldCmpltEventPulsePolarity(self, data):
        return DAQmxSetExportedAIHoldCmpltEventPulsePolarity(self.task_handle, data)

    def ResetExportedAIHoldCmpltEventPulsePolarity(self):
        return DAQmxResetExportedAIHoldCmpltEventPulsePolarity(self.task_handle)

    def GetExportedChangeDetectEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedChangeDetectEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedChangeDetectEventOutputTerm(self, data):
        return DAQmxSetExportedChangeDetectEventOutputTerm(self.task_handle, data)

    def ResetExportedChangeDetectEventOutputTerm(self):
        return DAQmxResetExportedChangeDetectEventOutputTerm(self.task_handle)

    def GetExportedChangeDetectEventPulsePolarity(self, data):
        return DAQmxGetExportedChangeDetectEventPulsePolarity(self.task_handle, data)

    def SetExportedChangeDetectEventPulsePolarity(self, data):
        return DAQmxSetExportedChangeDetectEventPulsePolarity(self.task_handle, data)

    def ResetExportedChangeDetectEventPulsePolarity(self):
        return DAQmxResetExportedChangeDetectEventPulsePolarity(self.task_handle)

    def GetExportedCtrOutEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedCtrOutEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedCtrOutEventOutputTerm(self, data):
        return DAQmxSetExportedCtrOutEventOutputTerm(self.task_handle, data)

    def ResetExportedCtrOutEventOutputTerm(self):
        return DAQmxResetExportedCtrOutEventOutputTerm(self.task_handle)

    def GetExportedCtrOutEventOutputBehavior(self, data):
        return DAQmxGetExportedCtrOutEventOutputBehavior(self.task_handle, data)

    def SetExportedCtrOutEventOutputBehavior(self, data):
        return DAQmxSetExportedCtrOutEventOutputBehavior(self.task_handle, data)

    def ResetExportedCtrOutEventOutputBehavior(self):
        return DAQmxResetExportedCtrOutEventOutputBehavior(self.task_handle)

    def GetExportedCtrOutEventPulsePolarity(self, data):
        return DAQmxGetExportedCtrOutEventPulsePolarity(self.task_handle, data)

    def SetExportedCtrOutEventPulsePolarity(self, data):
        return DAQmxSetExportedCtrOutEventPulsePolarity(self.task_handle, data)

    def ResetExportedCtrOutEventPulsePolarity(self):
        return DAQmxResetExportedCtrOutEventPulsePolarity(self.task_handle)

    def GetExportedCtrOutEventToggleIdleState(self, data):
        return DAQmxGetExportedCtrOutEventToggleIdleState(self.task_handle, data)

    def SetExportedCtrOutEventToggleIdleState(self, data):
        return DAQmxSetExportedCtrOutEventToggleIdleState(self.task_handle, data)

    def ResetExportedCtrOutEventToggleIdleState(self):
        return DAQmxResetExportedCtrOutEventToggleIdleState(self.task_handle)

    def GetExportedHshkEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedHshkEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedHshkEventOutputTerm(self, data):
        return DAQmxSetExportedHshkEventOutputTerm(self.task_handle, data)

    def ResetExportedHshkEventOutputTerm(self):
        return DAQmxResetExportedHshkEventOutputTerm(self.task_handle)

    def GetExportedHshkEventOutputBehavior(self, data):
        return DAQmxGetExportedHshkEventOutputBehavior(self.task_handle, data)

    def SetExportedHshkEventOutputBehavior(self, data):
        return DAQmxSetExportedHshkEventOutputBehavior(self.task_handle, data)

    def ResetExportedHshkEventOutputBehavior(self):
        return DAQmxResetExportedHshkEventOutputBehavior(self.task_handle)

    def GetExportedHshkEventDelay(self, data):
        return DAQmxGetExportedHshkEventDelay(self.task_handle, data)

    def SetExportedHshkEventDelay(self, data):
        return DAQmxSetExportedHshkEventDelay(self.task_handle, data)

    def ResetExportedHshkEventDelay(self):
        return DAQmxResetExportedHshkEventDelay(self.task_handle)

    def GetExportedHshkEventInterlockedAssertedLvl(self, data):
        return DAQmxGetExportedHshkEventInterlockedAssertedLvl(self.task_handle, data)

    def SetExportedHshkEventInterlockedAssertedLvl(self, data):
        return DAQmxSetExportedHshkEventInterlockedAssertedLvl(self.task_handle, data)

    def ResetExportedHshkEventInterlockedAssertedLvl(self):
        return DAQmxResetExportedHshkEventInterlockedAssertedLvl(self.task_handle)

    def GetExportedHshkEventInterlockedAssertOnStart(self, data):
        return DAQmxGetExportedHshkEventInterlockedAssertOnStart(self.task_handle, data)

    def SetExportedHshkEventInterlockedAssertOnStart(self, data):
        return DAQmxSetExportedHshkEventInterlockedAssertOnStart(self.task_handle, data)

    def ResetExportedHshkEventInterlockedAssertOnStart(self):
        return DAQmxResetExportedHshkEventInterlockedAssertOnStart(self.task_handle)

    def GetExportedHshkEventInterlockedDeassertDelay(self, data):
        return DAQmxGetExportedHshkEventInterlockedDeassertDelay(self.task_handle, data)

    def SetExportedHshkEventInterlockedDeassertDelay(self, data):
        return DAQmxSetExportedHshkEventInterlockedDeassertDelay(self.task_handle, data)

    def ResetExportedHshkEventInterlockedDeassertDelay(self):
        return DAQmxResetExportedHshkEventInterlockedDeassertDelay(self.task_handle)

    def GetExportedHshkEventPulsePolarity(self, data):
        return DAQmxGetExportedHshkEventPulsePolarity(self.task_handle, data)

    def SetExportedHshkEventPulsePolarity(self, data):
        return DAQmxSetExportedHshkEventPulsePolarity(self.task_handle, data)

    def ResetExportedHshkEventPulsePolarity(self):
        return DAQmxResetExportedHshkEventPulsePolarity(self.task_handle)

    def GetExportedHshkEventPulseWidth(self, data):
        return DAQmxGetExportedHshkEventPulseWidth(self.task_handle, data)

    def SetExportedHshkEventPulseWidth(self, data):
        return DAQmxSetExportedHshkEventPulseWidth(self.task_handle, data)

    def ResetExportedHshkEventPulseWidth(self):
        return DAQmxResetExportedHshkEventPulseWidth(self.task_handle)

    def GetExportedRdyForXferEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedRdyForXferEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedRdyForXferEventOutputTerm(self, data):
        return DAQmxSetExportedRdyForXferEventOutputTerm(self.task_handle, data)

    def ResetExportedRdyForXferEventOutputTerm(self):
        return DAQmxResetExportedRdyForXferEventOutputTerm(self.task_handle)

    def GetExportedRdyForXferEventLvlActiveLvl(self, data):
        return DAQmxGetExportedRdyForXferEventLvlActiveLvl(self.task_handle, data)

    def SetExportedRdyForXferEventLvlActiveLvl(self, data):
        return DAQmxSetExportedRdyForXferEventLvlActiveLvl(self.task_handle, data)

    def ResetExportedRdyForXferEventLvlActiveLvl(self):
        return DAQmxResetExportedRdyForXferEventLvlActiveLvl(self.task_handle)

    def GetExportedRdyForXferEventDeassertCond(self, data):
        return DAQmxGetExportedRdyForXferEventDeassertCond(self.task_handle, data)

    def SetExportedRdyForXferEventDeassertCond(self, data):
        return DAQmxSetExportedRdyForXferEventDeassertCond(self.task_handle, data)

    def ResetExportedRdyForXferEventDeassertCond(self):
        return DAQmxResetExportedRdyForXferEventDeassertCond(self.task_handle)

    def GetExportedRdyForXferEventDeassertCondCustomThreshold(self, data):
        return DAQmxGetExportedRdyForXferEventDeassertCondCustomThreshold(self.task_handle, data)

    def SetExportedRdyForXferEventDeassertCondCustomThreshold(self, data):
        return DAQmxSetExportedRdyForXferEventDeassertCondCustomThreshold(self.task_handle, data)

    def ResetExportedRdyForXferEventDeassertCondCustomThreshold(self):
        return DAQmxResetExportedRdyForXferEventDeassertCondCustomThreshold(self.task_handle)

    def GetExportedDataActiveEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedDataActiveEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedDataActiveEventOutputTerm(self, data):
        return DAQmxSetExportedDataActiveEventOutputTerm(self.task_handle, data)

    def ResetExportedDataActiveEventOutputTerm(self):
        return DAQmxResetExportedDataActiveEventOutputTerm(self.task_handle)

    def GetExportedDataActiveEventLvlActiveLvl(self, data):
        return DAQmxGetExportedDataActiveEventLvlActiveLvl(self.task_handle, data)

    def SetExportedDataActiveEventLvlActiveLvl(self, data):
        return DAQmxSetExportedDataActiveEventLvlActiveLvl(self.task_handle, data)

    def ResetExportedDataActiveEventLvlActiveLvl(self):
        return DAQmxResetExportedDataActiveEventLvlActiveLvl(self.task_handle)

    def GetExportedRdyForStartEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedRdyForStartEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedRdyForStartEventOutputTerm(self, data):
        return DAQmxSetExportedRdyForStartEventOutputTerm(self.task_handle, data)

    def ResetExportedRdyForStartEventOutputTerm(self):
        return DAQmxResetExportedRdyForStartEventOutputTerm(self.task_handle)

    def GetExportedRdyForStartEventLvlActiveLvl(self, data):
        return DAQmxGetExportedRdyForStartEventLvlActiveLvl(self.task_handle, data)

    def SetExportedRdyForStartEventLvlActiveLvl(self, data):
        return DAQmxSetExportedRdyForStartEventLvlActiveLvl(self.task_handle, data)

    def ResetExportedRdyForStartEventLvlActiveLvl(self):
        return DAQmxResetExportedRdyForStartEventLvlActiveLvl(self.task_handle)

    def GetExportedSyncPulseEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedSyncPulseEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedSyncPulseEventOutputTerm(self, data):
        return DAQmxSetExportedSyncPulseEventOutputTerm(self.task_handle, data)

    def ResetExportedSyncPulseEventOutputTerm(self):
        return DAQmxResetExportedSyncPulseEventOutputTerm(self.task_handle)

    def GetExportedWatchdogExpiredEventOutputTerm(self, data, bufferSize):
        return DAQmxGetExportedWatchdogExpiredEventOutputTerm(self.task_handle, data, bufferSize)

    def SetExportedWatchdogExpiredEventOutputTerm(self, data):
        return DAQmxSetExportedWatchdogExpiredEventOutputTerm(self.task_handle, data)

    def ResetExportedWatchdogExpiredEventOutputTerm(self):
        return DAQmxResetExportedWatchdogExpiredEventOutputTerm(self.task_handle)

    def GetReadRelativeTo(self, data):
        return DAQmxGetReadRelativeTo(self.task_handle, data)

    def SetReadRelativeTo(self, data):
        return DAQmxSetReadRelativeTo(self.task_handle, data)

    def ResetReadRelativeTo(self):
        return DAQmxResetReadRelativeTo(self.task_handle)

    def GetReadOffset(self, data):
        return DAQmxGetReadOffset(self.task_handle, data)

    def SetReadOffset(self, data):
        return DAQmxSetReadOffset(self.task_handle, data)

    def ResetReadOffset(self):
        return DAQmxResetReadOffset(self.task_handle)

    def GetReadChannelsToRead(self, data, bufferSize):
        return DAQmxGetReadChannelsToRead(self.task_handle, data, bufferSize)

    def SetReadChannelsToRead(self, data):
        return DAQmxSetReadChannelsToRead(self.task_handle, data)

    def ResetReadChannelsToRead(self):
        return DAQmxResetReadChannelsToRead(self.task_handle)

    def GetReadReadAllAvailSamp(self, data):
        return DAQmxGetReadReadAllAvailSamp(self.task_handle, data)

    def SetReadReadAllAvailSamp(self, data):
        return DAQmxSetReadReadAllAvailSamp(self.task_handle, data)

    def ResetReadReadAllAvailSamp(self):
        return DAQmxResetReadReadAllAvailSamp(self.task_handle)

    def GetReadAutoStart(self, data):
        return DAQmxGetReadAutoStart(self.task_handle, data)

    def SetReadAutoStart(self, data):
        return DAQmxSetReadAutoStart(self.task_handle, data)

    def ResetReadAutoStart(self):
        return DAQmxResetReadAutoStart(self.task_handle)

    def GetReadOverWrite(self, data):
        return DAQmxGetReadOverWrite(self.task_handle, data)

    def SetReadOverWrite(self, data):
        return DAQmxSetReadOverWrite(self.task_handle, data)

    def ResetReadOverWrite(self):
        return DAQmxResetReadOverWrite(self.task_handle)

    def GetReadCurrReadPos(self, data):
        return DAQmxGetReadCurrReadPos(self.task_handle, data)

    def GetReadAvailSampPerChan(self, data):
        return DAQmxGetReadAvailSampPerChan(self.task_handle, data)

    def GetLoggingFilePath(self, data, bufferSize):
        return DAQmxGetLoggingFilePath(self.task_handle, data, bufferSize)

    def SetLoggingFilePath(self, data):
        return DAQmxSetLoggingFilePath(self.task_handle, data)

    def ResetLoggingFilePath(self):
        return DAQmxResetLoggingFilePath(self.task_handle)

    def GetLoggingMode(self, data):
        return DAQmxGetLoggingMode(self.task_handle, data)

    def SetLoggingMode(self, data):
        return DAQmxSetLoggingMode(self.task_handle, data)

    def ResetLoggingMode(self):
        return DAQmxResetLoggingMode(self.task_handle)

    def GetLoggingTDMSGroupName(self, data, bufferSize):
        return DAQmxGetLoggingTDMSGroupName(self.task_handle, data, bufferSize)

    def SetLoggingTDMSGroupName(self, data):
        return DAQmxSetLoggingTDMSGroupName(self.task_handle, data)

    def ResetLoggingTDMSGroupName(self):
        return DAQmxResetLoggingTDMSGroupName(self.task_handle)

    def GetLoggingTDMSOperation(self, data):
        return DAQmxGetLoggingTDMSOperation(self.task_handle, data)

    def SetLoggingTDMSOperation(self, data):
        return DAQmxSetLoggingTDMSOperation(self.task_handle, data)

    def ResetLoggingTDMSOperation(self):
        return DAQmxResetLoggingTDMSOperation(self.task_handle)

    def GetLoggingPause(self, data):
        return DAQmxGetLoggingPause(self.task_handle, data)

    def SetLoggingPause(self, data):
        return DAQmxSetLoggingPause(self.task_handle, data)

    def ResetLoggingPause(self):
        return DAQmxResetLoggingPause(self.task_handle)

    def GetLoggingSampsPerFile(self, data):
        return DAQmxGetLoggingSampsPerFile(self.task_handle, data)

    def SetLoggingSampsPerFile(self, data):
        return DAQmxSetLoggingSampsPerFile(self.task_handle, data)

    def ResetLoggingSampsPerFile(self):
        return DAQmxResetLoggingSampsPerFile(self.task_handle)

    def GetLoggingFileWriteSize(self, data):
        return DAQmxGetLoggingFileWriteSize(self.task_handle, data)

    def SetLoggingFileWriteSize(self, data):
        return DAQmxSetLoggingFileWriteSize(self.task_handle, data)

    def ResetLoggingFileWriteSize(self):
        return DAQmxResetLoggingFileWriteSize(self.task_handle)

    def GetLoggingFilePreallocationSize(self, data):
        return DAQmxGetLoggingFilePreallocationSize(self.task_handle, data)

    def SetLoggingFilePreallocationSize(self, data):
        return DAQmxSetLoggingFilePreallocationSize(self.task_handle, data)

    def ResetLoggingFilePreallocationSize(self):
        return DAQmxResetLoggingFilePreallocationSize(self.task_handle)

    def GetReadTotalSampPerChanAcquired(self, data):
        return DAQmxGetReadTotalSampPerChanAcquired(self.task_handle, data)

    def GetReadCommonModeRangeErrorChansExist(self, data):
        return DAQmxGetReadCommonModeRangeErrorChansExist(self.task_handle, data)

    def GetReadCommonModeRangeErrorChans(self, data, bufferSize):
        return DAQmxGetReadCommonModeRangeErrorChans(self.task_handle, data, bufferSize)

    def GetReadOvercurrentChansExist(self, data):
        return DAQmxGetReadOvercurrentChansExist(self.task_handle, data)

    def GetReadOvercurrentChans(self, data, bufferSize):
        return DAQmxGetReadOvercurrentChans(self.task_handle, data, bufferSize)

    def GetReadOpenCurrentLoopChansExist(self, data):
        return DAQmxGetReadOpenCurrentLoopChansExist(self.task_handle, data)

    def GetReadOpenCurrentLoopChans(self, data, bufferSize):
        return DAQmxGetReadOpenCurrentLoopChans(self.task_handle, data, bufferSize)

    def GetReadOpenThrmcplChansExist(self, data):
        return DAQmxGetReadOpenThrmcplChansExist(self.task_handle, data)

    def GetReadOpenThrmcplChans(self, data, bufferSize):
        return DAQmxGetReadOpenThrmcplChans(self.task_handle, data, bufferSize)

    def GetReadOverloadedChansExist(self, data):
        return DAQmxGetReadOverloadedChansExist(self.task_handle, data)

    def GetReadOverloadedChans(self, data, bufferSize):
        return DAQmxGetReadOverloadedChans(self.task_handle, data, bufferSize)

    def GetReadAccessoryInsertionOrRemovalDetected(self, data):
        return DAQmxGetReadAccessoryInsertionOrRemovalDetected(self.task_handle, data)

    def GetReadDevsWithInsertedOrRemovedAccessories(self, data, bufferSize):
        return DAQmxGetReadDevsWithInsertedOrRemovedAccessories(self.task_handle, data, bufferSize)

    def GetReadChangeDetectHasOverflowed(self, data):
        return DAQmxGetReadChangeDetectHasOverflowed(self.task_handle, data)

    def GetReadRawDataWidth(self, data):
        return DAQmxGetReadRawDataWidth(self.task_handle, data)

    def GetReadNumChans(self, data):
        return DAQmxGetReadNumChans(self.task_handle, data)

    def GetReadDigitalLinesBytesPerChan(self, data):
        return DAQmxGetReadDigitalLinesBytesPerChan(self.task_handle, data)

    def GetReadWaitMode(self, data):
        return DAQmxGetReadWaitMode(self.task_handle, data)

    def SetReadWaitMode(self, data):
        return DAQmxSetReadWaitMode(self.task_handle, data)

    def ResetReadWaitMode(self):
        return DAQmxResetReadWaitMode(self.task_handle)

    def GetReadSleepTime(self, data):
        return DAQmxGetReadSleepTime(self.task_handle, data)

    def SetReadSleepTime(self, data):
        return DAQmxSetReadSleepTime(self.task_handle, data)

    def ResetReadSleepTime(self):
        return DAQmxResetReadSleepTime(self.task_handle)

    def GetRealTimeConvLateErrorsToWarnings(self, data):
        return DAQmxGetRealTimeConvLateErrorsToWarnings(self.task_handle, data)

    def SetRealTimeConvLateErrorsToWarnings(self, data):
        return DAQmxSetRealTimeConvLateErrorsToWarnings(self.task_handle, data)

    def ResetRealTimeConvLateErrorsToWarnings(self):
        return DAQmxResetRealTimeConvLateErrorsToWarnings(self.task_handle)

    def GetRealTimeNumOfWarmupIters(self, data):
        return DAQmxGetRealTimeNumOfWarmupIters(self.task_handle, data)

    def SetRealTimeNumOfWarmupIters(self, data):
        return DAQmxSetRealTimeNumOfWarmupIters(self.task_handle, data)

    def ResetRealTimeNumOfWarmupIters(self):
        return DAQmxResetRealTimeNumOfWarmupIters(self.task_handle)

    def GetRealTimeWaitForNextSampClkWaitMode(self, data):
        return DAQmxGetRealTimeWaitForNextSampClkWaitMode(self.task_handle, data)

    def SetRealTimeWaitForNextSampClkWaitMode(self, data):
        return DAQmxSetRealTimeWaitForNextSampClkWaitMode(self.task_handle, data)

    def ResetRealTimeWaitForNextSampClkWaitMode(self):
        return DAQmxResetRealTimeWaitForNextSampClkWaitMode(self.task_handle)

    def GetRealTimeReportMissedSamp(self, data):
        return DAQmxGetRealTimeReportMissedSamp(self.task_handle, data)

    def SetRealTimeReportMissedSamp(self, data):
        return DAQmxSetRealTimeReportMissedSamp(self.task_handle, data)

    def ResetRealTimeReportMissedSamp(self):
        return DAQmxResetRealTimeReportMissedSamp(self.task_handle)

    def GetRealTimeWriteRecoveryMode(self, data):
        return DAQmxGetRealTimeWriteRecoveryMode(self.task_handle, data)

    def SetRealTimeWriteRecoveryMode(self, data):
        return DAQmxSetRealTimeWriteRecoveryMode(self.task_handle, data)

    def ResetRealTimeWriteRecoveryMode(self):
        return DAQmxResetRealTimeWriteRecoveryMode(self.task_handle)

    def GetSwitchScanBreakMode(self, data):
        return DAQmxGetSwitchScanBreakMode(self.task_handle, data)

    def SetSwitchScanBreakMode(self, data):
        return DAQmxSetSwitchScanBreakMode(self.task_handle, data)

    def ResetSwitchScanBreakMode(self):
        return DAQmxResetSwitchScanBreakMode(self.task_handle)

    def GetSwitchScanRepeatMode(self, data):
        return DAQmxGetSwitchScanRepeatMode(self.task_handle, data)

    def SetSwitchScanRepeatMode(self, data):
        return DAQmxSetSwitchScanRepeatMode(self.task_handle, data)

    def ResetSwitchScanRepeatMode(self):
        return DAQmxResetSwitchScanRepeatMode(self.task_handle)

    def GetSwitchScanWaitingForAdv(self, data):
        return DAQmxGetSwitchScanWaitingForAdv(self.task_handle, data)

    def GetTaskName(self, data, bufferSize):
        return DAQmxGetTaskName(self.task_handle, data, bufferSize)

    def GetTaskChannels(self, data, bufferSize):
        return DAQmxGetTaskChannels(self.task_handle, data, bufferSize)

    def GetTaskNumChans(self, data):
        return DAQmxGetTaskNumChans(self.task_handle, data)

    def GetTaskDevices(self, data, bufferSize):
        return DAQmxGetTaskDevices(self.task_handle, data, bufferSize)

    def GetTaskNumDevices(self, data):
        return DAQmxGetTaskNumDevices(self.task_handle, data)

    def GetTaskComplete(self, data):
        return DAQmxGetTaskComplete(self.task_handle, data)

    def GetSampQuantSampMode(self, data):
        return DAQmxGetSampQuantSampMode(self.task_handle, data)

    def SetSampQuantSampMode(self, data):
        return DAQmxSetSampQuantSampMode(self.task_handle, data)

    def ResetSampQuantSampMode(self):
        return DAQmxResetSampQuantSampMode(self.task_handle)

    def GetSampQuantSampPerChan(self, data):
        return DAQmxGetSampQuantSampPerChan(self.task_handle, data)

    def SetSampQuantSampPerChan(self, data):
        return DAQmxSetSampQuantSampPerChan(self.task_handle, data)

    def ResetSampQuantSampPerChan(self):
        return DAQmxResetSampQuantSampPerChan(self.task_handle)

    def GetSampTimingType(self, data):
        return DAQmxGetSampTimingType(self.task_handle, data)

    def SetSampTimingType(self, data):
        return DAQmxSetSampTimingType(self.task_handle, data)

    def ResetSampTimingType(self):
        return DAQmxResetSampTimingType(self.task_handle)

    def GetSampClkRate(self, data):
        return DAQmxGetSampClkRate(self.task_handle, data)

    def SetSampClkRate(self, data):
        return DAQmxSetSampClkRate(self.task_handle, data)

    def ResetSampClkRate(self):
        return DAQmxResetSampClkRate(self.task_handle)

    def GetSampClkMaxRate(self, data):
        return DAQmxGetSampClkMaxRate(self.task_handle, data)

    def GetSampClkSrc(self, data, bufferSize):
        return DAQmxGetSampClkSrc(self.task_handle, data, bufferSize)

    def SetSampClkSrc(self, data):
        return DAQmxSetSampClkSrc(self.task_handle, data)

    def ResetSampClkSrc(self):
        return DAQmxResetSampClkSrc(self.task_handle)

    def GetSampClkActiveEdge(self, data):
        return DAQmxGetSampClkActiveEdge(self.task_handle, data)

    def SetSampClkActiveEdge(self, data):
        return DAQmxSetSampClkActiveEdge(self.task_handle, data)

    def ResetSampClkActiveEdge(self):
        return DAQmxResetSampClkActiveEdge(self.task_handle)

    def GetSampClkOverrunBehavior(self, data):
        return DAQmxGetSampClkOverrunBehavior(self.task_handle, data)

    def SetSampClkOverrunBehavior(self, data):
        return DAQmxSetSampClkOverrunBehavior(self.task_handle, data)

    def ResetSampClkOverrunBehavior(self):
        return DAQmxResetSampClkOverrunBehavior(self.task_handle)

    def GetSampClkUnderflowBehavior(self, data):
        return DAQmxGetSampClkUnderflowBehavior(self.task_handle, data)

    def SetSampClkUnderflowBehavior(self, data):
        return DAQmxSetSampClkUnderflowBehavior(self.task_handle, data)

    def ResetSampClkUnderflowBehavior(self):
        return DAQmxResetSampClkUnderflowBehavior(self.task_handle)

    def GetSampClkTimebaseDiv(self, data):
        return DAQmxGetSampClkTimebaseDiv(self.task_handle, data)

    def SetSampClkTimebaseDiv(self, data):
        return DAQmxSetSampClkTimebaseDiv(self.task_handle, data)

    def ResetSampClkTimebaseDiv(self):
        return DAQmxResetSampClkTimebaseDiv(self.task_handle)

    def GetSampClkTerm(self, data, bufferSize):
        return DAQmxGetSampClkTerm(self.task_handle, data, bufferSize)

    def GetSampClkTimebaseRate(self, data):
        return DAQmxGetSampClkTimebaseRate(self.task_handle, data)

    def SetSampClkTimebaseRate(self, data):
        return DAQmxSetSampClkTimebaseRate(self.task_handle, data)

    def ResetSampClkTimebaseRate(self):
        return DAQmxResetSampClkTimebaseRate(self.task_handle)

    def GetSampClkTimebaseSrc(self, data, bufferSize):
        return DAQmxGetSampClkTimebaseSrc(self.task_handle, data, bufferSize)

    def SetSampClkTimebaseSrc(self, data):
        return DAQmxSetSampClkTimebaseSrc(self.task_handle, data)

    def ResetSampClkTimebaseSrc(self):
        return DAQmxResetSampClkTimebaseSrc(self.task_handle)

    def GetSampClkTimebaseActiveEdge(self, data):
        return DAQmxGetSampClkTimebaseActiveEdge(self.task_handle, data)

    def SetSampClkTimebaseActiveEdge(self, data):
        return DAQmxSetSampClkTimebaseActiveEdge(self.task_handle, data)

    def ResetSampClkTimebaseActiveEdge(self):
        return DAQmxResetSampClkTimebaseActiveEdge(self.task_handle)

    def GetSampClkTimebaseMasterTimebaseDiv(self, data):
        return DAQmxGetSampClkTimebaseMasterTimebaseDiv(self.task_handle, data)

    def SetSampClkTimebaseMasterTimebaseDiv(self, data):
        return DAQmxSetSampClkTimebaseMasterTimebaseDiv(self.task_handle, data)

    def ResetSampClkTimebaseMasterTimebaseDiv(self):
        return DAQmxResetSampClkTimebaseMasterTimebaseDiv(self.task_handle)

    def GetSampClkTimebaseTerm(self, data, bufferSize):
        return DAQmxGetSampClkTimebaseTerm(self.task_handle, data, bufferSize)

    def GetSampClkDigFltrEnable(self, data):
        return DAQmxGetSampClkDigFltrEnable(self.task_handle, data)

    def SetSampClkDigFltrEnable(self, data):
        return DAQmxSetSampClkDigFltrEnable(self.task_handle, data)

    def ResetSampClkDigFltrEnable(self):
        return DAQmxResetSampClkDigFltrEnable(self.task_handle)

    def GetSampClkDigFltrMinPulseWidth(self, data):
        return DAQmxGetSampClkDigFltrMinPulseWidth(self.task_handle, data)

    def SetSampClkDigFltrMinPulseWidth(self, data):
        return DAQmxSetSampClkDigFltrMinPulseWidth(self.task_handle, data)

    def ResetSampClkDigFltrMinPulseWidth(self):
        return DAQmxResetSampClkDigFltrMinPulseWidth(self.task_handle)

    def GetSampClkDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetSampClkDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetSampClkDigFltrTimebaseSrc(self, data):
        return DAQmxSetSampClkDigFltrTimebaseSrc(self.task_handle, data)

    def ResetSampClkDigFltrTimebaseSrc(self):
        return DAQmxResetSampClkDigFltrTimebaseSrc(self.task_handle)

    def GetSampClkDigFltrTimebaseRate(self, data):
        return DAQmxGetSampClkDigFltrTimebaseRate(self.task_handle, data)

    def SetSampClkDigFltrTimebaseRate(self, data):
        return DAQmxSetSampClkDigFltrTimebaseRate(self.task_handle, data)

    def ResetSampClkDigFltrTimebaseRate(self):
        return DAQmxResetSampClkDigFltrTimebaseRate(self.task_handle)

    def GetSampClkDigSyncEnable(self, data):
        return DAQmxGetSampClkDigSyncEnable(self.task_handle, data)

    def SetSampClkDigSyncEnable(self, data):
        return DAQmxSetSampClkDigSyncEnable(self.task_handle, data)

    def ResetSampClkDigSyncEnable(self):
        return DAQmxResetSampClkDigSyncEnable(self.task_handle)

    def GetHshkDelayAfterXfer(self, data):
        return DAQmxGetHshkDelayAfterXfer(self.task_handle, data)

    def SetHshkDelayAfterXfer(self, data):
        return DAQmxSetHshkDelayAfterXfer(self.task_handle, data)

    def ResetHshkDelayAfterXfer(self):
        return DAQmxResetHshkDelayAfterXfer(self.task_handle)

    def GetHshkStartCond(self, data):
        return DAQmxGetHshkStartCond(self.task_handle, data)

    def SetHshkStartCond(self, data):
        return DAQmxSetHshkStartCond(self.task_handle, data)

    def ResetHshkStartCond(self):
        return DAQmxResetHshkStartCond(self.task_handle)

    def GetHshkSampleInputDataWhen(self, data):
        return DAQmxGetHshkSampleInputDataWhen(self.task_handle, data)

    def SetHshkSampleInputDataWhen(self, data):
        return DAQmxSetHshkSampleInputDataWhen(self.task_handle, data)

    def ResetHshkSampleInputDataWhen(self):
        return DAQmxResetHshkSampleInputDataWhen(self.task_handle)

    def GetChangeDetectDIRisingEdgePhysicalChans(self, data, bufferSize):
        return DAQmxGetChangeDetectDIRisingEdgePhysicalChans(self.task_handle, data, bufferSize)

    def SetChangeDetectDIRisingEdgePhysicalChans(self, data):
        return DAQmxSetChangeDetectDIRisingEdgePhysicalChans(self.task_handle, data)

    def ResetChangeDetectDIRisingEdgePhysicalChans(self):
        return DAQmxResetChangeDetectDIRisingEdgePhysicalChans(self.task_handle)

    def GetChangeDetectDIFallingEdgePhysicalChans(self, data, bufferSize):
        return DAQmxGetChangeDetectDIFallingEdgePhysicalChans(self.task_handle, data, bufferSize)

    def SetChangeDetectDIFallingEdgePhysicalChans(self, data):
        return DAQmxSetChangeDetectDIFallingEdgePhysicalChans(self.task_handle, data)

    def ResetChangeDetectDIFallingEdgePhysicalChans(self):
        return DAQmxResetChangeDetectDIFallingEdgePhysicalChans(self.task_handle)

    def GetChangeDetectDITristate(self, data):
        return DAQmxGetChangeDetectDITristate(self.task_handle, data)

    def SetChangeDetectDITristate(self, data):
        return DAQmxSetChangeDetectDITristate(self.task_handle, data)

    def ResetChangeDetectDITristate(self):
        return DAQmxResetChangeDetectDITristate(self.task_handle)

    def GetOnDemandSimultaneousAOEnable(self, data):
        return DAQmxGetOnDemandSimultaneousAOEnable(self.task_handle, data)

    def SetOnDemandSimultaneousAOEnable(self, data):
        return DAQmxSetOnDemandSimultaneousAOEnable(self.task_handle, data)

    def ResetOnDemandSimultaneousAOEnable(self):
        return DAQmxResetOnDemandSimultaneousAOEnable(self.task_handle)

    def GetImplicitUnderflowBehavior(self, data):
        return DAQmxGetImplicitUnderflowBehavior(self.task_handle, data)

    def SetImplicitUnderflowBehavior(self, data):
        return DAQmxSetImplicitUnderflowBehavior(self.task_handle, data)

    def ResetImplicitUnderflowBehavior(self):
        return DAQmxResetImplicitUnderflowBehavior(self.task_handle)

    def GetAIConvRate(self, data):
        return DAQmxGetAIConvRate(self.task_handle, data)

    def SetAIConvRate(self, data):
        return DAQmxSetAIConvRate(self.task_handle, data)

    def ResetAIConvRate(self):
        return DAQmxResetAIConvRate(self.task_handle)

    def GetAIConvRateEx(self, deviceNames, data):
        return DAQmxGetAIConvRateEx(self.task_handle, deviceNames, data)

    def SetAIConvRateEx(self, deviceNames, data):
        return DAQmxSetAIConvRateEx(self.task_handle, deviceNames, data)

    def ResetAIConvRateEx(self, deviceNames):
        return DAQmxResetAIConvRateEx(self.task_handle, deviceNames)

    def GetAIConvMaxRate(self, data):
        return DAQmxGetAIConvMaxRate(self.task_handle, data)

    def GetAIConvMaxRateEx(self, deviceNames, data):
        return DAQmxGetAIConvMaxRateEx(self.task_handle, deviceNames, data)

    def GetAIConvSrc(self, data, bufferSize):
        return DAQmxGetAIConvSrc(self.task_handle, data, bufferSize)

    def SetAIConvSrc(self, data):
        return DAQmxSetAIConvSrc(self.task_handle, data)

    def ResetAIConvSrc(self):
        return DAQmxResetAIConvSrc(self.task_handle)

    def GetAIConvSrcEx(self, deviceNames, data, bufferSize):
        return DAQmxGetAIConvSrcEx(self.task_handle, deviceNames, data, bufferSize)

    def SetAIConvSrcEx(self, deviceNames, data):
        return DAQmxSetAIConvSrcEx(self.task_handle, deviceNames, data)

    def ResetAIConvSrcEx(self, deviceNames):
        return DAQmxResetAIConvSrcEx(self.task_handle, deviceNames)

    def GetAIConvActiveEdge(self, data):
        return DAQmxGetAIConvActiveEdge(self.task_handle, data)

    def SetAIConvActiveEdge(self, data):
        return DAQmxSetAIConvActiveEdge(self.task_handle, data)

    def ResetAIConvActiveEdge(self):
        return DAQmxResetAIConvActiveEdge(self.task_handle)

    def GetAIConvActiveEdgeEx(self, deviceNames, data):
        return DAQmxGetAIConvActiveEdgeEx(self.task_handle, deviceNames, data)

    def SetAIConvActiveEdgeEx(self, deviceNames, data):
        return DAQmxSetAIConvActiveEdgeEx(self.task_handle, deviceNames, data)

    def ResetAIConvActiveEdgeEx(self, deviceNames):
        return DAQmxResetAIConvActiveEdgeEx(self.task_handle, deviceNames)

    def GetAIConvTimebaseDiv(self, data):
        return DAQmxGetAIConvTimebaseDiv(self.task_handle, data)

    def SetAIConvTimebaseDiv(self, data):
        return DAQmxSetAIConvTimebaseDiv(self.task_handle, data)

    def ResetAIConvTimebaseDiv(self):
        return DAQmxResetAIConvTimebaseDiv(self.task_handle)

    def GetAIConvTimebaseDivEx(self, deviceNames, data):
        return DAQmxGetAIConvTimebaseDivEx(self.task_handle, deviceNames, data)

    def SetAIConvTimebaseDivEx(self, deviceNames, data):
        return DAQmxSetAIConvTimebaseDivEx(self.task_handle, deviceNames, data)

    def ResetAIConvTimebaseDivEx(self, deviceNames):
        return DAQmxResetAIConvTimebaseDivEx(self.task_handle, deviceNames)

    def GetAIConvTimebaseSrc(self, data):
        return DAQmxGetAIConvTimebaseSrc(self.task_handle, data)

    def SetAIConvTimebaseSrc(self, data):
        return DAQmxSetAIConvTimebaseSrc(self.task_handle, data)

    def ResetAIConvTimebaseSrc(self):
        return DAQmxResetAIConvTimebaseSrc(self.task_handle)

    def GetAIConvTimebaseSrcEx(self, deviceNames, data):
        return DAQmxGetAIConvTimebaseSrcEx(self.task_handle, deviceNames, data)

    def SetAIConvTimebaseSrcEx(self, deviceNames, data):
        return DAQmxSetAIConvTimebaseSrcEx(self.task_handle, deviceNames, data)

    def ResetAIConvTimebaseSrcEx(self, deviceNames):
        return DAQmxResetAIConvTimebaseSrcEx(self.task_handle, deviceNames)

    def GetDelayFromSampClkDelayUnits(self, data):
        return DAQmxGetDelayFromSampClkDelayUnits(self.task_handle, data)

    def SetDelayFromSampClkDelayUnits(self, data):
        return DAQmxSetDelayFromSampClkDelayUnits(self.task_handle, data)

    def ResetDelayFromSampClkDelayUnits(self):
        return DAQmxResetDelayFromSampClkDelayUnits(self.task_handle)

    def GetDelayFromSampClkDelayUnitsEx(self, deviceNames, data):
        return DAQmxGetDelayFromSampClkDelayUnitsEx(self.task_handle, deviceNames, data)

    def SetDelayFromSampClkDelayUnitsEx(self, deviceNames, data):
        return DAQmxSetDelayFromSampClkDelayUnitsEx(self.task_handle, deviceNames, data)

    def ResetDelayFromSampClkDelayUnitsEx(self, deviceNames):
        return DAQmxResetDelayFromSampClkDelayUnitsEx(self.task_handle, deviceNames)

    def GetDelayFromSampClkDelay(self, data):
        return DAQmxGetDelayFromSampClkDelay(self.task_handle, data)

    def SetDelayFromSampClkDelay(self, data):
        return DAQmxSetDelayFromSampClkDelay(self.task_handle, data)

    def ResetDelayFromSampClkDelay(self):
        return DAQmxResetDelayFromSampClkDelay(self.task_handle)

    def GetDelayFromSampClkDelayEx(self, deviceNames, data):
        return DAQmxGetDelayFromSampClkDelayEx(self.task_handle, deviceNames, data)

    def SetDelayFromSampClkDelayEx(self, deviceNames, data):
        return DAQmxSetDelayFromSampClkDelayEx(self.task_handle, deviceNames, data)

    def ResetDelayFromSampClkDelayEx(self, deviceNames):
        return DAQmxResetDelayFromSampClkDelayEx(self.task_handle, deviceNames)

    def GetAIConvDigFltrEnable(self, data):
        return DAQmxGetAIConvDigFltrEnable(self.task_handle, data)

    def SetAIConvDigFltrEnable(self, data):
        return DAQmxSetAIConvDigFltrEnable(self.task_handle, data)

    def ResetAIConvDigFltrEnable(self):
        return DAQmxResetAIConvDigFltrEnable(self.task_handle)

    def GetAIConvDigFltrEnableEx(self, deviceNames, data):
        return DAQmxGetAIConvDigFltrEnableEx(self.task_handle, deviceNames, data)

    def SetAIConvDigFltrEnableEx(self, deviceNames, data):
        return DAQmxSetAIConvDigFltrEnableEx(self.task_handle, deviceNames, data)

    def ResetAIConvDigFltrEnableEx(self, deviceNames):
        return DAQmxResetAIConvDigFltrEnableEx(self.task_handle, deviceNames)

    def GetAIConvDigFltrMinPulseWidth(self, data):
        return DAQmxGetAIConvDigFltrMinPulseWidth(self.task_handle, data)

    def SetAIConvDigFltrMinPulseWidth(self, data):
        return DAQmxSetAIConvDigFltrMinPulseWidth(self.task_handle, data)

    def ResetAIConvDigFltrMinPulseWidth(self):
        return DAQmxResetAIConvDigFltrMinPulseWidth(self.task_handle)

    def GetAIConvDigFltrMinPulseWidthEx(self, deviceNames, data):
        return DAQmxGetAIConvDigFltrMinPulseWidthEx(self.task_handle, deviceNames, data)

    def SetAIConvDigFltrMinPulseWidthEx(self, deviceNames, data):
        return DAQmxSetAIConvDigFltrMinPulseWidthEx(self.task_handle, deviceNames, data)

    def ResetAIConvDigFltrMinPulseWidthEx(self, deviceNames):
        return DAQmxResetAIConvDigFltrMinPulseWidthEx(self.task_handle, deviceNames)

    def GetAIConvDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAIConvDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetAIConvDigFltrTimebaseSrc(self, data):
        return DAQmxSetAIConvDigFltrTimebaseSrc(self.task_handle, data)

    def ResetAIConvDigFltrTimebaseSrc(self):
        return DAQmxResetAIConvDigFltrTimebaseSrc(self.task_handle)

    def GetAIConvDigFltrTimebaseSrcEx(self, deviceNames, data, bufferSize):
        return DAQmxGetAIConvDigFltrTimebaseSrcEx(self.task_handle, deviceNames, data, bufferSize)

    def SetAIConvDigFltrTimebaseSrcEx(self, deviceNames, data):
        return DAQmxSetAIConvDigFltrTimebaseSrcEx(self.task_handle, deviceNames, data)

    def ResetAIConvDigFltrTimebaseSrcEx(self, deviceNames):
        return DAQmxResetAIConvDigFltrTimebaseSrcEx(self.task_handle, deviceNames)

    def GetAIConvDigFltrTimebaseRate(self, data):
        return DAQmxGetAIConvDigFltrTimebaseRate(self.task_handle, data)

    def SetAIConvDigFltrTimebaseRate(self, data):
        return DAQmxSetAIConvDigFltrTimebaseRate(self.task_handle, data)

    def ResetAIConvDigFltrTimebaseRate(self):
        return DAQmxResetAIConvDigFltrTimebaseRate(self.task_handle)

    def GetAIConvDigFltrTimebaseRateEx(self, deviceNames, data):
        return DAQmxGetAIConvDigFltrTimebaseRateEx(self.task_handle, deviceNames, data)

    def SetAIConvDigFltrTimebaseRateEx(self, deviceNames, data):
        return DAQmxSetAIConvDigFltrTimebaseRateEx(self.task_handle, deviceNames, data)

    def ResetAIConvDigFltrTimebaseRateEx(self, deviceNames):
        return DAQmxResetAIConvDigFltrTimebaseRateEx(self.task_handle, deviceNames)

    def GetAIConvDigSyncEnable(self, data):
        return DAQmxGetAIConvDigSyncEnable(self.task_handle, data)

    def SetAIConvDigSyncEnable(self, data):
        return DAQmxSetAIConvDigSyncEnable(self.task_handle, data)

    def ResetAIConvDigSyncEnable(self):
        return DAQmxResetAIConvDigSyncEnable(self.task_handle)

    def GetAIConvDigSyncEnableEx(self, deviceNames, data):
        return DAQmxGetAIConvDigSyncEnableEx(self.task_handle, deviceNames, data)

    def SetAIConvDigSyncEnableEx(self, deviceNames, data):
        return DAQmxSetAIConvDigSyncEnableEx(self.task_handle, deviceNames, data)

    def ResetAIConvDigSyncEnableEx(self, deviceNames):
        return DAQmxResetAIConvDigSyncEnableEx(self.task_handle, deviceNames)

    def GetMasterTimebaseRate(self, data):
        return DAQmxGetMasterTimebaseRate(self.task_handle, data)

    def SetMasterTimebaseRate(self, data):
        return DAQmxSetMasterTimebaseRate(self.task_handle, data)

    def ResetMasterTimebaseRate(self):
        return DAQmxResetMasterTimebaseRate(self.task_handle)

    def GetMasterTimebaseSrc(self, data, bufferSize):
        return DAQmxGetMasterTimebaseSrc(self.task_handle, data, bufferSize)

    def SetMasterTimebaseSrc(self, data):
        return DAQmxSetMasterTimebaseSrc(self.task_handle, data)

    def ResetMasterTimebaseSrc(self):
        return DAQmxResetMasterTimebaseSrc(self.task_handle)

    def GetRefClkRate(self, data):
        return DAQmxGetRefClkRate(self.task_handle, data)

    def SetRefClkRate(self, data):
        return DAQmxSetRefClkRate(self.task_handle, data)

    def ResetRefClkRate(self):
        return DAQmxResetRefClkRate(self.task_handle)

    def GetRefClkSrc(self, data, bufferSize):
        return DAQmxGetRefClkSrc(self.task_handle, data, bufferSize)

    def SetRefClkSrc(self, data):
        return DAQmxSetRefClkSrc(self.task_handle, data)

    def ResetRefClkSrc(self):
        return DAQmxResetRefClkSrc(self.task_handle)

    def GetSyncPulseSrc(self, data, bufferSize):
        return DAQmxGetSyncPulseSrc(self.task_handle, data, bufferSize)

    def SetSyncPulseSrc(self, data):
        return DAQmxSetSyncPulseSrc(self.task_handle, data)

    def ResetSyncPulseSrc(self):
        return DAQmxResetSyncPulseSrc(self.task_handle)

    def GetSyncPulseSyncTime(self, data):
        return DAQmxGetSyncPulseSyncTime(self.task_handle, data)

    def GetSyncPulseMinDelayToStart(self, data):
        return DAQmxGetSyncPulseMinDelayToStart(self.task_handle, data)

    def SetSyncPulseMinDelayToStart(self, data):
        return DAQmxSetSyncPulseMinDelayToStart(self.task_handle, data)

    def ResetSyncPulseMinDelayToStart(self):
        return DAQmxResetSyncPulseMinDelayToStart(self.task_handle)

    def GetSyncPulseResetTime(self, data):
        return DAQmxGetSyncPulseResetTime(self.task_handle, data)

    def GetSyncPulseResetDelay(self, data):
        return DAQmxGetSyncPulseResetDelay(self.task_handle, data)

    def SetSyncPulseResetDelay(self, data):
        return DAQmxSetSyncPulseResetDelay(self.task_handle, data)

    def ResetSyncPulseResetDelay(self):
        return DAQmxResetSyncPulseResetDelay(self.task_handle)

    def GetSyncPulseTerm(self, data, bufferSize):
        return DAQmxGetSyncPulseTerm(self.task_handle, data, bufferSize)

    def GetSyncClkInterval(self, data):
        return DAQmxGetSyncClkInterval(self.task_handle, data)

    def SetSyncClkInterval(self, data):
        return DAQmxSetSyncClkInterval(self.task_handle, data)

    def ResetSyncClkInterval(self):
        return DAQmxResetSyncClkInterval(self.task_handle)

    def GetSampTimingEngine(self, data):
        return DAQmxGetSampTimingEngine(self.task_handle, data)

    def SetSampTimingEngine(self, data):
        return DAQmxSetSampTimingEngine(self.task_handle, data)

    def ResetSampTimingEngine(self):
        return DAQmxResetSampTimingEngine(self.task_handle)

    def GetStartTrigType(self, data):
        return DAQmxGetStartTrigType(self.task_handle, data)

    def SetStartTrigType(self, data):
        return DAQmxSetStartTrigType(self.task_handle, data)

    def ResetStartTrigType(self):
        return DAQmxResetStartTrigType(self.task_handle)

    def GetStartTrigTerm(self, data, bufferSize):
        return DAQmxGetStartTrigTerm(self.task_handle, data, bufferSize)

    def GetDigEdgeStartTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeStartTrigSrc(self.task_handle, data, bufferSize)

    def SetDigEdgeStartTrigSrc(self, data):
        return DAQmxSetDigEdgeStartTrigSrc(self.task_handle, data)

    def ResetDigEdgeStartTrigSrc(self):
        return DAQmxResetDigEdgeStartTrigSrc(self.task_handle)

    def GetDigEdgeStartTrigEdge(self, data):
        return DAQmxGetDigEdgeStartTrigEdge(self.task_handle, data)

    def SetDigEdgeStartTrigEdge(self, data):
        return DAQmxSetDigEdgeStartTrigEdge(self.task_handle, data)

    def ResetDigEdgeStartTrigEdge(self):
        return DAQmxResetDigEdgeStartTrigEdge(self.task_handle)

    def GetDigEdgeStartTrigDigFltrEnable(self, data):
        return DAQmxGetDigEdgeStartTrigDigFltrEnable(self.task_handle, data)

    def SetDigEdgeStartTrigDigFltrEnable(self, data):
        return DAQmxSetDigEdgeStartTrigDigFltrEnable(self.task_handle, data)

    def ResetDigEdgeStartTrigDigFltrEnable(self):
        return DAQmxResetDigEdgeStartTrigDigFltrEnable(self.task_handle)

    def GetDigEdgeStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetDigEdgeStartTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetDigEdgeStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetDigEdgeStartTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetDigEdgeStartTrigDigFltrMinPulseWidth(self):
        return DAQmxResetDigEdgeStartTrigDigFltrMinPulseWidth(self.task_handle)

    def GetDigEdgeStartTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeStartTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetDigEdgeStartTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetDigEdgeStartTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetDigEdgeStartTrigDigFltrTimebaseSrc(self):
        return DAQmxResetDigEdgeStartTrigDigFltrTimebaseSrc(self.task_handle)

    def GetDigEdgeStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetDigEdgeStartTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetDigEdgeStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetDigEdgeStartTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetDigEdgeStartTrigDigFltrTimebaseRate(self):
        return DAQmxResetDigEdgeStartTrigDigFltrTimebaseRate(self.task_handle)

    def GetDigEdgeStartTrigDigSyncEnable(self, data):
        return DAQmxGetDigEdgeStartTrigDigSyncEnable(self.task_handle, data)

    def SetDigEdgeStartTrigDigSyncEnable(self, data):
        return DAQmxSetDigEdgeStartTrigDigSyncEnable(self.task_handle, data)

    def ResetDigEdgeStartTrigDigSyncEnable(self):
        return DAQmxResetDigEdgeStartTrigDigSyncEnable(self.task_handle)

    def GetDigPatternStartTrigSrc(self, data, bufferSize):
        return DAQmxGetDigPatternStartTrigSrc(self.task_handle, data, bufferSize)

    def SetDigPatternStartTrigSrc(self, data):
        return DAQmxSetDigPatternStartTrigSrc(self.task_handle, data)

    def ResetDigPatternStartTrigSrc(self):
        return DAQmxResetDigPatternStartTrigSrc(self.task_handle)

    def GetDigPatternStartTrigPattern(self, data, bufferSize):
        return DAQmxGetDigPatternStartTrigPattern(self.task_handle, data, bufferSize)

    def SetDigPatternStartTrigPattern(self, data):
        return DAQmxSetDigPatternStartTrigPattern(self.task_handle, data)

    def ResetDigPatternStartTrigPattern(self):
        return DAQmxResetDigPatternStartTrigPattern(self.task_handle)

    def GetDigPatternStartTrigWhen(self, data):
        return DAQmxGetDigPatternStartTrigWhen(self.task_handle, data)

    def SetDigPatternStartTrigWhen(self, data):
        return DAQmxSetDigPatternStartTrigWhen(self.task_handle, data)

    def ResetDigPatternStartTrigWhen(self):
        return DAQmxResetDigPatternStartTrigWhen(self.task_handle)

    def GetAnlgEdgeStartTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgEdgeStartTrigSrc(self.task_handle, data, bufferSize)

    def SetAnlgEdgeStartTrigSrc(self, data):
        return DAQmxSetAnlgEdgeStartTrigSrc(self.task_handle, data)

    def ResetAnlgEdgeStartTrigSrc(self):
        return DAQmxResetAnlgEdgeStartTrigSrc(self.task_handle)

    def GetAnlgEdgeStartTrigSlope(self, data):
        return DAQmxGetAnlgEdgeStartTrigSlope(self.task_handle, data)

    def SetAnlgEdgeStartTrigSlope(self, data):
        return DAQmxSetAnlgEdgeStartTrigSlope(self.task_handle, data)

    def ResetAnlgEdgeStartTrigSlope(self):
        return DAQmxResetAnlgEdgeStartTrigSlope(self.task_handle)

    def GetAnlgEdgeStartTrigLvl(self, data):
        return DAQmxGetAnlgEdgeStartTrigLvl(self.task_handle, data)

    def SetAnlgEdgeStartTrigLvl(self, data):
        return DAQmxSetAnlgEdgeStartTrigLvl(self.task_handle, data)

    def ResetAnlgEdgeStartTrigLvl(self):
        return DAQmxResetAnlgEdgeStartTrigLvl(self.task_handle)

    def GetAnlgEdgeStartTrigHyst(self, data):
        return DAQmxGetAnlgEdgeStartTrigHyst(self.task_handle, data)

    def SetAnlgEdgeStartTrigHyst(self, data):
        return DAQmxSetAnlgEdgeStartTrigHyst(self.task_handle, data)

    def ResetAnlgEdgeStartTrigHyst(self):
        return DAQmxResetAnlgEdgeStartTrigHyst(self.task_handle)

    def GetAnlgEdgeStartTrigCoupling(self, data):
        return DAQmxGetAnlgEdgeStartTrigCoupling(self.task_handle, data)

    def SetAnlgEdgeStartTrigCoupling(self, data):
        return DAQmxSetAnlgEdgeStartTrigCoupling(self.task_handle, data)

    def ResetAnlgEdgeStartTrigCoupling(self):
        return DAQmxResetAnlgEdgeStartTrigCoupling(self.task_handle)

    def GetAnlgEdgeStartTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgEdgeStartTrigDigFltrEnable(self.task_handle, data)

    def SetAnlgEdgeStartTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigFltrEnable(self.task_handle, data)

    def ResetAnlgEdgeStartTrigDigFltrEnable(self):
        return DAQmxResetAnlgEdgeStartTrigDigFltrEnable(self.task_handle)

    def GetAnlgEdgeStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgEdgeStartTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetAnlgEdgeStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetAnlgEdgeStartTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgEdgeStartTrigDigFltrMinPulseWidth(self.task_handle)

    def GetAnlgEdgeStartTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetAnlgEdgeStartTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetAnlgEdgeStartTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseSrc(self.task_handle)

    def GetAnlgEdgeStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgEdgeStartTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetAnlgEdgeStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetAnlgEdgeStartTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgEdgeStartTrigDigFltrTimebaseRate(self.task_handle)

    def GetAnlgEdgeStartTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgEdgeStartTrigDigSyncEnable(self.task_handle, data)

    def SetAnlgEdgeStartTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgEdgeStartTrigDigSyncEnable(self.task_handle, data)

    def ResetAnlgEdgeStartTrigDigSyncEnable(self):
        return DAQmxResetAnlgEdgeStartTrigDigSyncEnable(self.task_handle)

    def GetAnlgWinStartTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinStartTrigSrc(self.task_handle, data, bufferSize)

    def SetAnlgWinStartTrigSrc(self, data):
        return DAQmxSetAnlgWinStartTrigSrc(self.task_handle, data)

    def ResetAnlgWinStartTrigSrc(self):
        return DAQmxResetAnlgWinStartTrigSrc(self.task_handle)

    def GetAnlgWinStartTrigWhen(self, data):
        return DAQmxGetAnlgWinStartTrigWhen(self.task_handle, data)

    def SetAnlgWinStartTrigWhen(self, data):
        return DAQmxSetAnlgWinStartTrigWhen(self.task_handle, data)

    def ResetAnlgWinStartTrigWhen(self):
        return DAQmxResetAnlgWinStartTrigWhen(self.task_handle)

    def GetAnlgWinStartTrigTop(self, data):
        return DAQmxGetAnlgWinStartTrigTop(self.task_handle, data)

    def SetAnlgWinStartTrigTop(self, data):
        return DAQmxSetAnlgWinStartTrigTop(self.task_handle, data)

    def ResetAnlgWinStartTrigTop(self):
        return DAQmxResetAnlgWinStartTrigTop(self.task_handle)

    def GetAnlgWinStartTrigBtm(self, data):
        return DAQmxGetAnlgWinStartTrigBtm(self.task_handle, data)

    def SetAnlgWinStartTrigBtm(self, data):
        return DAQmxSetAnlgWinStartTrigBtm(self.task_handle, data)

    def ResetAnlgWinStartTrigBtm(self):
        return DAQmxResetAnlgWinStartTrigBtm(self.task_handle)

    def GetAnlgWinStartTrigCoupling(self, data):
        return DAQmxGetAnlgWinStartTrigCoupling(self.task_handle, data)

    def SetAnlgWinStartTrigCoupling(self, data):
        return DAQmxSetAnlgWinStartTrigCoupling(self.task_handle, data)

    def ResetAnlgWinStartTrigCoupling(self):
        return DAQmxResetAnlgWinStartTrigCoupling(self.task_handle)

    def GetAnlgWinStartTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgWinStartTrigDigFltrEnable(self.task_handle, data)

    def SetAnlgWinStartTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgWinStartTrigDigFltrEnable(self.task_handle, data)

    def ResetAnlgWinStartTrigDigFltrEnable(self):
        return DAQmxResetAnlgWinStartTrigDigFltrEnable(self.task_handle)

    def GetAnlgWinStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgWinStartTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetAnlgWinStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgWinStartTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetAnlgWinStartTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgWinStartTrigDigFltrMinPulseWidth(self.task_handle)

    def GetAnlgWinStartTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinStartTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetAnlgWinStartTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgWinStartTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetAnlgWinStartTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgWinStartTrigDigFltrTimebaseSrc(self.task_handle)

    def GetAnlgWinStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgWinStartTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetAnlgWinStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgWinStartTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetAnlgWinStartTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgWinStartTrigDigFltrTimebaseRate(self.task_handle)

    def GetAnlgWinStartTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgWinStartTrigDigSyncEnable(self.task_handle, data)

    def SetAnlgWinStartTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgWinStartTrigDigSyncEnable(self.task_handle, data)

    def ResetAnlgWinStartTrigDigSyncEnable(self):
        return DAQmxResetAnlgWinStartTrigDigSyncEnable(self.task_handle)

    def GetStartTrigDelay(self, data):
        return DAQmxGetStartTrigDelay(self.task_handle, data)

    def SetStartTrigDelay(self, data):
        return DAQmxSetStartTrigDelay(self.task_handle, data)

    def ResetStartTrigDelay(self):
        return DAQmxResetStartTrigDelay(self.task_handle)

    def GetStartTrigDelayUnits(self, data):
        return DAQmxGetStartTrigDelayUnits(self.task_handle, data)

    def SetStartTrigDelayUnits(self, data):
        return DAQmxSetStartTrigDelayUnits(self.task_handle, data)

    def ResetStartTrigDelayUnits(self):
        return DAQmxResetStartTrigDelayUnits(self.task_handle)

    def GetStartTrigRetriggerable(self, data):
        return DAQmxGetStartTrigRetriggerable(self.task_handle, data)

    def SetStartTrigRetriggerable(self, data):
        return DAQmxSetStartTrigRetriggerable(self.task_handle, data)

    def ResetStartTrigRetriggerable(self):
        return DAQmxResetStartTrigRetriggerable(self.task_handle)

    def GetRefTrigType(self, data):
        return DAQmxGetRefTrigType(self.task_handle, data)

    def SetRefTrigType(self, data):
        return DAQmxSetRefTrigType(self.task_handle, data)

    def ResetRefTrigType(self):
        return DAQmxResetRefTrigType(self.task_handle)

    def GetRefTrigPretrigSamples(self, data):
        return DAQmxGetRefTrigPretrigSamples(self.task_handle, data)

    def SetRefTrigPretrigSamples(self, data):
        return DAQmxSetRefTrigPretrigSamples(self.task_handle, data)

    def ResetRefTrigPretrigSamples(self):
        return DAQmxResetRefTrigPretrigSamples(self.task_handle)

    def GetRefTrigTerm(self, data, bufferSize):
        return DAQmxGetRefTrigTerm(self.task_handle, data, bufferSize)

    def GetDigEdgeRefTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeRefTrigSrc(self.task_handle, data, bufferSize)

    def SetDigEdgeRefTrigSrc(self, data):
        return DAQmxSetDigEdgeRefTrigSrc(self.task_handle, data)

    def ResetDigEdgeRefTrigSrc(self):
        return DAQmxResetDigEdgeRefTrigSrc(self.task_handle)

    def GetDigEdgeRefTrigEdge(self, data):
        return DAQmxGetDigEdgeRefTrigEdge(self.task_handle, data)

    def SetDigEdgeRefTrigEdge(self, data):
        return DAQmxSetDigEdgeRefTrigEdge(self.task_handle, data)

    def ResetDigEdgeRefTrigEdge(self):
        return DAQmxResetDigEdgeRefTrigEdge(self.task_handle)

    def GetDigEdgeRefTrigDigFltrEnable(self, data):
        return DAQmxGetDigEdgeRefTrigDigFltrEnable(self.task_handle, data)

    def SetDigEdgeRefTrigDigFltrEnable(self, data):
        return DAQmxSetDigEdgeRefTrigDigFltrEnable(self.task_handle, data)

    def ResetDigEdgeRefTrigDigFltrEnable(self):
        return DAQmxResetDigEdgeRefTrigDigFltrEnable(self.task_handle)

    def GetDigEdgeRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetDigEdgeRefTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetDigEdgeRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetDigEdgeRefTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetDigEdgeRefTrigDigFltrMinPulseWidth(self):
        return DAQmxResetDigEdgeRefTrigDigFltrMinPulseWidth(self.task_handle)

    def GetDigEdgeRefTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeRefTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetDigEdgeRefTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetDigEdgeRefTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetDigEdgeRefTrigDigFltrTimebaseSrc(self):
        return DAQmxResetDigEdgeRefTrigDigFltrTimebaseSrc(self.task_handle)

    def GetDigEdgeRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetDigEdgeRefTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetDigEdgeRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetDigEdgeRefTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetDigEdgeRefTrigDigFltrTimebaseRate(self):
        return DAQmxResetDigEdgeRefTrigDigFltrTimebaseRate(self.task_handle)

    def GetDigEdgeRefTrigDigSyncEnable(self, data):
        return DAQmxGetDigEdgeRefTrigDigSyncEnable(self.task_handle, data)

    def SetDigEdgeRefTrigDigSyncEnable(self, data):
        return DAQmxSetDigEdgeRefTrigDigSyncEnable(self.task_handle, data)

    def ResetDigEdgeRefTrigDigSyncEnable(self):
        return DAQmxResetDigEdgeRefTrigDigSyncEnable(self.task_handle)

    def GetDigPatternRefTrigSrc(self, data, bufferSize):
        return DAQmxGetDigPatternRefTrigSrc(self.task_handle, data, bufferSize)

    def SetDigPatternRefTrigSrc(self, data):
        return DAQmxSetDigPatternRefTrigSrc(self.task_handle, data)

    def ResetDigPatternRefTrigSrc(self):
        return DAQmxResetDigPatternRefTrigSrc(self.task_handle)

    def GetDigPatternRefTrigPattern(self, data, bufferSize):
        return DAQmxGetDigPatternRefTrigPattern(self.task_handle, data, bufferSize)

    def SetDigPatternRefTrigPattern(self, data):
        return DAQmxSetDigPatternRefTrigPattern(self.task_handle, data)

    def ResetDigPatternRefTrigPattern(self):
        return DAQmxResetDigPatternRefTrigPattern(self.task_handle)

    def GetDigPatternRefTrigWhen(self, data):
        return DAQmxGetDigPatternRefTrigWhen(self.task_handle, data)

    def SetDigPatternRefTrigWhen(self, data):
        return DAQmxSetDigPatternRefTrigWhen(self.task_handle, data)

    def ResetDigPatternRefTrigWhen(self):
        return DAQmxResetDigPatternRefTrigWhen(self.task_handle)

    def GetAnlgEdgeRefTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgEdgeRefTrigSrc(self.task_handle, data, bufferSize)

    def SetAnlgEdgeRefTrigSrc(self, data):
        return DAQmxSetAnlgEdgeRefTrigSrc(self.task_handle, data)

    def ResetAnlgEdgeRefTrigSrc(self):
        return DAQmxResetAnlgEdgeRefTrigSrc(self.task_handle)

    def GetAnlgEdgeRefTrigSlope(self, data):
        return DAQmxGetAnlgEdgeRefTrigSlope(self.task_handle, data)

    def SetAnlgEdgeRefTrigSlope(self, data):
        return DAQmxSetAnlgEdgeRefTrigSlope(self.task_handle, data)

    def ResetAnlgEdgeRefTrigSlope(self):
        return DAQmxResetAnlgEdgeRefTrigSlope(self.task_handle)

    def GetAnlgEdgeRefTrigLvl(self, data):
        return DAQmxGetAnlgEdgeRefTrigLvl(self.task_handle, data)

    def SetAnlgEdgeRefTrigLvl(self, data):
        return DAQmxSetAnlgEdgeRefTrigLvl(self.task_handle, data)

    def ResetAnlgEdgeRefTrigLvl(self):
        return DAQmxResetAnlgEdgeRefTrigLvl(self.task_handle)

    def GetAnlgEdgeRefTrigHyst(self, data):
        return DAQmxGetAnlgEdgeRefTrigHyst(self.task_handle, data)

    def SetAnlgEdgeRefTrigHyst(self, data):
        return DAQmxSetAnlgEdgeRefTrigHyst(self.task_handle, data)

    def ResetAnlgEdgeRefTrigHyst(self):
        return DAQmxResetAnlgEdgeRefTrigHyst(self.task_handle)

    def GetAnlgEdgeRefTrigCoupling(self, data):
        return DAQmxGetAnlgEdgeRefTrigCoupling(self.task_handle, data)

    def SetAnlgEdgeRefTrigCoupling(self, data):
        return DAQmxSetAnlgEdgeRefTrigCoupling(self.task_handle, data)

    def ResetAnlgEdgeRefTrigCoupling(self):
        return DAQmxResetAnlgEdgeRefTrigCoupling(self.task_handle)

    def GetAnlgEdgeRefTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgEdgeRefTrigDigFltrEnable(self.task_handle, data)

    def SetAnlgEdgeRefTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigFltrEnable(self.task_handle, data)

    def ResetAnlgEdgeRefTrigDigFltrEnable(self):
        return DAQmxResetAnlgEdgeRefTrigDigFltrEnable(self.task_handle)

    def GetAnlgEdgeRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgEdgeRefTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetAnlgEdgeRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetAnlgEdgeRefTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgEdgeRefTrigDigFltrMinPulseWidth(self.task_handle)

    def GetAnlgEdgeRefTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetAnlgEdgeRefTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetAnlgEdgeRefTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseSrc(self.task_handle)

    def GetAnlgEdgeRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgEdgeRefTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetAnlgEdgeRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetAnlgEdgeRefTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgEdgeRefTrigDigFltrTimebaseRate(self.task_handle)

    def GetAnlgEdgeRefTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgEdgeRefTrigDigSyncEnable(self.task_handle, data)

    def SetAnlgEdgeRefTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgEdgeRefTrigDigSyncEnable(self.task_handle, data)

    def ResetAnlgEdgeRefTrigDigSyncEnable(self):
        return DAQmxResetAnlgEdgeRefTrigDigSyncEnable(self.task_handle)

    def GetAnlgWinRefTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinRefTrigSrc(self.task_handle, data, bufferSize)

    def SetAnlgWinRefTrigSrc(self, data):
        return DAQmxSetAnlgWinRefTrigSrc(self.task_handle, data)

    def ResetAnlgWinRefTrigSrc(self):
        return DAQmxResetAnlgWinRefTrigSrc(self.task_handle)

    def GetAnlgWinRefTrigWhen(self, data):
        return DAQmxGetAnlgWinRefTrigWhen(self.task_handle, data)

    def SetAnlgWinRefTrigWhen(self, data):
        return DAQmxSetAnlgWinRefTrigWhen(self.task_handle, data)

    def ResetAnlgWinRefTrigWhen(self):
        return DAQmxResetAnlgWinRefTrigWhen(self.task_handle)

    def GetAnlgWinRefTrigTop(self, data):
        return DAQmxGetAnlgWinRefTrigTop(self.task_handle, data)

    def SetAnlgWinRefTrigTop(self, data):
        return DAQmxSetAnlgWinRefTrigTop(self.task_handle, data)

    def ResetAnlgWinRefTrigTop(self):
        return DAQmxResetAnlgWinRefTrigTop(self.task_handle)

    def GetAnlgWinRefTrigBtm(self, data):
        return DAQmxGetAnlgWinRefTrigBtm(self.task_handle, data)

    def SetAnlgWinRefTrigBtm(self, data):
        return DAQmxSetAnlgWinRefTrigBtm(self.task_handle, data)

    def ResetAnlgWinRefTrigBtm(self):
        return DAQmxResetAnlgWinRefTrigBtm(self.task_handle)

    def GetAnlgWinRefTrigCoupling(self, data):
        return DAQmxGetAnlgWinRefTrigCoupling(self.task_handle, data)

    def SetAnlgWinRefTrigCoupling(self, data):
        return DAQmxSetAnlgWinRefTrigCoupling(self.task_handle, data)

    def ResetAnlgWinRefTrigCoupling(self):
        return DAQmxResetAnlgWinRefTrigCoupling(self.task_handle)

    def GetAnlgWinRefTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgWinRefTrigDigFltrEnable(self.task_handle, data)

    def SetAnlgWinRefTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgWinRefTrigDigFltrEnable(self.task_handle, data)

    def ResetAnlgWinRefTrigDigFltrEnable(self):
        return DAQmxResetAnlgWinRefTrigDigFltrEnable(self.task_handle)

    def GetAnlgWinRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgWinRefTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetAnlgWinRefTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgWinRefTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetAnlgWinRefTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgWinRefTrigDigFltrMinPulseWidth(self.task_handle)

    def GetAnlgWinRefTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinRefTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetAnlgWinRefTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgWinRefTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetAnlgWinRefTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgWinRefTrigDigFltrTimebaseSrc(self.task_handle)

    def GetAnlgWinRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgWinRefTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetAnlgWinRefTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgWinRefTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetAnlgWinRefTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgWinRefTrigDigFltrTimebaseRate(self.task_handle)

    def GetAnlgWinRefTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgWinRefTrigDigSyncEnable(self.task_handle, data)

    def SetAnlgWinRefTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgWinRefTrigDigSyncEnable(self.task_handle, data)

    def ResetAnlgWinRefTrigDigSyncEnable(self):
        return DAQmxResetAnlgWinRefTrigDigSyncEnable(self.task_handle)

    def GetRefTrigAutoTrigEnable(self, data):
        return DAQmxGetRefTrigAutoTrigEnable(self.task_handle, data)

    def SetRefTrigAutoTrigEnable(self, data):
        return DAQmxSetRefTrigAutoTrigEnable(self.task_handle, data)

    def ResetRefTrigAutoTrigEnable(self):
        return DAQmxResetRefTrigAutoTrigEnable(self.task_handle)

    def GetRefTrigAutoTriggered(self, data):
        return DAQmxGetRefTrigAutoTriggered(self.task_handle, data)

    def GetRefTrigDelay(self, data):
        return DAQmxGetRefTrigDelay(self.task_handle, data)

    def SetRefTrigDelay(self, data):
        return DAQmxSetRefTrigDelay(self.task_handle, data)

    def ResetRefTrigDelay(self):
        return DAQmxResetRefTrigDelay(self.task_handle)

    def GetAdvTrigType(self, data):
        return DAQmxGetAdvTrigType(self.task_handle, data)

    def SetAdvTrigType(self, data):
        return DAQmxSetAdvTrigType(self.task_handle, data)

    def ResetAdvTrigType(self):
        return DAQmxResetAdvTrigType(self.task_handle)

    def GetDigEdgeAdvTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeAdvTrigSrc(self.task_handle, data, bufferSize)

    def SetDigEdgeAdvTrigSrc(self, data):
        return DAQmxSetDigEdgeAdvTrigSrc(self.task_handle, data)

    def ResetDigEdgeAdvTrigSrc(self):
        return DAQmxResetDigEdgeAdvTrigSrc(self.task_handle)

    def GetDigEdgeAdvTrigEdge(self, data):
        return DAQmxGetDigEdgeAdvTrigEdge(self.task_handle, data)

    def SetDigEdgeAdvTrigEdge(self, data):
        return DAQmxSetDigEdgeAdvTrigEdge(self.task_handle, data)

    def ResetDigEdgeAdvTrigEdge(self):
        return DAQmxResetDigEdgeAdvTrigEdge(self.task_handle)

    def GetDigEdgeAdvTrigDigFltrEnable(self, data):
        return DAQmxGetDigEdgeAdvTrigDigFltrEnable(self.task_handle, data)

    def SetDigEdgeAdvTrigDigFltrEnable(self, data):
        return DAQmxSetDigEdgeAdvTrigDigFltrEnable(self.task_handle, data)

    def ResetDigEdgeAdvTrigDigFltrEnable(self):
        return DAQmxResetDigEdgeAdvTrigDigFltrEnable(self.task_handle)

    def GetHshkTrigType(self, data):
        return DAQmxGetHshkTrigType(self.task_handle, data)

    def SetHshkTrigType(self, data):
        return DAQmxSetHshkTrigType(self.task_handle, data)

    def ResetHshkTrigType(self):
        return DAQmxResetHshkTrigType(self.task_handle)

    def GetInterlockedHshkTrigSrc(self, data, bufferSize):
        return DAQmxGetInterlockedHshkTrigSrc(self.task_handle, data, bufferSize)

    def SetInterlockedHshkTrigSrc(self, data):
        return DAQmxSetInterlockedHshkTrigSrc(self.task_handle, data)

    def ResetInterlockedHshkTrigSrc(self):
        return DAQmxResetInterlockedHshkTrigSrc(self.task_handle)

    def GetInterlockedHshkTrigAssertedLvl(self, data):
        return DAQmxGetInterlockedHshkTrigAssertedLvl(self.task_handle, data)

    def SetInterlockedHshkTrigAssertedLvl(self, data):
        return DAQmxSetInterlockedHshkTrigAssertedLvl(self.task_handle, data)

    def ResetInterlockedHshkTrigAssertedLvl(self):
        return DAQmxResetInterlockedHshkTrigAssertedLvl(self.task_handle)

    def GetPauseTrigType(self, data):
        return DAQmxGetPauseTrigType(self.task_handle, data)

    def SetPauseTrigType(self, data):
        return DAQmxSetPauseTrigType(self.task_handle, data)

    def ResetPauseTrigType(self):
        return DAQmxResetPauseTrigType(self.task_handle)

    def GetPauseTrigTerm(self, data, bufferSize):
        return DAQmxGetPauseTrigTerm(self.task_handle, data, bufferSize)

    def GetAnlgLvlPauseTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgLvlPauseTrigSrc(self.task_handle, data, bufferSize)

    def SetAnlgLvlPauseTrigSrc(self, data):
        return DAQmxSetAnlgLvlPauseTrigSrc(self.task_handle, data)

    def ResetAnlgLvlPauseTrigSrc(self):
        return DAQmxResetAnlgLvlPauseTrigSrc(self.task_handle)

    def GetAnlgLvlPauseTrigWhen(self, data):
        return DAQmxGetAnlgLvlPauseTrigWhen(self.task_handle, data)

    def SetAnlgLvlPauseTrigWhen(self, data):
        return DAQmxSetAnlgLvlPauseTrigWhen(self.task_handle, data)

    def ResetAnlgLvlPauseTrigWhen(self):
        return DAQmxResetAnlgLvlPauseTrigWhen(self.task_handle)

    def GetAnlgLvlPauseTrigLvl(self, data):
        return DAQmxGetAnlgLvlPauseTrigLvl(self.task_handle, data)

    def SetAnlgLvlPauseTrigLvl(self, data):
        return DAQmxSetAnlgLvlPauseTrigLvl(self.task_handle, data)

    def ResetAnlgLvlPauseTrigLvl(self):
        return DAQmxResetAnlgLvlPauseTrigLvl(self.task_handle)

    def GetAnlgLvlPauseTrigHyst(self, data):
        return DAQmxGetAnlgLvlPauseTrigHyst(self.task_handle, data)

    def SetAnlgLvlPauseTrigHyst(self, data):
        return DAQmxSetAnlgLvlPauseTrigHyst(self.task_handle, data)

    def ResetAnlgLvlPauseTrigHyst(self):
        return DAQmxResetAnlgLvlPauseTrigHyst(self.task_handle)

    def GetAnlgLvlPauseTrigCoupling(self, data):
        return DAQmxGetAnlgLvlPauseTrigCoupling(self.task_handle, data)

    def SetAnlgLvlPauseTrigCoupling(self, data):
        return DAQmxSetAnlgLvlPauseTrigCoupling(self.task_handle, data)

    def ResetAnlgLvlPauseTrigCoupling(self):
        return DAQmxResetAnlgLvlPauseTrigCoupling(self.task_handle)

    def GetAnlgLvlPauseTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgLvlPauseTrigDigFltrEnable(self.task_handle, data)

    def SetAnlgLvlPauseTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigFltrEnable(self.task_handle, data)

    def ResetAnlgLvlPauseTrigDigFltrEnable(self):
        return DAQmxResetAnlgLvlPauseTrigDigFltrEnable(self.task_handle)

    def GetAnlgLvlPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgLvlPauseTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetAnlgLvlPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetAnlgLvlPauseTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgLvlPauseTrigDigFltrMinPulseWidth(self.task_handle)

    def GetAnlgLvlPauseTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetAnlgLvlPauseTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetAnlgLvlPauseTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseSrc(self.task_handle)

    def GetAnlgLvlPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgLvlPauseTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetAnlgLvlPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetAnlgLvlPauseTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgLvlPauseTrigDigFltrTimebaseRate(self.task_handle)

    def GetAnlgLvlPauseTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgLvlPauseTrigDigSyncEnable(self.task_handle, data)

    def SetAnlgLvlPauseTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgLvlPauseTrigDigSyncEnable(self.task_handle, data)

    def ResetAnlgLvlPauseTrigDigSyncEnable(self):
        return DAQmxResetAnlgLvlPauseTrigDigSyncEnable(self.task_handle)

    def GetAnlgWinPauseTrigSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinPauseTrigSrc(self.task_handle, data, bufferSize)

    def SetAnlgWinPauseTrigSrc(self, data):
        return DAQmxSetAnlgWinPauseTrigSrc(self.task_handle, data)

    def ResetAnlgWinPauseTrigSrc(self):
        return DAQmxResetAnlgWinPauseTrigSrc(self.task_handle)

    def GetAnlgWinPauseTrigWhen(self, data):
        return DAQmxGetAnlgWinPauseTrigWhen(self.task_handle, data)

    def SetAnlgWinPauseTrigWhen(self, data):
        return DAQmxSetAnlgWinPauseTrigWhen(self.task_handle, data)

    def ResetAnlgWinPauseTrigWhen(self):
        return DAQmxResetAnlgWinPauseTrigWhen(self.task_handle)

    def GetAnlgWinPauseTrigTop(self, data):
        return DAQmxGetAnlgWinPauseTrigTop(self.task_handle, data)

    def SetAnlgWinPauseTrigTop(self, data):
        return DAQmxSetAnlgWinPauseTrigTop(self.task_handle, data)

    def ResetAnlgWinPauseTrigTop(self):
        return DAQmxResetAnlgWinPauseTrigTop(self.task_handle)

    def GetAnlgWinPauseTrigBtm(self, data):
        return DAQmxGetAnlgWinPauseTrigBtm(self.task_handle, data)

    def SetAnlgWinPauseTrigBtm(self, data):
        return DAQmxSetAnlgWinPauseTrigBtm(self.task_handle, data)

    def ResetAnlgWinPauseTrigBtm(self):
        return DAQmxResetAnlgWinPauseTrigBtm(self.task_handle)

    def GetAnlgWinPauseTrigCoupling(self, data):
        return DAQmxGetAnlgWinPauseTrigCoupling(self.task_handle, data)

    def SetAnlgWinPauseTrigCoupling(self, data):
        return DAQmxSetAnlgWinPauseTrigCoupling(self.task_handle, data)

    def ResetAnlgWinPauseTrigCoupling(self):
        return DAQmxResetAnlgWinPauseTrigCoupling(self.task_handle)

    def GetAnlgWinPauseTrigDigFltrEnable(self, data):
        return DAQmxGetAnlgWinPauseTrigDigFltrEnable(self.task_handle, data)

    def SetAnlgWinPauseTrigDigFltrEnable(self, data):
        return DAQmxSetAnlgWinPauseTrigDigFltrEnable(self.task_handle, data)

    def ResetAnlgWinPauseTrigDigFltrEnable(self):
        return DAQmxResetAnlgWinPauseTrigDigFltrEnable(self.task_handle)

    def GetAnlgWinPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetAnlgWinPauseTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetAnlgWinPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetAnlgWinPauseTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetAnlgWinPauseTrigDigFltrMinPulseWidth(self):
        return DAQmxResetAnlgWinPauseTrigDigFltrMinPulseWidth(self.task_handle)

    def GetAnlgWinPauseTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetAnlgWinPauseTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetAnlgWinPauseTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetAnlgWinPauseTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetAnlgWinPauseTrigDigFltrTimebaseSrc(self):
        return DAQmxResetAnlgWinPauseTrigDigFltrTimebaseSrc(self.task_handle)

    def GetAnlgWinPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetAnlgWinPauseTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetAnlgWinPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetAnlgWinPauseTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetAnlgWinPauseTrigDigFltrTimebaseRate(self):
        return DAQmxResetAnlgWinPauseTrigDigFltrTimebaseRate(self.task_handle)

    def GetAnlgWinPauseTrigDigSyncEnable(self, data):
        return DAQmxGetAnlgWinPauseTrigDigSyncEnable(self.task_handle, data)

    def SetAnlgWinPauseTrigDigSyncEnable(self, data):
        return DAQmxSetAnlgWinPauseTrigDigSyncEnable(self.task_handle, data)

    def ResetAnlgWinPauseTrigDigSyncEnable(self):
        return DAQmxResetAnlgWinPauseTrigDigSyncEnable(self.task_handle)

    def GetDigLvlPauseTrigSrc(self, data, bufferSize):
        return DAQmxGetDigLvlPauseTrigSrc(self.task_handle, data, bufferSize)

    def SetDigLvlPauseTrigSrc(self, data):
        return DAQmxSetDigLvlPauseTrigSrc(self.task_handle, data)

    def ResetDigLvlPauseTrigSrc(self):
        return DAQmxResetDigLvlPauseTrigSrc(self.task_handle)

    def GetDigLvlPauseTrigWhen(self, data):
        return DAQmxGetDigLvlPauseTrigWhen(self.task_handle, data)

    def SetDigLvlPauseTrigWhen(self, data):
        return DAQmxSetDigLvlPauseTrigWhen(self.task_handle, data)

    def ResetDigLvlPauseTrigWhen(self):
        return DAQmxResetDigLvlPauseTrigWhen(self.task_handle)

    def GetDigLvlPauseTrigDigFltrEnable(self, data):
        return DAQmxGetDigLvlPauseTrigDigFltrEnable(self.task_handle, data)

    def SetDigLvlPauseTrigDigFltrEnable(self, data):
        return DAQmxSetDigLvlPauseTrigDigFltrEnable(self.task_handle, data)

    def ResetDigLvlPauseTrigDigFltrEnable(self):
        return DAQmxResetDigLvlPauseTrigDigFltrEnable(self.task_handle)

    def GetDigLvlPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetDigLvlPauseTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetDigLvlPauseTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetDigLvlPauseTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetDigLvlPauseTrigDigFltrMinPulseWidth(self):
        return DAQmxResetDigLvlPauseTrigDigFltrMinPulseWidth(self.task_handle)

    def GetDigLvlPauseTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetDigLvlPauseTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetDigLvlPauseTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetDigLvlPauseTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetDigLvlPauseTrigDigFltrTimebaseSrc(self):
        return DAQmxResetDigLvlPauseTrigDigFltrTimebaseSrc(self.task_handle)

    def GetDigLvlPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetDigLvlPauseTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetDigLvlPauseTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetDigLvlPauseTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetDigLvlPauseTrigDigFltrTimebaseRate(self):
        return DAQmxResetDigLvlPauseTrigDigFltrTimebaseRate(self.task_handle)

    def GetDigLvlPauseTrigDigSyncEnable(self, data):
        return DAQmxGetDigLvlPauseTrigDigSyncEnable(self.task_handle, data)

    def SetDigLvlPauseTrigDigSyncEnable(self, data):
        return DAQmxSetDigLvlPauseTrigDigSyncEnable(self.task_handle, data)

    def ResetDigLvlPauseTrigDigSyncEnable(self):
        return DAQmxResetDigLvlPauseTrigDigSyncEnable(self.task_handle)

    def GetDigPatternPauseTrigSrc(self, data, bufferSize):
        return DAQmxGetDigPatternPauseTrigSrc(self.task_handle, data, bufferSize)

    def SetDigPatternPauseTrigSrc(self, data):
        return DAQmxSetDigPatternPauseTrigSrc(self.task_handle, data)

    def ResetDigPatternPauseTrigSrc(self):
        return DAQmxResetDigPatternPauseTrigSrc(self.task_handle)

    def GetDigPatternPauseTrigPattern(self, data, bufferSize):
        return DAQmxGetDigPatternPauseTrigPattern(self.task_handle, data, bufferSize)

    def SetDigPatternPauseTrigPattern(self, data):
        return DAQmxSetDigPatternPauseTrigPattern(self.task_handle, data)

    def ResetDigPatternPauseTrigPattern(self):
        return DAQmxResetDigPatternPauseTrigPattern(self.task_handle)

    def GetDigPatternPauseTrigWhen(self, data):
        return DAQmxGetDigPatternPauseTrigWhen(self.task_handle, data)

    def SetDigPatternPauseTrigWhen(self, data):
        return DAQmxSetDigPatternPauseTrigWhen(self.task_handle, data)

    def ResetDigPatternPauseTrigWhen(self):
        return DAQmxResetDigPatternPauseTrigWhen(self.task_handle)

    def GetArmStartTrigType(self, data):
        return DAQmxGetArmStartTrigType(self.task_handle, data)

    def SetArmStartTrigType(self, data):
        return DAQmxSetArmStartTrigType(self.task_handle, data)

    def ResetArmStartTrigType(self):
        return DAQmxResetArmStartTrigType(self.task_handle)

    def GetArmStartTerm(self, data, bufferSize):
        return DAQmxGetArmStartTerm(self.task_handle, data, bufferSize)

    def GetDigEdgeArmStartTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeArmStartTrigSrc(self.task_handle, data, bufferSize)

    def SetDigEdgeArmStartTrigSrc(self, data):
        return DAQmxSetDigEdgeArmStartTrigSrc(self.task_handle, data)

    def ResetDigEdgeArmStartTrigSrc(self):
        return DAQmxResetDigEdgeArmStartTrigSrc(self.task_handle)

    def GetDigEdgeArmStartTrigEdge(self, data):
        return DAQmxGetDigEdgeArmStartTrigEdge(self.task_handle, data)

    def SetDigEdgeArmStartTrigEdge(self, data):
        return DAQmxSetDigEdgeArmStartTrigEdge(self.task_handle, data)

    def ResetDigEdgeArmStartTrigEdge(self):
        return DAQmxResetDigEdgeArmStartTrigEdge(self.task_handle)

    def GetDigEdgeArmStartTrigDigFltrEnable(self, data):
        return DAQmxGetDigEdgeArmStartTrigDigFltrEnable(self.task_handle, data)

    def SetDigEdgeArmStartTrigDigFltrEnable(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigFltrEnable(self.task_handle, data)

    def ResetDigEdgeArmStartTrigDigFltrEnable(self):
        return DAQmxResetDigEdgeArmStartTrigDigFltrEnable(self.task_handle)

    def GetDigEdgeArmStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxGetDigEdgeArmStartTrigDigFltrMinPulseWidth(self.task_handle, data)

    def SetDigEdgeArmStartTrigDigFltrMinPulseWidth(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigFltrMinPulseWidth(self.task_handle, data)

    def ResetDigEdgeArmStartTrigDigFltrMinPulseWidth(self):
        return DAQmxResetDigEdgeArmStartTrigDigFltrMinPulseWidth(self.task_handle)

    def GetDigEdgeArmStartTrigDigFltrTimebaseSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseSrc(self.task_handle, data, bufferSize)

    def SetDigEdgeArmStartTrigDigFltrTimebaseSrc(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseSrc(self.task_handle, data)

    def ResetDigEdgeArmStartTrigDigFltrTimebaseSrc(self):
        return DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseSrc(self.task_handle)

    def GetDigEdgeArmStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxGetDigEdgeArmStartTrigDigFltrTimebaseRate(self.task_handle, data)

    def SetDigEdgeArmStartTrigDigFltrTimebaseRate(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigFltrTimebaseRate(self.task_handle, data)

    def ResetDigEdgeArmStartTrigDigFltrTimebaseRate(self):
        return DAQmxResetDigEdgeArmStartTrigDigFltrTimebaseRate(self.task_handle)

    def GetDigEdgeArmStartTrigDigSyncEnable(self, data):
        return DAQmxGetDigEdgeArmStartTrigDigSyncEnable(self.task_handle, data)

    def SetDigEdgeArmStartTrigDigSyncEnable(self, data):
        return DAQmxSetDigEdgeArmStartTrigDigSyncEnable(self.task_handle, data)

    def ResetDigEdgeArmStartTrigDigSyncEnable(self):
        return DAQmxResetDigEdgeArmStartTrigDigSyncEnable(self.task_handle)

    def GetTriggerSyncType(self, data):
        return DAQmxGetTriggerSyncType(self.task_handle, data)

    def SetTriggerSyncType(self, data):
        return DAQmxSetTriggerSyncType(self.task_handle, data)

    def ResetTriggerSyncType(self):
        return DAQmxResetTriggerSyncType(self.task_handle)

    def GetWatchdogTimeout(self, data):
        return DAQmxGetWatchdogTimeout(self.task_handle, data)

    def SetWatchdogTimeout(self, data):
        return DAQmxSetWatchdogTimeout(self.task_handle, data)

    def ResetWatchdogTimeout(self):
        return DAQmxResetWatchdogTimeout(self.task_handle)

    def GetWatchdogExpirTrigType(self, data):
        return DAQmxGetWatchdogExpirTrigType(self.task_handle, data)

    def SetWatchdogExpirTrigType(self, data):
        return DAQmxSetWatchdogExpirTrigType(self.task_handle, data)

    def ResetWatchdogExpirTrigType(self):
        return DAQmxResetWatchdogExpirTrigType(self.task_handle)

    def GetWatchdogExpirTrigTrigOnNetworkConnLoss(self, data):
        return DAQmxGetWatchdogExpirTrigTrigOnNetworkConnLoss(self.task_handle, data)

    def SetWatchdogExpirTrigTrigOnNetworkConnLoss(self, data):
        return DAQmxSetWatchdogExpirTrigTrigOnNetworkConnLoss(self.task_handle, data)

    def ResetWatchdogExpirTrigTrigOnNetworkConnLoss(self):
        return DAQmxResetWatchdogExpirTrigTrigOnNetworkConnLoss(self.task_handle)

    def GetDigEdgeWatchdogExpirTrigSrc(self, data, bufferSize):
        return DAQmxGetDigEdgeWatchdogExpirTrigSrc(self.task_handle, data, bufferSize)

    def SetDigEdgeWatchdogExpirTrigSrc(self, data):
        return DAQmxSetDigEdgeWatchdogExpirTrigSrc(self.task_handle, data)

    def ResetDigEdgeWatchdogExpirTrigSrc(self):
        return DAQmxResetDigEdgeWatchdogExpirTrigSrc(self.task_handle)

    def GetDigEdgeWatchdogExpirTrigEdge(self, data):
        return DAQmxGetDigEdgeWatchdogExpirTrigEdge(self.task_handle, data)

    def SetDigEdgeWatchdogExpirTrigEdge(self, data):
        return DAQmxSetDigEdgeWatchdogExpirTrigEdge(self.task_handle, data)

    def ResetDigEdgeWatchdogExpirTrigEdge(self):
        return DAQmxResetDigEdgeWatchdogExpirTrigEdge(self.task_handle)

    def GetWatchdogDOExpirState(self, lines, data):
        return DAQmxGetWatchdogDOExpirState(self.task_handle, lines, data)

    def SetWatchdogDOExpirState(self, lines, data):
        return DAQmxSetWatchdogDOExpirState(self.task_handle, lines, data)

    def ResetWatchdogDOExpirState(self, lines):
        return DAQmxResetWatchdogDOExpirState(self.task_handle, lines)

    def GetWatchdogAOOutputType(self, lines, data):
        return DAQmxGetWatchdogAOOutputType(self.task_handle, lines, data)

    def SetWatchdogAOOutputType(self, lines, data):
        return DAQmxSetWatchdogAOOutputType(self.task_handle, lines, data)

    def ResetWatchdogAOOutputType(self, lines):
        return DAQmxResetWatchdogAOOutputType(self.task_handle, lines)

    def GetWatchdogAOExpirState(self, lines, data):
        return DAQmxGetWatchdogAOExpirState(self.task_handle, lines, data)

    def SetWatchdogAOExpirState(self, lines, data):
        return DAQmxSetWatchdogAOExpirState(self.task_handle, lines, data)

    def ResetWatchdogAOExpirState(self, lines):
        return DAQmxResetWatchdogAOExpirState(self.task_handle, lines)

    def GetWatchdogCOExpirState(self, lines, data):
        return DAQmxGetWatchdogCOExpirState(self.task_handle, lines, data)

    def SetWatchdogCOExpirState(self, lines, data):
        return DAQmxSetWatchdogCOExpirState(self.task_handle, lines, data)

    def ResetWatchdogCOExpirState(self, lines):
        return DAQmxResetWatchdogCOExpirState(self.task_handle, lines)

    def GetWatchdogHasExpired(self, data):
        return DAQmxGetWatchdogHasExpired(self.task_handle, data)

    def GetWriteRelativeTo(self, data):
        return DAQmxGetWriteRelativeTo(self.task_handle, data)

    def SetWriteRelativeTo(self, data):
        return DAQmxSetWriteRelativeTo(self.task_handle, data)

    def ResetWriteRelativeTo(self):
        return DAQmxResetWriteRelativeTo(self.task_handle)

    def GetWriteOffset(self, data):
        return DAQmxGetWriteOffset(self.task_handle, data)

    def SetWriteOffset(self, data):
        return DAQmxSetWriteOffset(self.task_handle, data)

    def ResetWriteOffset(self):
        return DAQmxResetWriteOffset(self.task_handle)

    def GetWriteRegenMode(self, data):
        return DAQmxGetWriteRegenMode(self.task_handle, data)

    def SetWriteRegenMode(self, data):
        return DAQmxSetWriteRegenMode(self.task_handle, data)

    def ResetWriteRegenMode(self):
        return DAQmxResetWriteRegenMode(self.task_handle)

    def GetWriteCurrWritePos(self, data):
        return DAQmxGetWriteCurrWritePos(self.task_handle, data)

    def GetWriteOvercurrentChansExist(self, data):
        return DAQmxGetWriteOvercurrentChansExist(self.task_handle, data)

    def GetWriteOvercurrentChans(self, data, bufferSize):
        return DAQmxGetWriteOvercurrentChans(self.task_handle, data, bufferSize)

    def GetWriteOvertemperatureChansExist(self, data):
        return DAQmxGetWriteOvertemperatureChansExist(self.task_handle, data)

    def GetWriteOpenCurrentLoopChansExist(self, data):
        return DAQmxGetWriteOpenCurrentLoopChansExist(self.task_handle, data)

    def GetWriteOpenCurrentLoopChans(self, data, bufferSize):
        return DAQmxGetWriteOpenCurrentLoopChans(self.task_handle, data, bufferSize)

    def GetWritePowerSupplyFaultChansExist(self, data):
        return DAQmxGetWritePowerSupplyFaultChansExist(self.task_handle, data)

    def GetWritePowerSupplyFaultChans(self, data, bufferSize):
        return DAQmxGetWritePowerSupplyFaultChans(self.task_handle, data, bufferSize)

    def GetWriteSpaceAvail(self, data):
        return DAQmxGetWriteSpaceAvail(self.task_handle, data)

    def GetWriteTotalSampPerChanGenerated(self, data):
        return DAQmxGetWriteTotalSampPerChanGenerated(self.task_handle, data)

    def GetWriteRawDataWidth(self, data):
        return DAQmxGetWriteRawDataWidth(self.task_handle, data)

    def GetWriteNumChans(self, data):
        return DAQmxGetWriteNumChans(self.task_handle, data)

    def GetWriteWaitMode(self, data):
        return DAQmxGetWriteWaitMode(self.task_handle, data)

    def SetWriteWaitMode(self, data):
        return DAQmxSetWriteWaitMode(self.task_handle, data)

    def ResetWriteWaitMode(self):
        return DAQmxResetWriteWaitMode(self.task_handle)

    def GetWriteSleepTime(self, data):
        return DAQmxGetWriteSleepTime(self.task_handle, data)

    def SetWriteSleepTime(self, data):
        return DAQmxSetWriteSleepTime(self.task_handle, data)

    def ResetWriteSleepTime(self):
        return DAQmxResetWriteSleepTime(self.task_handle)

    def GetWriteNextWriteIsLast(self, data):
        return DAQmxGetWriteNextWriteIsLast(self.task_handle, data)

    def SetWriteNextWriteIsLast(self, data):
        return DAQmxSetWriteNextWriteIsLast(self.task_handle, data)

    def ResetWriteNextWriteIsLast(self):
        return DAQmxResetWriteNextWriteIsLast(self.task_handle)

    def GetWriteDigitalLinesBytesPerChan(self, data):
        return DAQmxGetWriteDigitalLinesBytesPerChan(self.task_handle, data)

    def GetWriteAccessoryInsertionOrRemovalDetected(self, data):
        return DAQmxGetWriteAccessoryInsertionOrRemovalDetected(self.task_handle, data)

    def GetWriteDevsWithInsertedOrRemovedAccessories(self, data, bufferSize):
        return DAQmxGetWriteDevsWithInsertedOrRemovedAccessories(self.task_handle, data, bufferSize)

    def GetSampClkTimingResponseMode(self, data):
        return DAQmxGetSampClkTimingResponseMode(self.task_handle, data)

    def SetSampClkTimingResponseMode(self, data):
        return DAQmxSetSampClkTimingResponseMode(self.task_handle, data)

    def ResetSampClkTimingResponseMode(self):
        return DAQmxResetSampClkTimingResponseMode(self.task_handle)

    stop = StopTask
    start = StartTask

    def __del__(self) -> None:
        try:
            self.ClearTask()
        except DAQError:
            pass
